/***********************************************************************

	File:	polytope.h
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	An abstract interface to a polytope.

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#ifndef POLYTOPE_H
#define	POLYTOPE_H

#include "bitmaskmacros.h"

/*
 * An object to abstractly represent a concrete polytope.
 */

struct Polytope {
	const char *	name;	/* Name of this polytope */
	int		dim;	/* Dimensionality of points of the	*/
				/* space within which this polytope	*/
				/* lives.  Not to be confused with	*/
				/* the dimensionality of the polytope	*/
				/* itself.				*/
	const char **	ename;	/* Printable names for each of the	*/
				/* `dim' elements of the space.  These	*/
				/* should probably be valid variable	*/
				/* names.				*/
	int		maxlen;	/* Length of longest ename[]		*/

	/* Operations provided by all polytopes. */

	/* Generate all of the extreme points of this polytope, and	*/
	/* return them to the caller in the form of a set of incidence	*/
	/* vectors.							*/
	struct IVecSet *
			(*get_extreme_points) (const struct Polytope * poly);

	/* Given the "name" or printable description of a constraint	*/
	/* (usually a facet) of this polytope, return a Constraint for	*/
	/* it.  Returns NULL if the particular name is not understodd.	*/
	struct Constraint *
			(*get_constraint) (const struct Polytope *	poly,
					   const char *			name);

#if 0
	/* Return the list of equations (if any) that define the affine	*/
	/* hull of this polytope.					*/
	struct Constraint *
			(*affine_hull) (struct Polytope * poly);
#endif

	/* Free up this polytope.  This is an intensionally virtual	*/
	/* destructor so that different implementations can have	*/
	/* additional data that this routine also frees up.		*/
	void (*free) (struct Polytope * poly);
};

/*
 * An object to represent a set of incidence vectors.
 */

struct IVecSet {
	int	nvecs;		/* Num incidence vectors */
	int	dim;		/* Num elements in each incidence vector */
	int	nmasks;		/* Num mask words in each incidence vector */
	bitmap_t *
		array;		/* Array of incidence vector bit-mask words */
};

/*
 * A structure to represent a single constraint, in a form that makes
 * it relatively easy to incorporate into a CPLEX formulation.
 *
 * There is also a `next' field making it easy to thread these together
 * into a linked list of constraints.
 */

struct Constraint {
	int	 nz;		/* Number of non-zeros on left-hand-side */
	int *	 rmatind;	/* Column index of each non-zero on LHS */
	double * rmatval;	/* Non-zero values for LHS */
	char	 sense;		/* Operator sense of this constraint */
	double	 rhs;		/* RHS value */
	struct Constraint *
		 next;		/* Next constraint in list */
};

extern void	clear_polytope (struct Polytope * poly);
extern void	free_constraint_list (struct Constraint * list);
extern void	free_incidence_vectors (struct IVecSet * ivecs);
extern int	get_extreme_points_incident_to_constraint (
				int **				list,
				const struct Constraint *	cp,
				const struct IVecSet *		xpts);
extern struct Constraint *
		make_constraint (int		nz,
				 const int *	rmatind,
				 const double *	rmatval,
				 char		sense,
				 double		rhs);
extern struct Polytope *
		make_polytope_by_name (const char * name);

#endif
