/***********************************************************************

	File:	tsp_indicators.h
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute strength indicators for polytope TSP(n) via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#ifndef TSP_INDICATORS_H
#define	TSP_INDICATORS_H

#include <gmp.h>

/*
 * Return the number of extreme points of TSP(n).
 */

extern void	tsp_num_extreme_points (mpz_ptr res, int n);

/*
 * Let e \in E.  Return the number of TSP(n) extreme points that contain
 * edge e.
 */

extern void	tsp_num_edge_extreme_points (mpz_ptr res, int n);

/*
 * Return the number of TSP(n) extreme points incident to the
 * non-negativity inequality for some edge e.
 */

extern void	tsp_nonneg_ep_count (mpz_ptr		res,
				     int		n);

/*
 * Return the Extreme Point Ratio of the TSP(n) non-negativity inequality
 * for some edge e.
 */

extern void	tsp_nonneg_epr (mpq_ptr		res,
				int		n);

/*
 * Return the Centroid Distance squared of the TSP(n) non-negativity
 * inequality for some edge e.
 */

extern void	tsp_nonneg_cd2 (mpq_ptr		res,
				int		n);

/*
 * Return the number of TSP(n) extreme points incident to a
 * subtour inequality S of cardinality k.
 */

extern void	tsp_subtour_ep_count (mpz_ptr		res,
				      int		n,
				      int		k);

/*
 * Return the Extreme Point Ratio of a TSP(n) subtour S of cardinality k.
 */

extern void	tsp_subtour_epr (mpq_ptr	res,
				 int		n,
				 int		k);

/*
 * Return the Centroid Distance squared of a TSP(n) subtour S of
 * cardinality k.
 */

extern void	tsp_subtour_cd2 (mpq_ptr	res,
				 int		n,
				 int		k);

/*
 * Return the number of TSP(n) extreme points incident to a
 * 3-toothed comb inequality.
 */

extern void	tsp_3tc_ep_count (mpz_ptr	res,
				  int		b1,
				  int		t1,
				  int		b2,
				  int		t2,
				  int		b3,
				  int		t3,
				  int		h,
				  int		o);

/*
 * Return the Extreme Point Ratio of a TSP(n) 3-toothed comb inequality.
 *
 * Note: the closed-form formula for EPR of 3-Toothed Combs is WRONG!
 * There are tours incident to the comb whose structure does not match
 * any of the "primtive comb" tours, so this formula undercounts the
 * incident tours.
 */

extern void	tsp_3tc_epr (mpq_ptr	res,
			     int	b1,
			     int	t1,
			     int	b2,
			     int	t2,
			     int	b3,
			     int	t3,
			     int	h,
			     int	o);

/*
 * Return the Centroid Distance squared of a TSP(n) 3-toothed comb
 * inequality.
 */

extern void	tsp_3tc_cd2 (mpq_ptr	res,
			     int	b1,
			     int	t1,
			     int	b2,
			     int	t2,
			     int	b3,
			     int	t3,
			     int	h,
			     int	o);

/*
 * Shut down the tsp_indicators module, freeing all memory it
 * statically holds.
 */

extern void	tsp_indicators_shutdown (void);

#endif
