/***********************************************************************

	$Id: binomial.c,v 1.1 2025/01/26 02:31:43 warme Exp $

	File:	binomial.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Binomial coefficients.

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "binomial.h"

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "new.h"
#include <stdlib.h>
#include <string.h>

/*
 * Local Types
 */

/*
 * An object to hold binomial coefficients.
 */

struct Binomial {
	int		n_alloc;	/* Largest row allocated to hold */
	int		n_fill;		/* Largest row currently computed */
	mpz_ptr *	rows;		/* Array of binomial rows */
};

/*
 * Local Functions
 */

static void			binomial_fill (struct Binomial *	bp,
						int			n);
static struct Binomial *	binomial_get (int n);

/*
 * Local Data
 */

static struct Binomial *	binomial_buf;

/*
 * Return the binomial coefficient "n choose k".
 */

	void
binomial (

mpz_ptr			res,		/* IN/OUT: Result */
int			n,		/* IN: Set cardinality */
int			k		/* IN: Number to choose */
)
{
struct Binomial *	bp;
mpz_ptr			row;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);
	FATAL_ERROR_IF (k > n);

	bp = binomial_get (n);

	row	= bp -> rows [n];

	mpz_set (res, &row [k]);
}

/*
 * Shut down binomial coefficients module, freeing all statically-held memory.
 */

	void
binomial_shutdown (void)

{
int			i, n_alloc, n_fill;
struct Binomial *	bp;
mpz_ptr *		rows;
mpz_ptr			row;

	bp = binomial_buf;
	if (bp NE NULL) {
		binomial_buf	= NULL;

		n_alloc	= bp -> n_alloc;
		n_fill	= bp -> n_fill;
		rows	= bp -> rows;
		FATAL_ERROR_IF (n_alloc < 0);

		for (i = 0; i <= n_fill; i++) {
			row = rows [i];
			rows [i] = NULL;

			FATAL_ERROR_IF (row EQ NULL);

			free_Z_vector (row, i + 1);
		}
		free (rows);

		memset (bp, 0, sizeof (*bp));
		free (bp);
	}
}

/*
 * Return the Binomial buffer, assuring that all rows up to cardinality
 * n have been computed.
 */

	static
	struct Binomial *
binomial_get (

int		n		/* IN: Cardinality required */
)
{
mpz_ptr *		rows;
struct Binomial *	bp;

	FATAL_ERROR_IF (n < 0);

	bp = binomial_buf;
	if (bp EQ NULL) {
		rows = NEWA (n + 1, mpz_ptr);
		memset (rows, 0, (n + 1) * sizeof (rows [0]));

		bp = NEW (struct Binomial);
		memset (bp, 0, sizeof (*bp));

		bp -> n_alloc	= n;
		bp -> n_fill	= -1;
		bp -> rows	= rows;

		binomial_buf = bp;
	}

	/* Assure rows up to n have been filled. */
	binomial_fill (bp, n);

	return (bp);
}

/*
 * Assure that binomial coefficient rows have been calculated for all sets
 * up to cardinality n.
 */

	static
	void
binomial_fill (

struct Binomial *	bp,		/* IN/OUT: Binomial buffer to fill */
int			n		/* IN: Max cardinality set to fill */
)
{
int		i, k, n_alloc, n_fill;
mpz_ptr		row, prev_row;
mpz_ptr *	rows;
mpz_ptr *	rows2;

	FATAL_ERROR_IF (bp EQ NULL);
	FATAL_ERROR_IF (n < 0);

	n_alloc	= bp -> n_alloc;
	n_fill	= bp -> n_fill;
	rows	= bp -> rows;

	FATAL_ERROR_IF (n_fill > n_alloc);

	if (n > n_alloc) {
		/* Must re-allocate the buffer. */
		rows2 = NEWA (n + 1, mpz_ptr);
		memcpy (rows2, rows, (n_fill + 1) * sizeof (rows2 [0]));
		memset (&rows2 [n_fill + 1],
			0,
			(n - n_fill) * sizeof (rows2 [0]));
		bp -> n_alloc	= n;
		bp -> rows	= rows2;
		memset (rows, 0, (n_fill + 1) * sizeof (rows [0]));
		free (rows);
		rows = rows2;
	}

	i = n_fill;
	if (i < n) {
		while (i < n) {
			++i;
			row = new_Z_vector (i + 1);
			mpz_set_ui (&row [i], 1);
			if (i > 0) {
				/* Initialize row i > 0. */
				prev_row = rows [i - 1];
				mpz_set_ui (&row [0], 1);
				for (k = 1; k < i; k++) {
					mpz_add (&row [k],
						 &prev_row [k - 1],
						 &prev_row [k]);
				}
			}
			rows [i] = row;
		}
		bp -> n_fill = n;
	}
}
