/***********************************************************************

	File:	polytope.c
	Rev:	a-3
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	An abstract interface to a polytope.

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-2:	08/09/2021	warme
		: Added STGP(n).
	a-3:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "polytope.h"

#include "fatal.h"
#include "logic.h"
#include "new.h"
#include "scanutils.h"
#include <stdio.h>
#include <stdlib.h>
#include "stgp_polytope.h"
#include "sthgp_polytope.h"
#include "stnp_polytope.h"
#include <string.h>
#include "tsp_polytope.h"

extern char *	me;

/*
 * Free up the space referenced by the given Polytope object, but
 * don't free the Polytope object itself, since it probably lives
 * inside of a larger object.
 *
 * NOTE: THIS IS A *BASE CLASS* ROUTINE!
 */

	void
clear_polytope (

struct Polytope *	poly	/* IN: Polytope to clear out */
)
{
int		i, dim;
const char **	ename;
const char *	name;

	if (poly NE NULL) {
		if (poly -> name NE NULL) {
			free ((void *) (poly -> name));
		}
		dim	= poly -> dim;
		ename	= poly -> ename;
		if (ename NE NULL) {
			for (i = 0; i < dim; i++) {
				name = ename [i];
				if (name NE NULL) {
					free ((void *) name);
					ename [i] = NULL;
				}
			}
			free (ename);
		}
		memset (poly, 0, sizeof (*poly));
	}
}

/*
 * Free up the given set of incidence vectors.
 */

	void
free_incidence_vectors (

struct IVecSet *	ivecs	/* IN: incidence vector set to free */
)
{
	if (ivecs NE NULL) {
		if (ivecs -> array NE NULL) {
			free (ivecs -> array);
			ivecs -> array = NULL;
		}
		ivecs -> nvecs	= -1;
		ivecs -> dim	= -1;
		ivecs -> nmasks	= -1;

		free (ivecs);
	}
}

/*
 * Generate a list of all extreme points that are incident to the
 * given constraint.
 */

	int
get_extreme_points_incident_to_constraint (

int **				_list,	/* IN/OUT: list of extreme point */
					/*	   indices */
const struct Constraint *	cp,	/* IN: constraint to test */
const struct IVecSet *		xpts	/* IN: extreme point incidence vectors */
)
{
int		i, j, k, count;
int		numxp, nmasks, nz;
int *		list;
int *		list2;
bitmap_t *	tp;
const int *	rmatind;
const double *	rmatval;
char		sense;
double		sum, rhs;

	FATAL_ERROR_IF ((cp EQ NULL) OR (xpts EQ NULL));

	numxp	= xpts -> nvecs;
	nmasks	= xpts -> nmasks;

	nz	= cp -> nz;
	rmatind	= cp -> rmatind;
	rmatval	= cp -> rmatval;
	sense	= cp -> sense;
	rhs	= cp -> rhs;

	/* Make maximum-size buffer -- will shorten it below. */
	list = NEWA (numxp, int);
	count = 0;
	tp = &(xpts -> array [0]);
	for (i = 0; i < numxp; i++, tp += nmasks) {
		sum = 0.0;
		for (k = 0; k < nz; k++) {
			j = rmatind [k];
			if (NOT BITON (tp, j)) continue;
			sum += rmatval [k];
		}

		/* Validate that constraint is not *violated*! */
		switch (sense) {
		case 'G':
			FATAL_ERROR_IF (sum < rhs);
			break;

		case 'L':
			FATAL_ERROR_IF (sum > rhs);
			break;

		default:
			FATAL_ERROR;
			break;
		}

		if (sum NE rhs) continue;

		list [count] = i;
		++count;
	}

	if (_list NE NULL) {
		list2 = NEWA (count, int);
		memcpy (list2, list, count * sizeof (int));
		free (list);

		*_list = list2;
	}

	return (count);
}

/*
 * Free up the given list of Constraints.
 */

	void
free_constraint_list (

struct Constraint *	list	/* IN: list of constraints to free */
)
{
struct Constraint *	cp;

	while (list NE NULL) {
		cp = list;
		list = cp -> next;

		if (cp -> rmatind NE NULL) {
			free (cp -> rmatind);
		}
		if (cp -> rmatval NE NULL) {
			free (cp -> rmatval);
		}
		memset (cp, 0, sizeof (*cp));
		free (cp);
	}
}

/*
 * Make a constraint with the given data.
 */

	struct Constraint *
make_constraint (

int		nz,		/* IN: Number of non-zeros on LHS */
const int *	_rmatind,	/* IN: Column indices of LHS non-zeros */
const double *	_rmatval,	/* IN: Non-zero values for LHS */
char		sense,		/* IN: Sense of constraint */
double		rhs		/* IN: Right-hand-side value */
)
{
struct Constraint *	cp;
int *			rmatind;
double *		rmatval;

	rmatind	= NEWA (nz, int);
	rmatval	= NEWA (nz, double);

	memcpy (rmatind, _rmatind, nz * sizeof (int));
	memcpy (rmatval, _rmatval, nz * sizeof (double));

	cp = NEW (struct Constraint);

	memset (cp, 0, sizeof (*cp));

	cp -> nz	= nz;
	cp -> rmatind	= rmatind;
	cp -> rmatval	= rmatval;
	cp -> sense	= sense;
	cp -> rhs	= rhs;
	cp -> next	= NULL;

	return (cp);
}

/*
 * Decode the polytope "name" string, and create the proper
 * type of polytope.
 */

	struct Polytope *
make_polytope_by_name (

const char *	name		/* IN: name of polytope */
)
{
int			n, nterms, nsteins;
const char *		s;
struct Polytope *	poly;

#define	LPAREN_STRING	"("
#define RPAREN_STRING	")"

	if (strncasecmp (name, "tsp" LPAREN_STRING, 4) EQ 0) {
		s = get_int (name + 4, &n);
		if ((s EQ NULL) OR
		    (strcmp (s, RPAREN_STRING) NE 0)) {
			fprintf (stderr,
				 "%s: Garbled polytope size: %s\n",
				 me, name);
			exit (1);
		}
		poly = make_TSP_polytope (n);
		if (poly EQ NULL) {
			fprintf (stderr,
				 "%s: invalid TSP polytope size\n", me);
			exit (1);
		}
		return (poly);
	}

	if (strncasecmp (name, "sthgp" LPAREN_STRING, 6) EQ 0) {
		s = get_int (name + 6, &n);
		if ((s EQ NULL) OR
		    (strcmp (s, RPAREN_STRING) NE 0)) {
			fprintf (stderr,
				 "%s: Garbled polytope size: %s\n",
				 me, name);
			return (NULL);
		}
		poly = make_STHGP_polytope (n);
		if (poly EQ NULL) {
			fprintf (stderr,
				 "%s: invalid STHGP polytope size\n", me);
		}
		return (poly);
	}

	if (strncasecmp (name, "stgp" LPAREN_STRING, 5) EQ 0) {
		s = get_int (name + 5, &n);
		if ((s EQ NULL) OR
		    (strcmp (s, RPAREN_STRING) NE 0)) {
			fprintf (stderr,
				 "%s: Garbled polytope size: %s\n",
				 me, name);
			return (NULL);
		}
		poly = make_STGP_polytope (n);
		if (poly EQ NULL) {
			fprintf (stderr,
				 "%s: invalid STGP polytope size\n", me);
		}
		return (poly);
	}

	if (strncasecmp (name, "stngp" LPAREN_STRING, 6) EQ 0) {
		s = get_int (name + 6, &nterms);
		if ((s EQ NULL) OR (s [0] NE ',')) {
			fprintf (stderr,
				 "%s: invalid STNGP polytope num terminals\n",
				 me);
			return (NULL);
		}
		s = get_int (s + 1, &nsteins);
		if ((s EQ NULL) OR (strcmp (s, RPAREN_STRING) NE 0)) {
			fprintf (stderr,
				 "%s: Garbled polytope num Steiners\n", me);
			return (NULL);
		}
		poly = make_STNGP_polytope (nterms, nsteins);
		if (poly EQ NULL) {
			fprintf (stderr, "%s: invalid STNGP(n,m) args\n", me);
		}
		return (poly);
	}

	if (strncasecmp (name, "stnhgp" LPAREN_STRING, 7) EQ 0) {
		s = get_int (name + 7, &nterms);
		if ((s EQ NULL) OR (s [0] NE ',')) {
			fprintf (stderr,
				 "%s: invalid STNHGP polytope num terminals\n",
				 me);
			return (NULL);
		}
		s = get_int (s + 1, &nsteins);
		if ((s EQ NULL) OR (strcmp (s, RPAREN_STRING) NE 0)) {
			fprintf (stderr,
				 "%s: Garbled polytope num Steiners\n", me);
			return (NULL);
		}
		poly = make_STNHGP_polytope (nterms, nsteins);
		poly = make_STNGP_polytope (nterms, nsteins);
		if (poly EQ NULL) {
			fprintf (stderr, "%s: invalid STNHGP(n,m) args\n", me);
		}
		return (poly);
	}

	fprintf (stderr, "%s: Unknown polytope `%s'\n", me, name);
	return (NULL);

#undef LPAREN_STRING
#undef RPAREN_STRING
}
