/***********************************************************************

	$Id: qpdist2.c,v 1.1 2025/01/26 02:31:43 warme Exp $

	File:	qpdist2.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Compute Centroid Distance (squared) via quadratic programming.

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "qpdist2.h"

#include "config.h"

#ifdef CPLEX

#include "cplex.h"
#include "fatal.h"
#include "logic.h"
#include "new.h"
#include "polytope.h"
#include <string.h>

/*
 * Local Functions
 */

static double *		compute_centroid (struct IVecSet * xpts);
static double		compute_min_dist2 (double *		P,
					   double *		alpha,
					   int			nf,
					   const int *		list,
					   const struct IVecSet * xpts,
					   const double *	C,
					   enum CD_Type		cd_type);

/*
 * Return the Centroid Distance squared for the given inequality with
 * respect to the given polytope.  This is a numeric approximation
 * (computed via quadratic programming), not an exact result.
 */
	double
cd2_via_qp (

const struct Constraint *	cp,	/* IN: Constraint/inequality */
const struct Polytope *		poly,	/* IN: Polytope */
enum CD_Type			cd_type	/* IN: Type of CD to compute */
)
{
int			dim, nf;
int *			list;
struct IVecSet *	xpts;
double *		C;
double *		P;
double *		alpha;
double			dist2;

	FATAL_ERROR_IF ((cp EQ NULL) OR (poly EQ NULL));

	dim	= poly -> dim;

	/* Generate all of the extreme points of this polytope. */
	xpts = poly -> get_extreme_points (poly);

	nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);

	C = compute_centroid (xpts);

	/* Compute the minimum distance from centroid to the	*/
	/* chosen hull (affine or convex) of the given set of	*/
	/* incidence vectors.  Return the point P that is	*/
	/* closest to C.					*/
	P = NEWA (dim, double);
	alpha = NEWA (nf, double);
	dist2 = compute_min_dist2 (P, alpha, nf, list, xpts, C, cd_type);

	free (alpha);
	free (P);
	free (C);
	free (list);

	free_incidence_vectors (xpts);

	return (dist2);
}

/*
 * Compute the coordinates of the centroid.
 */

	static
	double *
compute_centroid (

struct IVecSet *	xpts	/* IN: set of extreme points */
)
{
int			i, j, k;
int			numxp, dim, nmasks;
const bitmap_t *	array;
const bitmap_t *	tp;
double *		C;

	numxp	= xpts -> nvecs;
	dim	= xpts -> dim;
	nmasks	= xpts -> nmasks;
	array	= xpts -> array;

	C = NEWA (dim, double);

	for (j = 0; j < dim; j++) {
		k = 0;
		tp = array;
		for (i = 0; i < numxp; i++, tp += nmasks) {
			if (BITON (tp, j)) {
				++k;
			}
		}
		C [j] = k;
		C [j] /= numxp;
	}

	return (C);
}

/*
 * Set up and solve the quadratic program that finds the minimum
 * distance-squared between the hull (either affine or convex) of the
 * given set of incidence vectors, and the centroid C.  The QP
 * has the following formulation:
 *
 *	Minimize:	Y dot Y
 *
 *	Subject To:
 *
 *		P = \sum_j t_j A_j
 *		Y = P - C
 *		\sum_j A_j = 1
 *		A >= 0			{if convex hull}
 *
 * where t_i are the incidence vectors of all extreme points that are
 * incident to the facet being studied.  Point P is either an affine
 * or convex combination of these incidence vectors.  Y = P-C, the
 * vector from our centroid to point P (in the hull), and we want to
 * minimize the magnitude of Y.
 *
 * This is actually *expressed* as follows within the QP tableaux:
 *
 *	      -P      + \sum_j t_j A_j = 0
 *	      -P  + Y		       = -C
 *			\sum_j	A_j    = 1
 *				A      >= 0	{if convex hull}
 *
 * All variables are (-infinty,+infinity), except that for the convex
 * hull case, there is one additional row and A >= 0.
 *
 * The column variables are ordered [P, Y, A], of which there are
 * dim + dim + nf of these.
 *
 * Returns the distance-squared, and also:
 *
 *	P:	Coordinates of the point P in the hull that is closest to C,
 *	alpha:	Combination of incidence vectors yielding P.
 */

	static
	double
compute_min_dist2 (

double *		P,		/* IN/OUT: coords of closest point */
double *		alpha,		/* IN/OUT: combination producing P */
int			nf,		/* IN: number of incidence vectors */
const int *		list,		/* IN: list of incidence vector */
					/*     indices */
const struct IVecSet *	xpts,		/* IN: set of extreme points */
const double *		C,		/* IN: coordinates of centroid */
enum CD_Type		cd_type		/* IN: type of centroid distance */
					/*     to compute */
)
{
int			i, j, k, item;
int			dim, nmasks;
int			ncoeff;
int			index;
int			status;
int			col;
int			mac, mar;
int			macsz, marsz, matsz;
int			objsen;
int			pfeas, dfeas;
double *		objx;
double *		qpobj;
double *		rhsx;
char *			senx;
double *		bdl;
double *		bdu;
int *			matbeg;
int *			matcnt;
int *			matind;
double *		matval;
const bitmap_t *	xp_array;
const bitmap_t *	tp;
double *		x;
CPXENVptr		cplex_env;
CPXLPptr 		lp;
double			z;

	/* Start up CPLEX... */

	/* Newer versions of CPLEX need to be initialized... */
	{ int status;
	  char msg [512];

	  cplex_env = CPXopenCPLEX (&status);
	  if (cplex_env EQ NULL) {
		if (CPXgeterrorstring (NULL, status, msg) EQ NULL) {
			strcpy (msg, "No CPLEX error message.");
		}
		fprintf (stderr, "%s\n", msg);
		/* Shut down CPLEX and get out. */
		CPXcloseCPLEX (&cplex_env);
		exit (1);
	  }

#if 0
	  /* Get rid of CPLEX's default message destinations. */
	  CPXCHANNELptr	_res, _warn, _error, _log;
	  FILE *	fp;
	  CPXgetchannels (cplex_env, &_res, &_warn, &_error, &_log);
	  CPXdisconnectchannel (cplex_env, _res);
	  CPXdisconnectchannel (cplex_env, _warn);
	  CPXdisconnectchannel (cplex_env, _error);
	  CPXdisconnectchannel (cplex_env, _log);

	  CPXsetintparam (cplex_env, CPX_PARAM_SCRIND, 0);

	  fp = fopen ("cplex.log", "a");
	  if (fp EQ NULL) {
		perror ("cplex.log");
		/* Shut down CPLEX and get out. */
		CPXcloseCPLEX (&cplex_env);
		exit (1);
	  }
	  /* Send all log stuff to the cplex.log file. */
	  CPXsetlogfile (cplex_env, fp);

	  /* But discard the results stuff... */
	  CPXdisconnectchannel (cplex_env, _res);
#endif
	}

	/* Here is the real algorithm! */

	dim	 = xpts -> dim;
	nmasks	 = xpts -> nmasks;
	xp_array = xpts -> array;

	mac	= dim + dim + nf;
	mar	= dim + dim + 1;
	marsz	= mar;
	macsz	= mac;

	/* Build the objective function... (linear portion). */
	objx = NEWA (macsz, double);
	for (i = 0; i < macsz; i++) {
		objx [i] = 0.0;
	}
	objsen = CPX_MIN;	/* Minimize.  Objective is purely quadratic. */

	/* Build the quadratic portion of the objective.	*/
	/* This is a simple diagonal matrix (separable).	*/
	qpobj = NEWA (macsz, double);
	memset (qpobj, 0, macsz * sizeof (qpobj [0]));
	for (j = 0; j < dim; j++) {
		qpobj [dim + j] = 2.0;
	}

	/* Build variable bound arrays... */
	bdl = NEWA (macsz, double);
	bdu = NEWA (macsz, double);
	/* Start with all vars unbounded. */
	for (i = 0; i < macsz; i++) {
		bdl [i] = -CPX_INFBOUND;
		bdu [i] = CPX_INFBOUND;
	}

	switch (cd_type) {
	case QPDIST2_NORMAL_CD2:
		/* Convex hull.  Make the A vars non-negative. */
		for (i = 2*dim; i < macsz; i++) {
			bdl [i] = 0.0;
		}

	case QPDIST2_WEAK_CD2:
		break;

	default:
		FATAL_ERROR;
	}

	/* Count the total number of non-zeros needed... */
	init_nbits ();
	ncoeff = 3 * dim + nf;
	for (i = 0; i < nf; i++) {
		k = list [i];
		tp = &xp_array [k * nmasks];
		for (j = 0; j < nmasks; j++) {
			ncoeff += NBITSON (tp [j]);
		}
	}

	matsz = ncoeff;

	rhsx	= NEWA (marsz, double);
	senx	= NEWA (marsz, char);
	matbeg	= NEWA (macsz, int);
	matcnt	= NEWA (macsz, int);
	matind	= NEWA (matsz, int);
	matval	= NEWA (matsz, double);

	for (i = 0; i < dim; i++) {
		rhsx [i]	= 0.0;
		senx [i]	= 'E';
		rhsx [dim + i]	= - C [i];
		senx [dim + i]	= 'E';
	}
	rhsx [2*dim] = 1.0;
	senx [2*dim] = 'E';

	/* Generate the matrix one column at a time... */
	index = 0;
	col = 0;
	/* Generate P columns. */
	for (j = 0; j < dim; j++) {
		matbeg [col] = index;

		matind [index] = j;
		matval [index] = -1.0;
		++index;

		matind [index] = dim + j;
		matval [index] = -1.0;
		++index;

		matcnt [col] = index - matbeg [col];
		++col;
	}
	/* Generate Y columns. */
	for (j = 0; j < dim; j++) {
		matbeg [col] = index;

		matind [index] = dim + j;
		matval [index] = 1.0;
		++index;

		matcnt [col] = index - matbeg [col];
		++col;
	}
	/* Generate A columns. */
	for (item = 0; item < nf; item++) {
		matbeg [col] = index;
		tp = &xp_array [list [item] * nmasks];
		for (j = 0; j < dim; j++) {
			if (BITON (tp, j)) {
				matind [index] = j;
				matval [index] = 1.0;
				++index;
			}
		}
		matind [index] = 2*dim;
		matval [index] = 1.0;
		++index;

		matcnt [col] = index - matbeg [col];
		++col;
	}

	FATAL_ERROR_IF (col NE macsz);
	FATAL_ERROR_IF (index NE matsz);

	lp = CPXcreateprob (cplex_env, &status, "my_qp");
	FATAL_ERROR_IF ((lp EQ NULL) OR (status NE 0));

	i  = CPXcopylp (cplex_env,
			lp,
			mac,
			mar,
			objsen,
			objx,
			rhsx,
			senx,
			matbeg,
			matcnt,
			matind,
			matval,
			bdl,
			bdu,
			NULL);
	FATAL_ERROR_IF (i NE 0);

	/* Load the quadratic objective matrix. */
	i = CPXcopyqpsep (cplex_env, lp, qpobj);
	FATAL_ERROR_IF (i NE 0);

	free (matval);
	free (matind);
	free (matcnt);
	free (matbeg);
	free (senx);
	free (rhsx);
	free (bdu);
	free (bdl);
	free (qpobj);
	free (objx);

	/* Optimize! */
	i = CPXqpopt (cplex_env, lp);
	FATAL_ERROR_IF (i NE 0);

	/* Get solution info. */
	i = CPXsolninfo (cplex_env, lp, NULL, NULL, &pfeas, &dfeas);
#if 0
	printf ("pfeas=%d, dfeas=%d\n", pfeas, dfeas);
#endif

	/* Get solution info. */
	x = NEWA (macsz, double);
	i = CPXsolution (cplex_env, lp, &status, &z, x, NULL, NULL, NULL);
	FATAL_ERROR_IF (i NE 0);
#if 0
	printf ("status = %d\n", status);
#endif

	for (j = 0; j < dim; j++) {
		P [j] = x [j];
	}
	for (j = 0; j < nf; j++) {
		alpha [j] = x [2*dim + j];
	}
	free (x);

	/* Shut down... */
	if (CPXcloseCPLEX (&cplex_env) NE 0) {
		fprintf (stderr, "Warning: Unable to close CPLEX.\n");
	}

	return (z);
}

#endif
