/***********************************************************************

	File:	scanutils.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Various string scanning utilities.

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "scanutils.h"

#include "fatal.h"
#include "logic.h"
#include <string.h>

/*
 * Global routines
 */

const char *	get_int (const char * s, int * value);
const char *	match_mixed_case (const char * s, const char ** table);

/*
 * Match the given string (in a case-insensitive manner) against the
 * given list of keywords.  If the given string does NOT match any
 * of the keywords, return NULL.
 *
 * Otherwise, return the first character of the given string *after*
 * the portion that matched.
 */

	const char *
match_mixed_case (

const char *		s,	/* IN: string to match */
const char **		table	/* IN: keyword table */
)
{
int			len;
const char *		keyword;

	for (;;) {
		keyword = *table++;
		if (keyword EQ NULL) break;

		len = strlen (keyword);
		if (strncasecmp (s, keyword, len) EQ 0) {
			return (s + len);
		}
	}

	/* None of the keywords matched. */
	return (NULL);
}

/*
 * Scan a string consisting of at least 1 decimal digit.
 *
 * Convert the decimal number into its integer value, and
 * return the first character after the decimal string.
 *
 * Returns NULL if there are no decimal digits at the start of the string.
 */

	const char *
get_int (

const char *		s,		/* IN: string to scan */
int *			value		/* IN/OUT: integer value */
)
{
int			ival;
char			c;

	FATAL_ERROR_IF ((s EQ NULL) OR (value EQ NULL));

	c = *s;
	if ((c < '0') OR (c > '9')) {
		*value = 0;
		return (NULL);
	}

	ival = c - '0';
	for (;;) {
		++s;
		c = *s;
		if ((c < '0') OR (c > '9')) break;
		ival = (ival * 10) + (c - '0');
	}

	*value = ival;

	return (s);
}
