/***********************************************************************

	File:	stgp_indicators.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute strength indicators for polytope STGP(n) via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "stgp_indicators.h"

#include "fatal.h"
#include <gmp.h>
#include "logic.h"

/*
 * Return the number of extreme points of STGP(n).
 */

	void
stgp_num_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n <= 0);

	mpz_ui_pow_ui (res, n, n - 2);
}

/*
 * Let e \in E.  Return the number of STHGP(n) extreme points that
 * contain edge e.
 */

	void
stgp_num_edge_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpz_t		TEMP;

	FATAL_ERROR_IF (res EQ NULL);

	if (n <= 2) {
		FATAL_ERROR_IF (n < 2);

		mpz_set_ui (res, 1);
	}
	else {
		mpz_init (TEMP);

		mpz_ui_pow_ui (TEMP, n, n - 3);
		mpz_mul_ui (res, TEMP, 2);

		mpz_clear (TEMP);
	}
}

/*
 * Return the number of STGP(n) extreme points incident to the
 * non-negativity inequality for some edge e.
 */

	void
stgp_nonneg_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpz_t		TOTAL, COUNT;

	mpz_init (TOTAL);
	mpz_init (COUNT);

	stgp_num_extreme_points (TOTAL, n);

	/* Compute number of spanning tree containing a given edge e. */
	mpz_ui_pow_ui (COUNT, n, n - 3);
	mpz_mul_ui (COUNT, COUNT, 2);

	/* Return number of spanning trees that DO NOT contain edge e. */
	mpz_sub (res, TOTAL, COUNT);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
}

/*
 * Return the Extreme Point Ratio for an STGP(n) non-negativity inequality
 * for some edge e.
 */

	void
stgp_nonneg_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpz_t		TOTAL, COUNT;
mpq_t		RATIO;

	mpq_init (RATIO);
	mpz_init (TOTAL);
	mpz_init (COUNT);

	stgp_num_extreme_points (TOTAL, n);
	stgp_nonneg_ep_count (COUNT, n);

	mpz_set (mpq_numref (RATIO), COUNT);
	mpz_set (mpq_denref (RATIO), TOTAL);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (RATIO);
}

/*
 * Return the Centroid Distance squared for an STGP(n) non-negativity
 * inequality for some edge e.
 */

	void
stgp_nonneg_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
int		s;
mpz_ptr		N, D;
mpq_t		DIST2;

	mpq_init (DIST2);

	N	= mpq_numref (DIST2);
	D	= mpq_denref (DIST2);

	/* dist2 = s*(b - a.C)^2/(q*s - r^2),				*/
	/* where: b = 0, a.C = 2/n, q=1, r=1, s={n choose 2}.		*/
	/* dist2 = s*(4/n^2) / (s-1).					*/
	s = n * (n - 1) / 2;

	mpz_set_ui (N, 4 * s);
	mpz_set_ui (D, n);
	mpz_mul_ui (D, D, n);
	mpz_mul_ui (D, D, s - 1);

	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpq_clear (DIST2);
}

/*
 * Return the number of STGP(n) extreme points incident to a
 * subtour inequality S of cardinality k.
 */

	void
stgp_subtour_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_t		TEMP1, TEMP2;

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	mpz_ui_pow_ui (TEMP1, k, k - 1);
	mpz_ui_pow_ui (TEMP2, n, n - k - 1);

	mpz_mul (res, TEMP1, TEMP2);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Return the Extreme Point Ratio for an STGP(n) subtour S of cardinality k.
 */

	void
stgp_subtour_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_t		TOTAL, COUNT;
mpq_t		RATIO;

	mpq_init (RATIO);
	mpz_init (TOTAL);
	mpz_init (COUNT);

	stgp_num_extreme_points (TOTAL, n);
	stgp_subtour_ep_count (COUNT, n, k);

	mpz_set (mpq_numref (RATIO), COUNT);
	mpz_set (mpq_denref (RATIO), TOTAL);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (RATIO);
}

/*
 * Return the Centroid Distance squared for an STGP(n) subtour S of
 * cardinality k.
 */

	void
stgp_subtour_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_ptr		N, D;
mpq_t		DIST2;

	mpq_init (DIST2);

	N	= mpq_numref (DIST2);
	D	= mpq_denref (DIST2);

	mpz_set_ui (N, 2);
	mpz_mul_ui (N, N, k - 1);
	mpz_mul_ui (N, N, n - 1);
	mpz_mul_ui (N, N, n - k);

	mpz_set_ui (D, k);
	mpz_mul_ui (D, D, n);
	mpz_mul_ui (D, D, n + k - 1);

	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpq_clear (DIST2);
}
