/***********************************************************************

	File:	sthgp_indicators.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute strength indicators for polytope STHGP(n) via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "sthgp_indicators.h"

#include "binomial.h"
#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "new.h"
#include "pmoment.h"
#include <stdlib.h>
#include "stirling2.h"
#include <string.h>

/*
 * Local Types
 */

/*
 * An object to represent a variable-sized row of numbers.
 */

struct ZRow {
	int		size;		/* Size of row array */
	mpz_ptr *	array;		/* Array *pointers* to numbers */
};

/*
 * An object to cache a variable-length sequence of Row objects.
 */

struct ZRCache {
	int		nrows;		/* Number of valid rows */
	int		size;		/* Size of array */
	struct ZRow **	array;		/* Array of pointers to Row objects */
};

/*
 * Local Functions
 */

static void		Alpha (mpz_ptr res, int n);
static void		b (mpz_ptr res, int n);
static void		Beta (mpz_ptr res, int n, int k);
static void		clear_ZRCache (struct ZRCache * cp);
static void		complex_pmoment (mpz_ptr res, int lambda, int m, int n);
static void		d (mpz_ptr res, int n);
static void		E (mpz_ptr result, int n, int k);
static void		Gamma (mpz_ptr res, int n, int k);
static void		g (mpz_ptr res, int n, int k);
static struct ZRow *	get_ZRCache_row (struct ZRCache * cp, int n);
static mpz_ptr		get_ZRow_column (struct ZRow * rp, int n);
static void		set_ZRow_column (struct ZRow *	cp,
					 int		n,
					 mpz_ptr	value);
static void		t (mpz_ptr res, int n, int k);
static void		V (mpz_ptr result, int n, int k);

/*
 * Local Data
 */

static struct ZRCache	e_cache;

/*
 * Return the number of extreme points of STHGP(n).
 */

	void
sthgp_num_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices in hypergraph */
)
{
unsigned long int	rem;
mpz_t			COUNT;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n <= 0);

	mpz_init (COUNT);

	pmoment (COUNT, n, n - 1);
	rem = mpz_tdiv_q_ui (COUNT, COUNT, n);

	FATAL_ERROR_IF (rem NE 0);

	mpz_set (res, COUNT);

	mpz_clear (COUNT);
}

/*
 * Let e \in E such that |e| = k.  Return the number of STHGP(n)
 * extreme points that contain edge e.
 */

	void
sthgp_num_edge_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of edge */
)
{
	g (res, n, k);
}

/*
 * Return the number of STHGP(n) extreme points incident to the
 * non-negativity inequality for an edge e of cardinality k.
 */

	void
sthgp_nonneg_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of edge */
)
{
mpz_t		TOTAL, COUNT;

	mpz_init (TOTAL);
	mpz_init (COUNT);

	sthgp_num_extreme_points (TOTAL, n);
	g (COUNT, n, k);

	/* Return number of spanning trees that DO NOT contain edge e. */
	mpz_sub (res, TOTAL, COUNT);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
}

/*
 * Return the Extreme Point Ratio for an STHGP(n) non-negativity inequality
 * for an edge e of cardinality k.
 */

	void
sthgp_nonneg_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of edge */
)
{
mpz_t		TOTAL, COUNT;
mpq_t		RATIO;

	mpq_init (RATIO);
	mpz_init (TOTAL);
	mpz_init (COUNT);

	sthgp_num_extreme_points (TOTAL, n);
	g (COUNT, n, k);
	mpz_sub (mpq_numref (RATIO), TOTAL, COUNT);
	mpz_set (mpq_denref (RATIO), TOTAL);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (RATIO);
}

/*
 * Return the Centroid Distance squared for an STHGP(n) non-negativity
 * inequality for an edge e of cardinality k.
 */

	void
sthgp_nonneg_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of edge */
)
{
mpz_t		TOTAL, COUNT, A;
mpq_t		DIST2;

	mpq_init (DIST2);
	mpz_init (TOTAL);
	mpz_init (COUNT);
	mpz_init (A);

	sthgp_num_extreme_points (TOTAL, n);
	g (COUNT, n, k);
	mpz_set (mpq_numref (DIST2), COUNT);
	mpz_set (mpq_denref (DIST2), TOTAL);
	mpq_canonicalize (DIST2);

	mpq_mul (DIST2, DIST2, DIST2);

	Alpha (A, n);

	mpz_mul (mpq_numref (DIST2), mpq_numref (DIST2), A);
	mpz_sub_ui (A, A, (k - 1) * (k - 1));
	mpz_mul (mpq_denref (DIST2), mpq_denref (DIST2), A);

	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpz_clear (A);
	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (DIST2);
}

/*
 * Return the number of STHGP(n) extreme points incident to a
 * subtour inequality S of cardinality k.
 */

	void
sthgp_subtour_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of subtour */
)
{
int			i, j, p;
unsigned long int	rem;
mpz_t			SUM1, TERM1, SUM2, TERM2, SUM3, TERM3, TEMP;

	mpz_init (SUM1);
	mpz_init (TERM1);
	mpz_init (SUM2);
	mpz_init (TERM2);
	mpz_init (SUM3);
	mpz_init (TERM3);
	mpz_init (TEMP);

	for (i = 0; i < k; i++) {
		mpz_set_ui (SUM2, 0);
		for (j = 1; j <= n - k; j++) {
			mpz_set_ui (SUM3, 0);
			for (p = 0; p <= j; p++) {
				pmoment (TERM3, k, p);
				mpz_ui_pow_ui (TEMP, i, j - p);
				mpz_mul (TERM3, TERM3, TEMP);
				binomial (TEMP, j, p);
				mpz_mul (TERM3, TERM3, TEMP);

				mpz_add (SUM3, SUM3, TERM3);
			}
			pmoment (TERM2, n - k, n - k - j);
			mpz_mul (TERM2, TERM2, SUM3);
			binomial (TEMP, n - k, j);
			mpz_mul (TERM2, TERM2, TEMP);
			mpz_mul_ui (TERM2, TERM2, j);

			mpz_add (SUM2, SUM2, TERM2);
		}
		stirling2 (TERM1, k - 1, i);
		mpz_ui_pow_ui (TEMP, k, i);
		mpz_mul (TERM1, TERM1, TEMP);
		mpz_mul (TERM1, TERM1, SUM2);

		mpz_add (SUM1, SUM1, TERM1);
	}

	rem = mpz_tdiv_q_ui (SUM1, SUM1, k);
	FATAL_ERROR_IF (rem NE 0);

	rem = mpz_tdiv_q_ui (SUM1, SUM1, n - k);
	FATAL_ERROR_IF (rem NE 0);

	mpz_set (res, SUM1);

	mpz_clear (TEMP);
	mpz_clear (TERM3);
	mpz_clear (SUM3);
	mpz_clear (TERM2);
	mpz_clear (SUM2);
	mpz_clear (TERM1);
	mpz_clear (SUM1);
}

/*
 * Return the number of STHGP(n) extreme points incident to a
 * subtour inequality S of cardinality k.  This uses an "alternate"
 * formula that is much more computationally efficient for large n.
 * This formula is not "closed-form," however, because it contains
 * non-standard functions defined only via non-standard recurrence
 * relations.
 */

	void
sthgp_subtour_alternate_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of subtour */
)
{
int		i, m;
mpz_t		SUM1, SUM2, TERM1, TERM2, KPOW, ST, BN, VTREES, ETREES;

	mpz_init (SUM1);
	mpz_init (SUM2);
	mpz_init (TERM1);
	mpz_init (TERM2);
	mpz_init (KPOW);
	mpz_init (ST);
	mpz_init (BN);
	mpz_init (VTREES);
	mpz_init (ETREES);

	mpz_set_ui (KPOW, 1);

	for (i = 1; i < k; i++) {
		stirling2 (ST, k - 1, i);
		mpz_mul (TERM1, ST, KPOW);

		mpz_set_ui (SUM2, 0);
		for (m = 0; m <= n - k; m++) {
			binomial (BN, n-k, m);
			V (VTREES, n-m-k, k);
			E (ETREES, m, i);
			mpz_mul (TERM2, BN, VTREES);
			mpz_mul (TERM2, TERM2, ETREES);
			mpz_add (SUM2, SUM2, TERM2);
		}

		mpz_mul (TERM1, TERM1, SUM2);
		mpz_add (SUM1, SUM1, TERM1);

		mpz_mul_ui (KPOW, KPOW, k);
	}

	mpz_set (res, SUM1);

	mpz_clear (ETREES);
	mpz_clear (VTREES);
	mpz_clear (BN);
	mpz_clear (ST);
	mpz_clear (KPOW);
	mpz_clear (TERM2);
	mpz_clear (TERM1);
	mpz_clear (SUM2);
	mpz_clear (SUM1);
}

/*
 * Compute V(n,k): the number of forests consisting of K subtrees that
 * connect N "loose" vertices to K existing root vertices (that are *NOT*
 * part of the N "loose" vertices).
 *
 * These are "Vertex-attached" subtrees.
 */

	static
	void
V (

mpz_ptr		result,		/* IN/OUT: resulting number of forests */
int		n,		/* IN: Number of "loose" vertices */
int		k		/* IN: Number of roots for subtrees */
)
{
	FATAL_ERROR_IF ((result EQ NULL) OR (n < 0) OR (k < 0));

	g (result, n + k, k);
}

/*
 * Compute the number of "Edge-attached" forests that partition
 * N "loose" vertices into K disjoint subtrees that can be attached to K
 * existing *hyperedges* (that lie outside of the N "loose" vertices).
 */

	static
	void
E (

mpz_ptr		result,		/* IN/OUT: resulting number of forests */
int		n,		/* IN: Number of "loose" vertices */
int		k		/* IN: Number of root edges */
)
{
int			i;
struct ZRow *		rp;
struct ZRCache *	cp;
mpz_ptr			zp;
mpz_t			TMP1, TMP2;

	FATAL_ERROR_IF ((result EQ NULL) OR (n < 0) OR (k < 0));

	if (n EQ 0) {
		mpz_set_ui (result, 1);
		return;
	}
	if (k EQ 0) {
		mpz_set_ui (result, 0);
		return;
	}

	cp = &e_cache;
	rp = get_ZRCache_row (cp, k);

	FATAL_ERROR_IF (rp EQ NULL);

	zp = get_ZRow_column (rp, n);

	if (zp NE NULL) {
		mpz_set (result, zp);
		return;
	}

	/* Compute E(n,k) recursively. */
	zp = new_Z_vector (1);
	if (k EQ 1) {
		pmoment (zp, n, n);
		mpz_tdiv_q_ui (zp, zp, n);
	}
	else {
		mpz_init (TMP1);
		mpz_init (TMP2);
		for (i = 0; i <= n; i++) {
			binomial (TMP1, n, i);
			E (TMP2, i, k - 1);
			mpz_mul (TMP1, TMP1, TMP2);
			E (TMP2, n - i, 1);
			mpz_mul (TMP1, TMP1, TMP2);
			mpz_add (zp, zp, TMP1);
		}
		mpz_clear (TMP2);
		mpz_clear (TMP1);
	}
	set_ZRow_column (rp, n, zp);

	mpz_set (result, zp);
}

/*
 * Return the Extreme Point Ratio for an STHGP(n) subtour S of cardinality k.
 */

	void
sthgp_subtour_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_t		TOTAL, COUNT;
mpq_t		RATIO;

	mpq_init (RATIO);
	mpz_init (TOTAL);
	mpz_init (COUNT);

	/* Use the more efficient formula. */
	sthgp_subtour_alternate_ep_count (COUNT, n, k);

	sthgp_num_extreme_points (TOTAL, n);

	mpz_set (mpq_numref (RATIO), COUNT);
	mpz_set (mpq_denref (RATIO), TOTAL);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (RATIO);
}

/*
 * Set array[k] to the number of STHGP(n) extreme points incident
 * to a subtour inequality S of cardinality k (for all 2 <= k <= n).
 * This means that array[n] is the total number of extreme points.
 */

	void
sthgp_all_subtour_ep_count (

mpz_ptr		array,		/* IN/OUT: Array of extreme point counts */
int		n		/* IN: Number of vertices in hypergraph */
)
{
int		k;

	FATAL_ERROR_IF ((array EQ NULL) OR (n < 3));

	mpz_set_ui (&array [0], 0);
	mpz_set_ui (&array [1], 0);
	for (k = 2; k < n; k++) {
		sthgp_subtour_alternate_ep_count (&array [k], n, k);
	}
	/* Subtour S = V corresponds to the affine hull hyperplane, so	*/
	/* it contains all all extreme points.				*/
	sthgp_num_extreme_points (&array [n], n);
}

/*
 * Return the Centroid Distance squared for an STHGP(n) subtour S of
 * cardinality k.
 */

	void
sthgp_subtour_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_t		A, B, G, T;
mpq_t		DIST2;

	mpq_init (DIST2);
	mpz_init (A);
	mpz_init (B);
	mpz_init (G);
	mpz_init (T);

	t (T, n, k);
	Alpha (A, n);
	Beta (B, n, k);
	Gamma (G, n, k);

	mpz_mul_ui (mpq_numref (DIST2), T, n - k);
	pmoment (mpq_denref (DIST2), n, n - 1);
	mpq_canonicalize (DIST2);

	mpq_mul (DIST2, DIST2, DIST2);

	mpz_mul (mpq_numref (DIST2), mpq_numref (DIST2), A);
	mpz_mul (A, A, G);
	mpz_mul (B, B, B);
	mpz_sub (A, A, B);
	mpz_mul (mpq_denref (DIST2), mpq_denref (DIST2), A);
	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpz_clear (T);
	mpz_clear (G);
	mpz_clear (B);
	mpz_clear (A);
	mpq_clear (DIST2);
}

/*
 * Set array[k] to the Centroid Distance squared for an STHGP(n) subtour
 * S of cardinality k (for all 2 <= k <= n).
 */

	void
sthgp_all_subtour_cd2 (

mpq_ptr		array,		/* IN/OUT: Array of Centroid Dist squared */
int		n		/* IN: Number of vertices in hypergraph */
)
{
int		k;

	mpq_set_ui (&array [0], 0, 1);
	mpq_set_ui (&array [1], 0, 1);

	for (k = 2; k < n; k++) {
		sthgp_subtour_cd2 (&array [k], n, k);
	}
	mpq_set_ui (&array [n], 0, 1);
}

/*
 * Return the number of STHGP(n) extreme points containing a
 * particular edge e of cardinality k.
 */

	static
	void
g (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices in hypergraph */
int		k		/* IN: Cardinality of edge */
)
{
unsigned long int	rem;
mpz_t			COUNT;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 2);
	FATAL_ERROR_IF (k < 2);
	FATAL_ERROR_IF (k > n);

	mpz_init (COUNT);

	pmoment (COUNT, n, n - k);

	mpz_mul_ui (COUNT, COUNT, k);

	rem = mpz_tdiv_q_ui (COUNT, COUNT, n);

	FATAL_ERROR_IF (rem NE 0);

	mpz_set (res, COUNT);

	mpz_clear (COUNT);
}

/*
 * Return Alpha(n) = (n^2 - 3*n + 4)*2^(n-2) - 1.
 */

	static
	void
Alpha (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Parameter n */
)
{
mpz_t		VAL;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);

	if (n < 2) {
		mpz_set_ui (res, 0);
		return;
	}

	mpz_init (VAL);

	mpz_set_ui (VAL, n);
	mpz_sub_ui (VAL, VAL, 3);
	mpz_mul_ui (VAL, VAL, n);
	mpz_add_ui (VAL, VAL, 4);
	mpz_mul_2exp (VAL, VAL, n - 2);
	mpz_sub_ui (VAL, VAL, 1);

	mpz_set (res, VAL);

	mpz_clear (VAL);
}

/*
 * Return Beta(n,k) = Gamma(n,k) + b(n-k) * d(k).
 */

	static
	void
Beta (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Parameter n */
int		k		/* IN: Parameter k */
)
{
mpz_t		TEMP1, TEMP2, TEMP3;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);
	FATAL_ERROR_IF (k > n);

	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpz_init (TEMP3);

	Gamma (TEMP1, n, k);
	b (TEMP2, n - k);
	d (TEMP3, k);

	mpz_mul (TEMP2, TEMP2, TEMP3);

	mpz_add (res, TEMP1, TEMP2);

	mpz_clear (TEMP3);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Return Gamma(n,k) = 2^(n-k) * Alpha(k).
 */

	static
	void
Gamma (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Parameter n */
int		k		/* IN: Parameter k */
)
{
mpz_t		TEMP;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);
	FATAL_ERROR_IF (k > n);

	mpz_init (TEMP);

	Alpha (TEMP, k);
	mpz_mul_2exp (res, TEMP, n - k);

	mpz_clear (TEMP);
}

/*
 * Return b(n) = n*2^(n-1).
 */

	static
	void
b (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Parameter n */
)
{
mpz_t		VAL;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);

	if (n < 1) {
		mpz_set_ui (res, n);
		return;
	}

	mpz_init (VAL);

	mpz_set_ui (VAL, n);
	mpz_mul_2exp (res, VAL, n - 1);

	mpz_clear (VAL);
}

/*
 * Return d(n) = 1 + n*2^(n-1) - 2^n.
 */

	static
	void
d (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Parameter n */
)
{
mpz_t		TEMP1, TEMP2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);

	if (n < 1) {
		mpz_set_ui (res, 0);
		return;
	}

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	mpz_set_ui (TEMP1, n);
	mpz_mul_2exp (TEMP1, TEMP1, n - 1);

	mpz_set_ui (TEMP2, 1);
	mpz_mul_2exp (TEMP2, TEMP2, n);

	mpz_sub (TEMP1, TEMP1, TEMP2);

	mpz_add_ui (res, TEMP1, 1);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * External interface to compute t(n,k).
 */

	void
sthgp_subtour_cd2_t (

mpz_ptr		res,
int		n,
int		k
)
{
	t (res, n, k);
}

/*
 * External interface to compute Alpha(n).
 */

	void
sthgp_subtour_cd2_Alpha (

mpz_ptr		res,
int		n
)
{
	Alpha (res, n);
}

/*
 * External interface to compute Beta(n,k).
 */

	void
sthgp_subtour_cd2_Beta (

mpz_ptr		res,
int		n,
int		k
)
{
	Beta (res, n, k);
}

/*
 * External interface to compute Gamma(n,k).
 */

	void
sthgp_subtour_cd2_Gamma (

mpz_ptr		res,
int		n,
int		k
)
{
	Gamma (res, n, k);
}

/*
 * Return t(n,k) = E[Xn*(Xn + 1)^(n-2)] - E[Xn^k*(Xn + 1)^(n-k-1)].
 */

	static
	void
t (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Parameter n */
int		k		/* IN: Parameter k */
)
{
mpz_t		TEMP1, TEMP2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);
	FATAL_ERROR_IF (k > n);

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	complex_pmoment (TEMP1, n, 1, n - 2);
	complex_pmoment (TEMP2, n, k, n - k - 1);

	mpz_sub (res, TEMP1, TEMP2);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Return t(n,k) = E[X_{\lambda}^m*(X_{\lambda} + 1)^n]
 *		 = \sum_{j=0}^n {n choose j} E[X_{\lambda}^{m+j}].
 */

	static
	void
complex_pmoment (

mpz_ptr		res,		/* IN/OUT: Result */
int		lambda,		/* IN: Mean of Poisson r.v. X_{\lambda} */
int		m,		/* IN: Exponent m */
int		n		/* IN: Exponent n */
)
{
int		j;
mpz_t		TEMP1, TEMP2, SUM;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (lambda < 0);
	FATAL_ERROR_IF (m < 0);
	FATAL_ERROR_IF (n < 0);

	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpz_init (SUM);

	/* Compute E[X_{\lambda}^m*(X_{\lambda} + 1)^n			*/
	/* = \sum_{j=0}^n {n \choose j} pmoment (\lambda, m+j).		*/

	for (j = 0; j <= n; j++) {
		binomial (TEMP1, n, j);
		pmoment (TEMP2, lambda, m + j);
		mpz_mul (TEMP1, TEMP1, TEMP2);
		mpz_add (SUM, SUM, TEMP1);
	}

	mpz_set (res, SUM);

	mpz_clear (SUM);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Shut down the sthgp_indicators module, freeing all memory it
 * statically holds.
 */

	void
sthgp_indicators_shutdown (void)

{
	clear_ZRCache (&e_cache);
	pmoment_shutdown ();
	stirling2_shutdown ();
	binomial_shutdown ();
}

/*
 * Return the N-th row from the given ZRCache object.  If no such
 * row exists yet, it creates an empty ZRow object and adds it to
 * the cache before returning it.
 */

	static
	struct ZRow *
get_ZRCache_row (

struct ZRCache *	cp,		/* IN: ZRCache to access */
int			n		/* IN: row number to return */
)
{
int		old_size, new_size;
struct ZRow **	old_array;
struct ZRow **	new_array;
struct ZRow **	hookp;
struct ZRow *	rp;

	FATAL_ERROR_IF ((cp EQ NULL) OR (n < 0));

	old_size	= cp -> size;
	old_array	= cp -> array;
	if (n >= old_size) {
		/* Need to expand the array. */
		new_size = 2 * old_size;
		if (n >= new_size) {
			new_size = n + 1;
		}
		if (n >= cp -> nrows) {
			cp -> nrows = n + 1;
		}
		new_array = NEWA (new_size, struct ZRow *);
		memset (new_array, 0, new_size * sizeof (new_array [0]));
		memcpy (new_array, old_array, old_size * sizeof (new_array [0]));
		cp -> array	= new_array;
		cp -> size	= new_size;
		if (old_array NE NULL) {
			free (old_array);
		}
		old_array = new_array;
	}

	hookp = &old_array [n];

	rp = *hookp;
	if (rp EQ NULL) {
		rp = NEW (struct ZRow);
		rp -> size	= 0;
		rp -> array	= NULL;
		*hookp = rp;
	}
	return (rp);
}

/*
 * Return the n-th column of the given ZRow object.
 * This returns NULL if no such element exists yet.
 */

	static
	mpz_ptr
get_ZRow_column (

struct ZRow *		rp,		/* IN: ZRow object to access */
int			n		/* IN: column number to access */
)
{
mpz_ptr		ptr;

	FATAL_ERROR_IF ((rp EQ NULL) OR (n < 0));

	ptr = NULL;
	if (n < rp -> size) {
		ptr = rp -> array [n];
	}
	return (ptr);
}

/*
 * Set the n-th column of the given ZRow object.
 * This returns NULL if no such element exists yet.
 */

	static
	void
set_ZRow_column (

struct ZRow *		rp,		/* IN: ZRow object to access */
int			n,		/* IN: column number to set */
mpz_ptr			value		/* IN: value to set */
)
{
int		old_size, new_size;
mpz_ptr *	old_array;
mpz_ptr *	new_array;

	FATAL_ERROR_IF ((rp EQ NULL) OR (n < 0) OR (value EQ NULL));

	old_size	= rp -> size;
	old_array	= rp -> array;
	if (n >= old_size) {
		/* Need to expand this row. */
		new_size = 2 * n + 1;
		new_array = NEWA (new_size, mpz_ptr);
		memset (new_array, 0, new_size * sizeof (new_array [0]));
		memcpy (new_array, old_array, old_size * sizeof (new_array [0]));
		rp -> array	= new_array;
		rp -> size	= new_size;
		if (old_array NE NULL) {
			free (old_array);
		}
		old_array = new_array;
	}

	/* Column N was previously set! */
	FATAL_ERROR_IF (rp -> array [n] NE NULL);

	rp -> array [n] = value;
}

/*
 * Clear out the given ZRCache, freeing all memory it holds.
 */

	static
	void
clear_ZRCache (

struct ZRCache *	cp	/* IN/OUT: ZRCache to clear */
)
{
int		i, j, nrows, size, rsize;
struct ZRow **	array;
struct ZRow *	rp;
mpz_ptr *	row;

	FATAL_ERROR_IF (cp EQ NULL);

	nrows	= cp -> nrows;
	size	= cp -> size;
	array	= cp -> array;

	FATAL_ERROR_IF (nrows < 0);
	FATAL_ERROR_IF (size < 0);
	FATAL_ERROR_IF ((nrows > 0) AND (nrows > size));

	if (nrows <= 0) {
		FATAL_ERROR_IF (array NE NULL);
	}
	else {
		FATAL_ERROR_IF (array EQ NULL);
		for (i = 0; i < nrows; i++) {
			rp = array [i];
			if (rp EQ NULL) continue;
			array [i] = NULL;
			rsize	= rp -> size;
			row	= rp -> array;
			for (j = 0; j < rsize; j++) {
				if (row [j] NE NULL) {
					free_Z_vector (row [j], 1);
					row [j] = NULL;
				}
			}
			free (row);
			memset (rp, 0, sizeof (*rp));
			free (rp);
		}
		free (array);
	}
	memset (cp, 0, sizeof (*cp));
}
