/***********************************************************************

	File:	sthgp_sec_angles.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute angle between a pair of STHGP(n) subtours via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "sthgp_sec_angles.h"

#include "fatal.h"
#include <gmp.h>
#include "logic.h"
#include <math.h>
#include <mpfr.h>
#include <stdio.h>
#include <stdlib.h>


/*
 * Local Routines
 */

static void		Alpha (mpz_ptr result, int n);
static void		B (mpz_ptr result, int n);
static void		Beta (mpz_ptr result, int n, int k);
static void		C (mpz_ptr result, int n);
static void		D (mpz_ptr result, int n);
static void		Gamma (mpz_ptr result, int n, int k);
static void		Mu (mpz_ptr result, int n, int k);
static void		W (mpz_ptr result, int n, int p, int q, int r);


/*
 * Local Data
 */

mpfr_prec_t		prec	= 128;
mpfr_rnd_t		rnd	= GMP_RNDN;

/*
 * The function Theta (n, p, q, r).
 * Return the angle (in *degrees*) between a pair of STHGP(n) subtours
 * S1 and S2:
 *
 *	n = total number of vertices
 *	p = number of vertices in S1 \ S2
 *	q = number of vertices in S2 \ S1
 *	r = number of vertices in S1 intersect S2
 */

	double
sthgp_subtour_angle_theta (

int		n,
int		p,
int		q,
int		r
)
{
double		cos_phi, phi, degrees, int_angle;

	cos_phi = sthgp_subtour_angle_f (n, p, q, r);

	phi = acos (cos_phi);

	degrees = 180.0 * phi / 3.14159265358979323;

	int_angle = 180.0 - degrees;

	return (int_angle);
}

/*
 * The function F (n, p, q, r).
 */

	double
sthgp_subtour_angle_f (

int		n,
int		p,
int		q,
int		r
)
{
double		result;
mpz_t		NUM, DEN, TEMP1, TEMP2;
mpfr_t		FNUM, FDEN;

	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF (p < 0);
	FATAL_ERROR_IF (q < 0);
	FATAL_ERROR_IF (r < 0);
	FATAL_ERROR_IF (p + r < 2);
	FATAL_ERROR_IF (q + r < 2);
	FATAL_ERROR_IF (p + q + r > n);

	mpz_init (NUM);
	mpz_init (DEN);
	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpfr_init2 (FNUM, prec);
	mpfr_init2 (FDEN, prec);

	Alpha (TEMP1, n);
	W (TEMP2, n, p, q, r);
	mpz_mul (NUM, TEMP1, TEMP2);

	Beta (TEMP1, n, p + r);
	Beta (TEMP2, n, q + r);
	mpz_mul (TEMP1, TEMP1, TEMP2);
	mpz_sub (NUM, NUM, TEMP1);

	Mu (TEMP1, n, p + r);
	Mu (TEMP2, n, q + r);
	mpz_mul (DEN, TEMP1, TEMP2);

	mpfr_set_z (FNUM, NUM, rnd);
	mpfr_set_z (FDEN, DEN, rnd);
	mpfr_sqrt (FDEN, FDEN, rnd);

	mpfr_div (FNUM, FNUM, FDEN, rnd);

	result = mpfr_get_d (FNUM, rnd);

	mpfr_clear (FDEN);
	mpfr_clear (FNUM);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
	mpz_clear (DEN);
	mpz_clear (NUM);

	return (result);
}

/*
 * The function W (n, p, q, r).
 */

	static
	void
W (

mpz_ptr		result,		/* IN/OUT: Result */
int		n,		/* IN: The value n */
int		p,		/* IN: The value p */
int		q,		/* IN: The value q */
int		r		/* IN: The value r */
)
{
mpz_t		TEMP1, TEMP2, TEMP3;

	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpz_init (TEMP3);

	Alpha (TEMP1, r);
	mpz_mul_2exp (TEMP1, TEMP1, p + q);

	D (TEMP2, p);
	D (TEMP3, q);
	mpz_mul (TEMP2, TEMP2, TEMP3);
	mpz_add (TEMP1, TEMP1, TEMP2);

	B (TEMP2, p);
	B (TEMP3, q);
	mpz_mul (TEMP2, TEMP2, TEMP3);
	C (TEMP3, r);
	mpz_mul (TEMP2, TEMP2, TEMP3);
	mpz_add (TEMP1, TEMP1, TEMP2);

	B (TEMP2, p + q);
	D (TEMP3, r);
	mpz_mul (TEMP2, TEMP2, TEMP3);
	mpz_add (TEMP1, TEMP1, TEMP2);

	mpz_mul_2exp (result, TEMP1, n - p - q - r);

	mpz_clear (TEMP3);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * The function Mu (n, k).
 */

	static
	void
Mu (

mpz_ptr		result,		/* IN/OUT: Result */
int		n,		/* IN: The value n */
int		k		/* IN: The value k */
)
{
mpz_t		TEMP1, TEMP2;

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	Alpha (TEMP1, n);
	Gamma (TEMP2, n, k);
	mpz_mul (TEMP1, TEMP1, TEMP2);

	Beta (TEMP2, n, k);
	mpz_mul (TEMP2, TEMP2, TEMP2);

	mpz_sub (result, TEMP1, TEMP2);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * The function Alpha (n).
 */

	static
	void
Alpha (

mpz_ptr		result,		/* IN/OUT: Result */
int		n		/* IN: The value n */
)
{
mpz_t		TEMP;

	if (n < 2) {
		FATAL_ERROR_IF (n < 0);
		mpz_set_ui (result, 0);
		return;
	}

	mpz_init (TEMP);

	mpz_set_ui (TEMP, n);
	mpz_sub_ui (TEMP, TEMP, 3);
	mpz_mul_si (TEMP, TEMP, n);
	mpz_add_ui (TEMP, TEMP, 4);
	mpz_mul_2exp (TEMP, TEMP, n - 2);
	mpz_sub_ui (result, TEMP, 1);

	mpz_clear (TEMP);
}

/*
 * The function Beta (n, k).
 */

	static
	void
Beta (

mpz_ptr		result,		/* IN/OUT: Result */
int		n,		/* IN: The value n */
int		k		/* IN: The value k */
)
{
mpz_t		TEMP1, TEMP2, TEMP3;

	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpz_init (TEMP3);

	Gamma (TEMP1, n, k);

	B (TEMP2, n - k);
	D (TEMP3, k);
	mpz_mul (TEMP2, TEMP2, TEMP3);

	mpz_add (result, TEMP1, TEMP2);

	mpz_clear (TEMP3);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * The function Gamma (n, k).
 */

	static
	void
Gamma (

mpz_ptr		result,		/* IN/OUT: Result */
int		n,		/* IN: The value n */
int		k		/* IN: The value k */
)
{
mpz_t		TEMP;

	mpz_init (TEMP);

	Alpha (TEMP, k);
	mpz_mul_2exp (result, TEMP, n - k);

	mpz_clear (TEMP);
}

/*
 * The function B (n).
 */

	static
	void
B (

mpz_ptr		result,		/* IN/OUT: Result */
int		n		/* IN: The value n */
)
{
mpz_t		TEMP;

	if (n < 1) {
		FATAL_ERROR_IF (n < 0);
		mpz_set_ui (result, 0);
		return;
	}

	mpz_init (TEMP);

	mpz_set_ui (TEMP, n);
	mpz_mul_2exp (result, TEMP, n - 1);

	mpz_clear (TEMP);
}

/*
 * The function C (n).
 */

	static
	void
C (

mpz_ptr		result,		/* IN/OUT: Result */
int		n		/* IN: The value n */
)
{
mpz_t		TEMP;

	FATAL_ERROR_IF (n < 0);

	mpz_init (TEMP);

	mpz_set_ui (TEMP, 1);
	mpz_mul_2exp (TEMP, TEMP, n);
	mpz_sub_ui (result, TEMP, 1);

	mpz_clear (TEMP);
}

/*
 * The function D (n).
 */

	static
	void
D (

mpz_ptr		result,		/* IN: Result */
int		n		/* IN: The value n */
)
{
mpz_t		TEMP1, TEMP2;

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	B (TEMP1, n);
	C (TEMP2, n);
	mpz_sub (result, TEMP1, TEMP2);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}
