/***********************************************************************

	File:	sthgp_sec_cd2_dx.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute various functions associated with delta-x values
	for STHGP(n) subtour Centroid Distance squared.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "sthgp_sec_cd2_dx.h"

#include "fatal.h"
#include <gmp.h>
#include "logic.h"
#include "pmoment.h"
#include "sthgp_indicators.h"

/*
 * Local Functions
 */

static void		mul_pow_2 (mpz_ptr res, mpz_srcptr op, int pow);

/*
 * Compute \sum_{e \in E(S)} (DeltaX_e)^2.
 */

	void
sthgp_subtour_cd2_dx_Delta_x_1 (

mpq_ptr		value,
int		n,
int		k
)
{
mpz_t		A, B, A2, B2, AB, TMP1, TMP2, TMP3, TMP4, TMP5;
mpq_t		VAL;

	mpq_init (VAL);

	mpz_init (A);
	mpz_init (B);
	mpz_init (A2);
	mpz_init (B2);
	mpz_init (AB);
	mpz_init (TMP1);
	mpz_init (TMP2);
	mpz_init (TMP3);
	mpz_init (TMP4);
	mpz_init (TMP5);

	sthgp_subtour_cd2_dx_TAU (VAL, n, k);
	mpq_mul (VAL, VAL, VAL);

	sthgp_subtour_cd2_Alpha (A, n);
	sthgp_subtour_cd2_Beta  (B, n, k);
	mpz_mul (A2, A, A);
	mpz_mul (B2, B, B);
	mpz_mul (AB, A, B);

	mpz_set_si (TMP1, k * k - 3*k + 4);
	mul_pow_2  (TMP1, TMP1, n - 2);
	mpz_set_si (TMP2, 1);
	mul_pow_2  (TMP2, TMP2, n - k);
	mpz_sub (TMP1, TMP1, TMP2);
	mpz_mul (TMP1, TMP1, A2);

	mpz_set_si (TMP2, (- k + 2)*n + k - 4);
	mul_pow_2  (TMP2, TMP2, n - 1);
	mpz_set_si (TMP3, n - k - 2);
	mul_pow_2  (TMP3, TMP3, n - k);
	mpz_sub (TMP2, TMP2, TMP3);
	mpz_mul (TMP2, TMP2, AB);

	mpz_set_si (TMP3, n*n - 3*n + 4);
	mul_pow_2  (TMP3, TMP3, n - 2);
	mpz_set_si (TMP4, k*k*k - 2*n*k*k + (n*n - n + 3)*k + n*n - 3*n + 4);
	mul_pow_2  (TMP4, TMP4, n - k - 2);
	mpz_sub (TMP3, TMP3, TMP4);
	mpz_mul (TMP3, TMP3, B2);

	mpz_add (TMP1, TMP1, TMP2);
	mpz_add (TMP1, TMP1, TMP3);

	mpz_mul (mpq_numref (VAL), mpq_numref (VAL), TMP1);
	mpq_canonicalize (VAL);
	mpq_set (value, VAL);

	mpz_clear (TMP5);
	mpz_clear (TMP4);
	mpz_clear (TMP3);
	mpz_clear (TMP2);
	mpz_clear (TMP1);
	mpz_clear (AB);
	mpz_clear (B2);
	mpz_clear (A2);
	mpz_clear (B);
	mpz_clear (A);

	mpq_clear (VAL);
}

/*
 * Compute \sum_{e \in (E - E(S))} (DeltaX_e)^2.
 */

	void
sthgp_subtour_cd2_dx_Delta_x_2 (

mpq_ptr		value,
int		n,
int		k
)
{
mpz_t		A, B, B2, TMP1, TMP2, TMP3;
mpq_t		VAL;

	mpq_init (VAL);

	mpz_init (A);
	mpz_init (B);
	mpz_init (B2);
	mpz_init (TMP1);
	mpz_init (TMP2);
	mpz_init (TMP3);

	sthgp_subtour_cd2_dx_TAU (VAL, n, k);
	mpq_mul (VAL, VAL, VAL);

	sthgp_subtour_cd2_Alpha (A, n - k);
	sthgp_subtour_cd2_Beta  (B, n, k);
	mpz_mul (B2, B, B);

	mpz_set_si (TMP1, k);
	mpz_mul_si (TMP1, TMP1, n-k);
	mpz_mul_si (TMP1, TMP1, n-k+1);
	mul_pow_2  (TMP1, TMP1, n-k-2);
	mpz_add (TMP1, TMP1, A);
	mpz_mul (TMP1, TMP1, B2);

	mpz_mul (mpq_numref (VAL), mpq_numref (VAL), TMP1);
	mpq_canonicalize (VAL);
	mpq_set (value, VAL);

	mpz_clear (TMP3);
	mpz_clear (TMP2);
	mpz_clear (TMP1);
	mpz_clear (B2);
	mpz_clear (B);
	mpz_clear (A);

	mpq_clear (VAL);
}

/*
 * Compute TAU(n,k) = (n-k)*t(n,k) /
 *		     ((Alpha(n)*Gamma(n,k) - Beta(n,k)^2) * E[X_n^{n-1}]).
 */

	void
sthgp_subtour_cd2_dx_TAU (

mpq_ptr		value,
int		n,
int		k
)
{
mpz_t		NUM, DEN, TMP1, TMP2, TMP3;
mpq_t		FRAC;

	mpq_init (FRAC);
	mpz_init (NUM);
	mpz_init (DEN);
	mpz_init (TMP1);
	mpz_init (TMP2);
	mpz_init (TMP3);

	sthgp_subtour_cd2_t (NUM, n, k);
	mpz_mul_ui (NUM, NUM, n - k);

	sthgp_subtour_cd2_Alpha (TMP1, n);
	sthgp_subtour_cd2_Beta  (TMP2, n, k);
	sthgp_subtour_cd2_Gamma (TMP3, n, k);

	mpz_mul (TMP1, TMP1, TMP3);
	mpz_mul (TMP2, TMP2, TMP2);
	mpz_sub (DEN, TMP1, TMP2);
	pmoment (TMP1, n, n - 1);
	mpz_mul (DEN, DEN, TMP1);

	mpz_set (mpq_numref (FRAC), NUM);
	mpz_set (mpq_denref (FRAC), DEN);
	mpq_canonicalize (FRAC);
	mpq_set (value, FRAC);

	mpz_clear (TMP3);
	mpz_clear (TMP2);
	mpz_clear (TMP1);
	mpz_clear (DEN);
	mpz_clear (NUM);
	mpq_clear (FRAC);
}

/*
 * Multiply the given integer by the given power of two.
 * It correctly handles negative powers (so long as the division is exact).
 */

	static
	void
mul_pow_2 (

mpz_ptr		res,		/* IN/OUT: Result */
mpz_srcptr	op,		/* IN: Operand to be multiplied */
int		pow		/* IN: Power of two by which to multiply */
)
{
mp_bitcnt_t	bit;

	if (pow EQ 0) {
		mpz_set (res, op);
	}
	else if (pow > 0) {
		mpz_mul_2exp (res, op, pow);
	}
	else {
		/* This is shift right. */
		pow = - pow;
		bit = 0;
		bit = mpz_scan1 (op, bit);

		FATAL_ERROR_IF (bit < pow);

		mpz_tdiv_q_2exp (res, op, pow);
	}
}
