/***********************************************************************

	File:	stnp_polytope.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	A wrapper for polytope STNGP(n,m) and STNHGP(n,m).)

************************************************************************

	Modification Log:

	a-1:	08/27/2021	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "stnp_polytope.h"

#include "bitmaskmacros.h"
#include "fatal.h"
#include <limits.h>
#include "logic.h"
#include <math.h>
#include "new.h"
#include "polytope.h"
#include "s3itypes.h"
#include "scanutils.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Global Routines
 */

struct Polytope *	make_STNGP_polytope (int nterms, int nsteins);
struct Polytope *	make_STNHGP_polytope (int nterms, int nsteins);

/*
 * The actual implementation of the STNGP(n,m) and STNHGP(n,m) Polytopes.
 * This is a "derived class" from Polytope.
 * The only difference between STNGP and STNHGP is the set of
 * edges stored in nedges/edge_list (either all edges or all hyperedges).
 */

struct STNGP {
	struct Polytope	base;		/* The base class instance */
	int		nterms;		/* Number of terminal vertices */
	int		nsteins;	/* Number of Steiner vertices */
	int		nverts;		/* Number of vertices */
	int		nedges;		/* Number of hyperedges */
	int **		edge_list;	/* To access all vertices of edge e: */
					/* ep1 = edge_list [e];		*/
					/* ep2 = edge_list [e+1];	*/
					/* while (ep1 < ep2) {		*/
					/*	v = *ep1++;		*/
					/* }				*/
};

struct gen {
	int		nverts;		/* Number of vertices */
	int		count;		/* Number of edges generated */
	int **		edge_list;	/* Array of edge vertex lists */
	int *		genp;		/* output buffer pointer */
	int *		stack;		/* base of stack */
};

/*
 * The stack-oriented data structures for enumerating hypertrees.
 */

struct sinfo {
	int			nterms;
	int			nsteins;
	int			nedges;
	int			nmasks;
	int			kmasks;
	int			vmasks;
	int *			fsets;
	int *			root_edges;
	bool *			st_chosen;
	int **			edge_list;
	bitmap_t *		solution;
	bitmap_t *		terms_left;
	bitmap_t *		compat;
	bitmap_t *		incompat;
	bitmap_t *		try;
	bitmap_t *		omit;
	bitmap_t *		cmasks;
	bitmap_t *		incmasks;
	bitmap_t *		edge_tmap;
	int8u **		one_bits_in_byte;
	int **			term_trees;
	int			ntrees;		/* Number of trees generated */
	struct iv_mask *	root;		/* Linked list of trees */
	struct iv_mask **	hookp;		/* For appending to root */
};

/*
 * Linked list of generated incidence vectors (bitmask form).
 */

struct iv_mask {
	bitmap_t *		mask;	/* Mask of edges in the tree */
	struct iv_mask *	next;	/* Next incident vector in list */
};

/*
 * Local Routines
 */

static int *		compute_binomial_row (int n);
static int8u **		create_one_bits_in_byte_table (void);
static const char **	create_var_names (int		nterms,
					  int		nsteins,
					  int		nedges,
					  int **	edge_list);
static void		enumerate_spanning_trees (struct sinfo * sip);
static struct IVecSet *	enumerate_steiner_trees (const struct Polytope * poly);
static void		free_STNHGP (struct Polytope * poly);
static void		gen_h (int, int, int *, struct gen *);
static int **		generate_edges (int nverts, int nedges);
static int **		generate_hyperedges (int nverts, int nedges);
static struct Constraint *
			get_STNGP_constraint (const struct Polytope *	_poly,
					      const char *		name);
static struct Constraint *
			get_STNGP_subtour (const struct STNGP *	poly,
					   const char *		name);
static struct Constraint *
			get_STNHGP_constraint (const struct Polytope *	_poly,
					       const char *		name);
static struct Constraint *
			get_STNHGP_nonneg_x_constraint (
					const struct STNGP *	poly,
					const char *		s);
static struct Constraint *
			get_STNHGP_subtour (const struct STNGP *	poly,
					   const char *			s);
static void		ht_recurse (struct sinfo * sip, int level, int nleft);
static void		recurse_steiners (struct sinfo *	sip,
					  int			level);

/*
 * Return an object representing the polytope STNGP(n).
 */

	struct Polytope *
make_STNGP_polytope (

int		nterms,		/* IN: number of terminal vertices */
int		nsteins		/* IN: number of Steiner vertices */
)
{
int			j, k, nverts;
int			nedges, maxlen;
struct STNGP *		poly;
int **			edge_list;
const char **		var_names;
char			buffer [32];

	FATAL_ERROR_IF (nterms < 1);
	FATAL_ERROR_IF (nsteins < 0);

	nverts = nterms + nsteins;

	if (nverts < 2) {
		fprintf (stderr, "STNGP(n) requires n >= 2.\n");
		return (NULL);
	}

	poly = NEW (struct STNGP);
	memset (poly, 0, sizeof (*poly));

	nedges	= (nverts * (nverts - 1)) >> 1;

	edge_list = generate_edges (nverts, nedges);

	var_names = create_var_names (nterms, nsteins, nedges, edge_list);

	sprintf (buffer, "STNGP(%d,%d)", nterms, nsteins);

	maxlen = 0;
	for (j = 0; j < nedges; j++) {
		k = strlen (var_names [j]);
		if (k > maxlen) {
			maxlen = k;
		}
	}

	poly -> base.name	= strdup (buffer);
	poly ->	base.dim	= nedges + nsteins;
	poly -> base.ename	= var_names;
	poly -> base.maxlen	= maxlen;
	poly -> base.get_extreme_points	= enumerate_steiner_trees;
	poly -> base.get_constraint	= get_STNGP_constraint;
	poly -> base.free		= free_STNHGP;

	poly -> nterms		= nterms;
	poly -> nsteins		= nsteins;
	poly -> nverts		= nverts;
	poly -> nedges		= nedges;
	poly -> edge_list	= edge_list;

	return (&(poly -> base));
}

/*
 * This routine generates all edges in the conventional graph.
 */

	static
	int **
generate_edges (

int			nverts,		/* IN: number of vertices */
int			nedges		/* IN: number of edges */
)
{
int		i, j, k;
int *		buf;
int *		ip;
int **		edge_list;

	edge_list = NEWA (nedges + 1, int *);
	buf	  = NEWA (2 * nedges, int);

	ip = buf;
	k = 0;
	for (i = 0; i < nverts; i++) {
		for (j = i + 1; j < nverts; j++) {
			edge_list [k] = ip;
			*ip++ = i;
			*ip++ = j;
			++k;
		}
	}
	edge_list [k] = ip;
	FATAL_ERROR_IF (k NE nedges);
	FATAL_ERROR_IF (ip NE &buf [2 * nedges]);

	return (edge_list);
}

/*
 * Return an object representing the polytope STNHGP(n).
 */

	struct Polytope *
make_STNHGP_polytope (

int		nterms,		/* IN: number of terminal vertices */
int		nsteins		/* IN: number of Steiner vertices */
)
{
int			j, k, nverts;
int			nedges, maxlen;
struct STNGP *		poly;
int **			edge_list;
const char **		var_names;
char			buffer [32];

	FATAL_ERROR_IF (nterms < 1);
	FATAL_ERROR_IF (nsteins < 0);

	nverts = nterms + nsteins;

	if (nverts < 2) {
		fprintf (stderr, "STNHGP(n) requires n >= 2.\n");
		return (NULL);
	}

	poly = NEW (struct STNGP);
	memset (poly, 0, sizeof (*poly));

	nedges	= (1 << nverts) - nverts - 1;

	edge_list = generate_hyperedges (nverts, nedges);

	var_names = create_var_names (nterms, nsteins, nedges, edge_list);

	sprintf (buffer, "STNHGP(%d,%d)", nterms, nsteins);

	maxlen = 0;
	for (j = 0; j < nedges; j++) {
		k = strlen (var_names [j]);
		if (k > maxlen) {
			maxlen = k;
		}
	}

	poly -> base.name	= strdup (buffer);
	poly ->	base.dim	= nedges + nsteins;
	poly -> base.ename	= var_names;
	poly -> base.maxlen	= maxlen;
	poly -> base.get_extreme_points	= enumerate_steiner_trees;
	poly -> base.get_constraint	= get_STNHGP_constraint;
	poly -> base.free		= free_STNHGP;

	poly -> nterms		= nterms;
	poly -> nsteins		= nsteins;
	poly -> nverts		= nverts;
	poly -> nedges		= nedges;
	poly -> edge_list	= edge_list;

	return (&(poly -> base));
}

/*
 * This routine generates all of the hyperedges of size 2 or more.
 */

	static
	int **
generate_hyperedges (

int			nverts,		/* IN: number of vertices */
int			nedges		/* IN: number of hyperedges */
)
{
int		i, k;
int *		ip;
int **		edge_list;
int *		binomial_row;
struct gen	gen;

	/* Compute sum of sizes of all hyperedges. */
	binomial_row = compute_binomial_row (nverts);
	k = 0;
	for (i = 2; i <= nverts; i++) {
		k += i * binomial_row [i];
	}

	edge_list = NEWA (nedges + 1, int *);
	ip	  = NEWA (k, int);

	gen.nverts		= nverts;
	gen.count		= 0;
	gen.edge_list		= edge_list;
	gen.genp		= ip;
	gen.stack		= NEWA (nverts, int);

	for (i = 2; i <= nverts; i++) {
		gen_h (0, i, gen.stack, &gen);
	}
	free (gen.stack);

	FATAL_ERROR_IF (ip + k NE gen.genp);
	FATAL_ERROR_IF (gen.count NE nedges);

	edge_list [nedges] = gen.genp;

	free (binomial_row);

	return (edge_list);
}

/*
 * Recursive portion of hyperedge generator.
 */

	static
	void
gen_h (

int		v,		/* IN - current vertex number */
int		nleft,		/* IN - number of vertices left to choose */
int *		sp,		/* IN - current stack pointer */
struct gen *	gp		/* IN/OUT - generation info */
)
{
int *		ip;

	if (nleft <= 0) {
		/* We have a complete hyperedge! */
		gp -> edge_list [gp -> count] = gp -> genp;
		++(gp -> count);
		for (ip = gp -> stack; ip < sp; ip++) {
			*(gp -> genp)++ = *ip;
		}
		return;
	}

	if (v + nleft <= gp -> nverts) {
		/* Include vertex v. */
		*sp = v;
		gen_h (v + 1, nleft - 1, sp + 1, gp);

		/* Exclude vertex v. */
		gen_h (v + 1, nleft, sp, gp);
	}
}

/*
 * Compute the Nth row of the binomial coefficients, via the
 * recurrence relation.
 */

	static
	int *
compute_binomial_row (

int		n	/* IN: which row to generate */
)
{
int		i;
int		j;
int *		prev;
int *		cur;
int *		temp;

	prev = NEWA (n + 1, int);
	cur  = NEWA (n + 1, int);

	/* {0 \choose 0} ==> 1 */
	j = 0;
	cur [0] = 1;

	for (; j <= n; j++) {
		temp = cur;
		cur  = prev;
		prev = temp;

		/* {n \choose 0} ==> 1 */
		cur [0] = 1;
		for (i = 1; i < j; i++) {
			cur [i] = prev [i - 1] + prev [i];
		}
		cur [j] = 1;
	}

	free (prev);

	return (cur);
}

/*
 * Create names for each of the variables.
 */

	static
	const char **
create_var_names (

int		nterms,		/* IN: number of terminal vertices */
int		nsteins,	/* IN: number of Steiner vertices */
int		nedges,		/* IN: number of edges */
int **		edge_list	/* IN: vertices contained in each edge */
)
{
int		j, v, nverts;
const char **	var_names;
char *		s;
int *		ip1;
int *		ip2;
char		buffer [128];

	nverts = nterms + nsteins;

	var_names = NEWA (nedges + nsteins, const char *);

	if (nverts <= 10) {
		/* We can use single digits for each vertex. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				*s++ = '0' + v;
			}
			*s = '\0';
			var_names [j] = strdup (buffer);
		}
	}
	else {
		/* At least one vertex requires more than 1 digit. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				sprintf (s, "_%d", v);
				s = strchr (s, '\0');
			}
			var_names [j] = strdup (buffer);
		}
	}
	for (j = 0; j < nsteins; j++) {
		sprintf (buffer, "y%d", nterms + j);
		var_names [nedges + j] = strdup (buffer);
	}
	return (var_names);
}

/*
 * Enumerate all Steiner trees.
 */

	static
	struct IVecSet *
enumerate_steiner_trees (

const struct Polytope *	_poly		/* IN: polytope to enumerate	*/
					/*     all hypertrees for	*/
)
{
int			i, j, k;
int			nterms, nsteins, nverts, nedges;
int			kmasks, nmasks, vmasks;
int			count;
size_t			size, csize;
const struct STNGP *	poly;
int8u **		one_bits_in_byte;
int **			edge_list;
bitmap_t *		edge_tmap;
int **			term_trees;
bitmap_t *		cmasks;
bitmap_t *		incmasks;
bitmap_t *		bp1;
bitmap_t *		array;
struct iv_mask *	list;
int *			root_edges;
int *			vcount;
int *			ip;
int *			endp;
int **			ptr;
struct IVecSet *	xpts;
struct sinfo		sinfo;

	FATAL_ERROR_IF (_poly EQ NULL);
	FATAL_ERROR_IF (_poly -> get_extreme_points NE enumerate_steiner_trees);

	poly = (const struct STNGP *) _poly;

	nterms		= poly -> nterms;
	nsteins		= poly -> nsteins;
	nverts		= poly -> nverts;
	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	FATAL_ERROR_IF (poly -> base.dim NE nedges + nsteins);

	nmasks = BMAP_ELTS (nedges);
	kmasks = BMAP_ELTS (nverts);
	vmasks = BMAP_ELTS (nedges + nsteins);

	init_nbits ();
	one_bits_in_byte = create_one_bits_in_byte_table ();

	/* Build edge_tmap -- a bitmap of vertices for each edge. */
	size = nedges * kmasks;
	edge_tmap = NEWA (size, bitmap_t);
	(void) memset (edge_tmap, 0, size * sizeof (bitmap_t));
	bp1 = edge_tmap;
	for (i = 0; i < nedges; i++) {
		ip = edge_list [i];
		endp = edge_list [i + 1];
		while (ip < endp) {
			j = *ip++;
			SETBIT (bp1, j);
		}
		bp1 += kmasks;
	}

	/* Build term_trees -- a list of hyperedges for each terminal. */
	vcount = NEWA (nverts, int);
	memset (vcount, 0, nverts * sizeof (vcount [0]));
	k = 0;
	for (i = 0; i < nedges; i++) {
		ip = edge_list [i];
		endp = edge_list [i + 1];
		while (ip < endp) {
			j = *ip++;
			FATAL_ERROR_IF ((j < 0) OR (j >= nverts));
			++(vcount [j]);
			++k;
		}
	}
	ip = NEWA (k, int);
	term_trees	= NEWA (nverts + 1, int *);
	ptr		= NEWA (nverts, int *);
	for (i = 0; i < nverts; i++) {
		term_trees [i]	= ip;
		ptr [i]		= ip;
		ip += vcount [i];
	}
	free (vcount);
	term_trees [i] = ip;
	for (i = 0; i < nedges; i++) {
		ip = edge_list [i];
		endp = edge_list [i + 1];
		while (ip < endp) {
			j = *ip++;
			*(ptr [j])++ = i;
		}
	}
	for (i = 0; i < nverts; i++) {
		FATAL_ERROR_IF (ptr [i] NE term_trees [i + 1]);
	}
	free (ptr);

	/* Build cmasks -- a mask of compatible edges for each edge. */
	/* Build incmasks -- a mask of incompatible edges for each edge. */
	size = nedges * nmasks;
	cmasks	 = NEWA (size, bitmap_t);
	incmasks = NEWA (size, bitmap_t);
	(void) memset (cmasks, 0, size * sizeof (bitmap_t));
	(void) memset (incmasks, 0, size * sizeof (bitmap_t));
	for (i = 0; i < nedges; i++) {
		SETBIT (&incmasks [i * nmasks], i);
		for (j = i + 1; j < nedges; j++) {
			count = 0;
			ip = edge_list [i];
			endp = edge_list [i + 1];
			bp1 = &edge_tmap [j * kmasks];
			while (ip < endp) {
				k = *ip++;
				if (BITON (bp1, k)) {
					++count;
				}
			}
			if (count EQ 1) {
				SETBIT (&cmasks [i * nmasks], j);
				SETBIT (&cmasks [j * nmasks], i);
			}
			else if (count > 1) {
				SETBIT (&incmasks [i * nmasks], j);
				SETBIT (&incmasks [j * nmasks], i);
			}
		}
	}

	/* Build root_edges.  For each vertex i > 0, j = root_edges [i]	*/
	/* gives edge j = (0,i).  When vertex i is an ignored Steiner	*/
	/* vertex, we set bit j of incompat (pretending that edge (0,i)	*/
	/* is part of the tree) so that we never use any edge that is	*/
	/* incompatible with this.					*/
	root_edges = NEWA (nverts, int);
	for (i = 0; i < nverts; i++) {
		root_edges [i] = -1;
	}
	for (j = 0; j < nedges; j++) {
		ip = edge_list [j];
		endp = edge_list [j + 1];
		if (ip + 2 NE endp) continue;
		if (ip [0] NE 0) continue;
		i = ip [1];
		FATAL_ERROR_IF ((i <= 0) OR (i >= nverts));
		root_edges [i] = j;
	}
	FATAL_ERROR_IF (root_edges [0] NE -1);
	for (i = 1; i < nverts; i++) {
		FATAL_ERROR_IF (root_edges [i] < 0);
	}

	memset (&sinfo, 0, sizeof (sinfo));

	csize = nverts + 1;
	csize *= nmasks;

	sinfo.nterms		= nterms;
	sinfo.nsteins		= nsteins;
	sinfo.nedges		= nedges;
	sinfo.nmasks		= nmasks;
	sinfo.kmasks		= kmasks;
	sinfo.vmasks		= vmasks;
	sinfo.fsets		= NEWA (nverts, int);
	sinfo.root_edges	= root_edges;
	sinfo.st_chosen		= NEWA (nsteins, bool);
	sinfo.edge_list		= edge_list;
	sinfo.solution		= NEWA (vmasks, bitmap_t);
	sinfo.terms_left	= NEWA ((nverts + 1) * kmasks, bitmap_t);
	sinfo.compat		= NEWA (csize, bitmap_t);
	sinfo.incompat		= NEWA (csize, bitmap_t);
	sinfo.try		= NEWA (csize, bitmap_t);
	sinfo.omit		= NEWA (csize, bitmap_t);
	sinfo.cmasks		= cmasks;
	sinfo.incmasks		= incmasks;
	sinfo.edge_tmap		= edge_tmap;
	sinfo.one_bits_in_byte	= one_bits_in_byte;
	sinfo.term_trees	= term_trees;
	sinfo.ntrees		= 0;
	sinfo.root		= NULL;
	sinfo.hookp		= &sinfo.root;

	/* Outer recursion choose subset of Steiner vertices.  Inner	*/
	/* recursion enumerates trees over terminals + chosen Steiners.	*/
	recurse_steiners (&sinfo, nsteins);

	/* Copy the incidence vectors into a single array of bit masks.	*/
	/* Each incidence vector is vmasks words long.			*/

	array = NEWA (sinfo.ntrees * vmasks, bitmap_t);
	bp1 = array;
	for (;;) {
		list = sinfo.root;
		if (list EQ NULL) break;

		sinfo.root = list -> next;
		list -> next = NULL;

		memcpy (bp1, list -> mask, vmasks * sizeof (bitmap_t));
		bp1 += vmasks;

		free (list -> mask);
		free (list);
	}

	xpts = NEW (struct IVecSet);

	xpts -> nvecs	= sinfo.ntrees;
	xpts -> dim	= nedges + nsteins;
	xpts -> nmasks	= vmasks;
	xpts -> array	= array;

	free (sinfo.st_chosen);
	free (sinfo.root_edges);
	free (sinfo.fsets);
	sinfo.edge_list = NULL;
	free (sinfo.solution);
	free (sinfo.terms_left);
	free (sinfo.compat);
	free (sinfo.incompat);
	free (sinfo.try);
	free (sinfo.omit);
	free (incmasks);
	free (cmasks);

	free (edge_tmap);
	free (one_bits_in_byte [0]);
	free (one_bits_in_byte);

	free (term_trees [0]);
	free (term_trees);

	return (xpts);
}

/*
 * This routine initializes a table that is used to rapidly convert an
 * 8-bit byte into a sequence of bit-positions that are on within that
 * byte.  For example, it translates the byte 0xB7 into {0, 1, 2, 4, 5, 7}.
 */

	static
	int8u **
create_one_bits_in_byte_table (void)

{
int		i;
int		j;
int		byte;
int8u *		byte_array;
int8u **	ptr_array;
int8u *		p;

	byte_array = NEWA (1024, int8u);
	ptr_array	= NEWA (256 + 1, int8u *);

	p = &byte_array [0];
	for (i = 0; i < 256; i++) {
		ptr_array [i] = p;
		byte = i;
		j = 0;
		while (byte > 0) {
			if ((byte & 0x01) NE 0) {
				*p++ = j;
			}
			++j;
			byte >>= 1;
		}
	}

	/* Terminate last table entry. */
	ptr_array [i] = p;

	return (ptr_array);
}

/*
 * Recursively choose all possible subsets of the Steiner vertices.
 * Set bits in sip->terms_left for those Steiner vertices that have been
 * chosen to be part of the tree.  Clear bits in sip->terms_left for
 * Steiner vertices that must be ignored (thereby pretending they have
 * already been connected to the current partial tree).
 * For each chosen subset, recursively enumerate all trees.
 */

	static
	void
recurse_steiners (

struct sinfo *		sip,	/* IN/OUT: search info */
int			level	/* IN: recursion level */
)
{
	if (level <= 0) {
		FATAL_ERROR_IF (level NE 0);

		/* Base case: we have chosen a subset of Steiner vertices. */
		/* Now enumerate all spanning trees of the combined set.   */
		enumerate_spanning_trees (sip);
		return;
	}

	--level;

	/* We are considering Steiner vertex v = nterms + level.	*/

	/* Enumerate subsets that omit v. */
	sip -> st_chosen [level] = FALSE;
	recurse_steiners (sip, level);

	/* Enumerate subsets tha include v. */
	sip -> st_chosen [level] = TRUE;
	recurse_steiners (sip, level);
}

/*
 * Recursively enumerate all trees that span all terminals and
 * chosen Steiner vertices.
 */

	static
	void
enumerate_spanning_trees (

struct sinfo *	sip		/* IN/OUT: search info */
)
{
int			i, j, t, v, e, size, nterms, nsteins, nedges, nverts;
int			nmasks, kmasks, vcount;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
int *			tp1;
int *			endp;
int **			edge_list;
int **			term_trees;
bitmap_t *		edge_tmap;
bitmap_t *		cmasks;
bitmap_t *		incmasks;
bitmap_t *		terms_left;
bitmap_t *		solution;

	nterms		= sip -> nterms;
	nsteins		= sip -> nsteins;
	nedges		= sip -> nedges;
	kmasks		= sip -> kmasks;
	nmasks		= sip -> nmasks;
	edge_list	= sip -> edge_list;
	term_trees	= sip -> term_trees;
	edge_tmap	= sip -> edge_tmap;
	cmasks		= sip -> cmasks;
	incmasks	= sip -> incmasks;
	terms_left	= sip -> terms_left;
	solution	= sip -> solution;

	nverts = nterms + nsteins;

	bp1 = terms_left;
	for (i = 0; i < kmasks; i++) {
		*bp1++ = 0;
	}
	for (i = 0; i < nverts; i++) {
		SETBIT (terms_left, i);
	}

	size = (nmasks * sizeof (bitmap_t));
	memset (sip -> compat, 0, size);
	memset (sip -> incompat, 0, size);
	memset (sip -> try, 0, size);
	memset (sip -> omit, 0, size);
	memset (solution, 0, size);

	/* Initialize for selected subset of Steiner vertices. */
	vcount = nterms + nsteins;
	for (i = 0; i < nsteins; i++) {
		v = nterms + i;
		j = nedges + i;
		if (sip -> st_chosen [i]) {
			/* We are using Steiner vertex v. */
			SETBIT (terms_left, v);
			SETBIT (solution, j);
		}
		else {
			/* We are ignoring Steiner vertex v. */
			CLRBIT (terms_left, v);
			CLRBIT (solution, j);

			/* Pretend edge e=(0,v) is part of solution.	*/
			/* Make all edges incompatiblt with e be	*/
			/* incompatible with successive partial trees.	*/
			e = sip -> root_edges [v];
			FATAL_ERROR_IF ((e < 0) OR (e >= nedges));
			bp1 = sip -> incompat;
			bp2 = &incmasks [e * nmasks];
			bp3 = bp1 + nmasks;
			while (bp1 < bp3) {
				*bp1++ |= *bp2++;
			}

			/* One fewer vertex to include in enumerated trees. */
			--vcount;
		}
	}

	/* We start at terminal 0... */

	tp1	= term_trees [0];
	endp	= term_trees [1];
	while (tp1 < endp) {
		t = *tp1++;

		if (BITON (sip -> incompat, t)) {
			/* Edge t contains an ignored Steiner vertex.	*/
			/* Skip it.					*/
			continue;
		}

		sip -> fsets [0] = t;
		SETBIT (solution, t);

		bp1 = terms_left;
		bp2 = &edge_tmap [t * kmasks];
		bp3 = bp1 + kmasks;
		for (i = 0; i < kmasks; i++) {
			*bp3++ = *bp1++ & ~*bp2++;
		}

		bp1 = &cmasks [t * nmasks];
		bp2 = sip -> compat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp2++ = *bp1++;
		}

		bp1 = &incmasks [t * nmasks];
		bp2 = sip -> omit;
		bp3 = sip -> incompat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp3++ = *bp1++ | *bp2++;
		}

		ht_recurse (sip,
			    1,
			    vcount - (edge_list [t + 1] - edge_list [t]));

		CLRBIT (solution, t);
		SETBIT (sip -> omit, t);
	}
}

/*
 * This routine performs the recursive portion of the backtrack search.
 */

	static
	void
ht_recurse (

struct sinfo *	sip,		/* IN/OUT - search info */
int		level,		/* IN - recursion level */
int		nleft		/* IN - number of terminals left to connect */
)
{
int			i, j, k;
int			nmasks, kmasks, vmasks;
int			fset_offset;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
bitmap_t		mask;
bitmap_t *		terms_left;
bitmap_t *		compat;
bitmap_t *		incompat;
bitmap_t *		try;
bitmap_t *		omit;
bitmap_t *		cmasks;
bitmap_t *		incmasks;
int			i1;
int **			edge_list;
int8u **		one_bits_in_byte;
bitmap_t *		edge_tmap;
bitmap_t *		wordp;
bitmap_t		word;
int			basebit;
int			bitno;
int8u *			p1;
int8u *			p2;
struct iv_mask *	list;
int			byte;

	nmasks	= sip -> nmasks;
	kmasks	= sip -> kmasks;

	if (nleft <= 0) {
		FATAL_ERROR_IF (nleft < 0);

		/* All terminals have been connected!  We have	*/
		/* a new hypertree.  Add it to the list.	*/

		vmasks	= sip -> vmasks;
		bp1 = NEWA (vmasks, bitmap_t);
		memcpy (bp1, sip -> solution, vmasks * sizeof (bitmap_t));

		list = NEW (struct iv_mask);
		list -> mask	= bp1;
		list -> next	= NULL;
		*(sip -> hookp) = list;
		sip -> hookp = &(list -> next);

		++(sip -> ntrees);

		return;
	}

	terms_left	= sip -> terms_left + level * kmasks;

	fset_offset	= level * nmasks;

	compat		= sip -> compat	  + fset_offset;
	incompat	= sip -> incompat + fset_offset;
	try		= sip -> try	  + fset_offset;
	omit		= sip -> omit	  + fset_offset;

	/* Determine the set of trees we will try to add to the	*/
	/* partial solution.					*/

	bp1 = compat;
	bp2 = incompat;
	bp3 = try;
	for (i = 0; i < nmasks; i++) {
		*bp3++ = *bp1++ & ~*bp2++;
	}

	bp1 = omit;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	/* Loop over each feasible hyperedge. */
	edge_list	 = sip -> edge_list;
	one_bits_in_byte = sip -> one_bits_in_byte;
	edge_tmap	 = sip -> edge_tmap;
	cmasks		 = sip -> cmasks;
	incmasks	 = sip -> incmasks;
	wordp = try;
	for (j = 0, basebit = 0; j < nmasks; basebit += BPW, j++) {
		word = *wordp++;
		if (word EQ 0) continue;
		for (bitno = basebit; word NE 0; bitno += 8, word >>= 8) {
			byte = (word & 0xFF);
			if (byte EQ 0) continue;
			p1 = one_bits_in_byte [byte];
			p2 = one_bits_in_byte [byte + 1];
			while (p1 < p2) {
				k = bitno + *p1++;
				/* Hyperedge K is feasible.  Test to see */
				/* if it is truly compatible...		 */
				bp1 = &edge_tmap [k * kmasks];
				bp2 = terms_left;
				i1 = 0;
				for (i = 0; i < kmasks; i++) {
					mask = *bp1++ & ~*bp2++;
					i1 += NBITSON (mask);
				}
				if (i1 NE 1) continue;
				/* Tree K intersects the current partial */
				/* solution at exactly one point!	 */

				bp1 = compat;
				bp2 = &cmasks [k * nmasks];
				bp3 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp3++ = *bp1++ | *bp2++;
				}

				bp1 = incompat;
				bp2 = &incmasks [k * nmasks];
				bp3 = omit;
				bp4 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp4++ = (*bp1++ | *bp2++ | *bp3++);
				}

				bp1 = terms_left;
				bp2 = &edge_tmap [k * kmasks];
				bp3 = bp1 + kmasks;
				for (i = 0; i < kmasks; i++) {
					*bp3++ = *bp1++ & ~*bp2++;
				}

				sip -> fsets [level] = k;
				SETBIT (sip -> solution, k);

				ht_recurse (sip,
					    level + 1,
					    nleft
					      - (edge_list [k + 1]
						 - edge_list [k]) + 1);

				CLRBIT (sip -> solution, k);
				SETBIT (omit, k);
			}
		}
	}
}

/*
 * Decode the given name and return the corresponding inequality.
 */

	static
	struct Constraint *
get_STNGP_constraint (

const struct Polytope *	_poly,		/* IN: polytope to get constraint of */
const char *		name		/* IN: name of constraint */
)
{
const struct STNGP *	poly;
const char *		s;
struct Constraint *	cp;

	static const char *	sec_names [] = {
		"sec-", "sec", "subtour-", "subtour", NULL
	};
	static const char *	nonneg_x_names [] = {
		"nnx", "nonnegx", NULL
	};
	static const char *	nonneg_y_names [] = {
		"nny", "nonnegy", NULL
	};

	FATAL_ERROR_IF ((_poly EQ NULL) OR (name EQ NULL));
	FATAL_ERROR_IF (_poly -> get_constraint NE get_STNGP_constraint);

	poly = (const struct STNGP *) _poly;

	s = match_mixed_case (name, sec_names);
	if (s NE NULL) {
		cp = get_STNGP_subtour (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, nonneg_x_names);
	if (s NE NULL) {
		/* x_e >= 0 */
		cp = NEW (struct Constraint);
		cp -> rmatind	= NEWA (1, int);
		cp -> rmatval	= NEWA (1, double);
		cp -> sense	= 'G';
		cp -> rhs	= 0.0;
		cp -> next	= NULL;

		cp -> rmatind [0] = 0;
		cp -> rmatval [0] = 1.0;

		return (cp);
	}

	s = match_mixed_case (name, nonneg_y_names);
	if (s NE NULL) {
		/* y_s >= 0 */
		if (s [0] NE '\0') {
			fprintf (stderr, "Invalid constraint name `%s'.\n",
				 name);
			return (NULL);
		}
		if (poly -> nsteins < 1) {
			fprintf (stderr, "No y variables present.\n");
			return (NULL);
		}
		cp = NEW (struct Constraint);
		cp -> rmatind	= NEWA (1, int);
		cp -> rmatval	= NEWA (1, double);
		cp -> sense	= 'G';
		cp -> rhs	= 0.0;
		cp -> next	= NULL;

		cp -> rmatind [0] = poly -> nedges;
		cp -> rmatval [0] = 1.0;

		return (cp);
	}

	return (NULL);
}

/*
 * Parse the subtour size (#nterms and #steins), and return either the
 * correct subtour inequality (for terminals 0...{nterms-1} and
 * Steiner vertices (nterms...nterms+nsteins-1) or NULL if the string
 * is invalid.
 */

	static
	struct Constraint *
get_STNGP_subtour (

const struct STNGP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, nterms, nedges, dim, nz, count, v;
int			nstterms, nststeins, stsize;
int **			edge_list;
int *			rmatind;
double *		rmatval;
int *			ip1;
int *			ip2;
struct Constraint *	cp;

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &nstterms);	/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE ',') return (NULL);

	s = get_int (s + 1, &nststeins);

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	stsize = nstterms + nststeins;

	if ((nstterms < 0)  OR (nstterms > poly -> nterms) OR
	    (nststeins < 0) OR (nststeins > poly -> nsteins) OR
	    (stsize < 2) OR (stsize > poly -> nverts)) {
		fprintf (stderr,
			 "Subtour of %d terms and %d Steiners is invalid.\n",
			 nstterms, nststeins);
		return (NULL);
	}

	dim		= poly -> base.dim;
	nterms		= poly -> nterms;
	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	rmatind	= NEWA (dim, int);
	rmatval	= NEWA (dim, double);

	if (nstterms > 0) {
		/* At least one terminal -- regular subtour. */
		nz = 0;
		for (j = 0; j < nedges; j++) {
			count = 0;
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				if ((v < nstterms) OR
				    ((v >= nterms) AND (v < nterms + nststeins))) {
					++count;
				}
			}
			if (count > 1) {
				rmatind [nz] = j;
				rmatval [nz] = count - 1;
				++nz;
			}
		}
		for (j = 0; j < nststeins; j++) {
			rmatind [nz] = nedges + j;
			rmatval [nz] = -1.0;
			++nz;
		}
		cp = make_constraint (nz, rmatind, rmatval, 'L', nstterms - 1);
	}
	else {
		/* Subtour contains only Steiner vertices. */
		nz = 0;
		for (j = 0; j < nedges; j++) {
			count = 0;
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				if ((v < nstterms) OR
				    ((v >= nterms) AND (v < nterms + nststeins))) {
					++count;
				}
			}
			if (count > 1) {
				rmatind [nz] = j;
				rmatval [nz] = count - 1;
				++nz;
			}
		}
		/* Omit first Steiner vertex from y(S). */
		for (j = 1; j < nststeins; j++) {
			rmatind [nz] = nedges + j;
			rmatval [nz] = -1.0;
			++nz;
		}
		cp = make_constraint (nz, rmatind, rmatval, 'L', 0.0);
	}

	free (rmatval);
	free (rmatind);

	return (cp);
}

/*
 * Decode the given name and return the corresponding inequality.
 */

	static
	struct Constraint *
get_STNHGP_constraint (

const struct Polytope *	_poly,		/* IN: polytope to get constraint of */
const char *		name		/* IN: name of constraint */
)
{
const struct STNGP *	poly;
const char *		s;
struct Constraint *	cp;

	static const char *	sec_names [] = {
		"sec-", "sec", "subtour-", "subtour", NULL
	};
	static const char *	nonneg_x_names [] = {
		"nnx-", "nnx", "nonnegx-", "nonnegx", NULL
	};
	static const char *	nonneg_y_names [] = {
		"nny-", "nny", "nonnegy-", "nonnegy", NULL
	};

	FATAL_ERROR_IF ((_poly EQ NULL) OR (name EQ NULL));
	FATAL_ERROR_IF (_poly -> get_constraint NE get_STNHGP_constraint);

	poly = (const struct STNGP *) _poly;

	s = match_mixed_case (name, sec_names);
	if (s NE NULL) {
		cp = get_STNHGP_subtour (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, nonneg_x_names);
	if (s NE NULL) {
		cp = get_STNHGP_nonneg_x_constraint (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, nonneg_y_names);
	if (s NE NULL) {
		if (s [0] NE '\0') {
			fprintf (stderr, "Invalid constraint name `%s'.\n",
				 name);
			return (NULL);
		}
		if (poly -> nsteins < 1) {
			fprintf (stderr, "No y variables present.\n");
			return (NULL);
		}
		cp = NEW (struct Constraint);
		cp -> rmatind	= NEWA (1, int);
		cp -> rmatval	= NEWA (1, double);
		cp -> sense	= 'G';
		cp -> rhs	= 0.0;
		cp -> next	= NULL;

		cp -> rmatind [0] = poly -> nedges;
		cp -> rmatval [0] = 1.0;

		return (cp);
	}

	return (NULL);
}

/*
 * Parse the subtour size, and return either the correct subtour
 * inequality (for vertices 0...{k-1}, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_STNHGP_subtour (

const struct STNGP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, nedges, dim, nz, count, v;
int			nstterms, nststeins, stsize;
int **			edge_list;
int *			rmatind;
double *		rmatval;
int *			ip1;
int *			ip2;
struct Constraint *	cp;

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &nstterms);	/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE ',') return (NULL);

	s = get_int (s + 1, &nststeins);

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	stsize = nstterms + nststeins;

	if ((nstterms < 0)  OR (nstterms > poly -> nterms) OR
	    (nststeins < 0) OR (nststeins > poly -> nsteins) OR
	    (stsize < 2) OR (stsize > poly -> nverts)) {
		fprintf (stderr,
			 "Subtour of %d terms and %d Steiners is invalid.\n",
			 nstterms, nststeins);
		return (NULL);
	}

	dim		= poly -> base.dim;
	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	rmatind	= NEWA (dim, int);
	rmatval	= NEWA (dim, double);

	if (nstterms > 0) {
		/* At least one terminal -- regular subtour. */
		nz = 0;
		for (j = 0; j < nedges; j++) {
			count = 0;
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				if (v < nstterms) {
					++count;
				}
			}
			if (count > 1) {
				rmatind [nz] = j;
				rmatval [nz] = count - 1;
				++nz;
			}
		}
		for (j = 0; j < nststeins; j++) {
			rmatind [nz] = nedges + j;
			rmatval [nz] = -1.0;
			++nz;
		}
		cp = make_constraint (nz, rmatind, rmatval, 'L', nstterms - 1);
	}
	else {
		/* Subtour contains only Steiner vertices. */
		nz = 0;
		for (j = 0; j < nedges; j++) {
			count = 0;
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				if (v < nstterms) {
					++count;
				}
			}
			if (count > 1) {
				rmatind [nz] = j;
				rmatval [nz] = count - 1;
				++nz;
			}
		}
		/* Omit first Steiner vertex from y(S). */
		for (j = 1; j < nststeins; j++) {
			rmatind [nz] = nedges + j;
			rmatval [nz] = -1.0;
			++nz;
		}
		cp = make_constraint (nz, rmatind, rmatval, 'L', 0.0);
	}

	free (rmatval);
	free (rmatind);

	return (cp);
}

/*
 * Parse the edge size, and return either the correct non-negativity
 * inequality (for vertices 0...{k-1}, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_STNHGP_nonneg_x_constraint (

const struct STNGP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, k;
int			nedges;
int **			edge_list;
int *			ip1;
int *			ip2;
struct Constraint *	cp;
int			rmatind [1];
double			rmatval [1];

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &k);		/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	if ((k < 2) OR (k > poly -> nverts)) {
		fprintf (stderr, "Edge size %d is out of range [2,%d]\n",
			 k, poly -> nverts);
		return (NULL);
	}

	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	for (j = 0; ; j++) {
		FATAL_ERROR_IF (j >= nedges);
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		if (ip2 - ip1 EQ k) {
			/* Found first k-vertex edge! */
			break;
		}
	}

	rmatind [0] = j;
	rmatval [0] = 1.0;

	cp = make_constraint (1, rmatind, rmatval, 'G', 0.0);

	return (cp);
}

/*
 * Return the edge_list data structure.
 */

	int **
get_STNGP_edge_list (

const struct Polytope *	_poly		/* IN: polytope to get edge_list of */
)
{
const struct STNGP *	poly;

	FATAL_ERROR_IF (_poly EQ NULL);
	FATAL_ERROR_IF (_poly -> free NE free_STNHGP);

	poly = (const struct STNGP *) _poly;

	return (poly -> edge_list);
}

/*
 * Free up the given STHGP polytope.
 */

	static
	void
free_STNHGP (

struct Polytope *	poly	/* IN: Polytope to free */
)
{
struct STNGP *		p;
int **			edge_list;
int *			ip;

	if (poly NE NULL) {
		FATAL_ERROR_IF (poly -> free NE free_STNHGP);

		p = (struct STNGP *) poly;

		edge_list = p -> edge_list;
		if (edge_list NE NULL) {
			ip = edge_list [0];
			if (ip NE NULL) {
				free (ip);
			}
			free (edge_list);
		}
		p -> edge_list = NULL;

		clear_polytope (&(p -> base));
		free (p);
	}
}
