/***********************************************************************

	File:	tsp_indicators.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute strength indicators for polytope TSP(n) via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "tsp_indicators.h"

#include "binomial.h"
#include "fatal.h"
#include <gmp.h>
#include "logic.h"

static int	compute_Cho (int h, int o);
static void	eval_terms (mpz_ptr		res,
			    const signed char *	terms,
			    const int *		val);

/*
 * Return the number of extreme points of TSP(n).
 */

	void
tsp_num_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
unsigned long int	rem;
mpz_t			TEMP;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n <= 0);

	mpz_init (TEMP);

	mpz_fac_ui (TEMP, n - 1);
	rem = mpz_tdiv_q_ui (TEMP, TEMP, 2);
	FATAL_ERROR_IF (rem NE 0);

	mpz_set (res, TEMP);

	mpz_clear (TEMP);
}

/*
 * Let e \in E.  Return the number of TSP(n) extreme points that contain
 * edge e.
 */

	void
tsp_num_edge_extreme_points (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
	mpz_fac_ui (res, n - 2);
}

/*
 * Return the number of TSP(n) extreme points incident to the
 * non-negativity inequality for some edge e.
 */

	void
tsp_nonneg_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpz_t		TEMP1, TEMP2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n <= 1);

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	tsp_num_extreme_points (TEMP1, n);

	/* Number of tours that contain edge e. */
	mpz_fac_ui (TEMP2, n - 2);

	/* Number of tours that DO NOT contain edge e. */
	mpz_sub (res, TEMP1, TEMP2);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Return the Extreme Point Ratio of the TSP(n) non-negativity inequality
 * for some edge e.
 */

	void
tsp_nonneg_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpq_t		RATIO;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 3);

	mpq_init (RATIO);

	mpz_set_ui (mpq_numref (RATIO), n - 3);
	mpz_set_ui (mpq_denref (RATIO), n - 1);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpq_clear (RATIO);
}

/*
 * Return the Centroid Distance squared of the TSP(n) non-negativity
 * inequality for some edge e.
 */

	void
tsp_nonneg_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n		/* IN: Number of vertices */
)
{
mpq_t		DIST2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 3);

	mpq_init (DIST2);

	mpz_set_ui (mpq_numref (DIST2), 4);
	mpz_set_ui (mpq_denref (DIST2), (n - 1) * (n - 3));
	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpq_clear (DIST2);
}

/*
 * Return the number of TSP(n) extreme points incident to a
 * subtour inequality S of cardinality k.
 */

	void
tsp_subtour_ep_count (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
unsigned long int	rem;
mpz_t			TEMP1, TEMP2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF (k < 2);
	FATAL_ERROR_IF (k >= n);

	mpz_init (TEMP1);
	mpz_init (TEMP2);

	mpz_fac_ui (TEMP1, k);
	mpz_fac_ui (TEMP2, n - k);
	mpz_mul (TEMP1, TEMP1, TEMP2);
	rem = mpz_tdiv_q_ui (TEMP1, TEMP1, 2);
	FATAL_ERROR_IF (rem NE 0);

	mpz_set (res, TEMP1);

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
}

/*
 * Return the Extreme Point Ratio of a TSP(n) subtour S of cardinality k.
 */

	void
tsp_subtour_epr (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
mpq_t		RATIO;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF (k < 2);
	FATAL_ERROR_IF (k >= n);

	mpq_init (RATIO);

	mpz_set_ui (mpq_numref (RATIO), n);
	binomial (mpq_denref (RATIO), n, k);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpq_clear (RATIO);
}

/*
 * Return the Centroid Distance squared of a TSP(n) subtour S of
 * cardinality k.
 */

	void
tsp_subtour_cd2 (

mpq_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Number of vertices */
int		k		/* IN: Cardinality of subtour */
)
{
mpz_t		N, D;
mpq_t		DIST2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF (k < 2);
	FATAL_ERROR_IF (k >= n);

	mpq_init (DIST2);
	mpz_init (N);
	mpz_init (D);

	mpz_set_ui (N, 2);
	mpz_mul_ui (N, N, k - 1);
	mpz_mul_ui (N, N, n - 2);
	mpz_mul_ui (N, N, n - k - 1);

	mpz_set_ui (D, k);
	mpz_mul_ui (D, D, n - 1);
	mpz_mul_ui (D, D, n - k);

	mpz_set (mpq_numref (DIST2), N);
	mpz_set (mpq_denref (DIST2), D);
	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpz_clear (D);
	mpz_clear (N);
	mpq_clear (DIST2);
}

/*
 * Return the number of TSP(n) extreme points incident to a
 * 3-toothed comb inequality.
 */

	void
tsp_3tc_ep_count (

mpz_ptr		res,	/* IN/OUT: Result */
int		b1,	/* IN: Parameter b1 */
int		t1,	/* IN: Parameter t1 */
int		b2,	/* IN: Parameter b2 */
int		t2,	/* IN: Parameter t2 */
int		b3,	/* IN: Parameter b3 */
int		t3,	/* IN: Parameter t3 */
int		h,	/* IN: Parameter h */
int		o	/* IN: Parameter o */
)
{
int		Cho;
mpz_t		PROD, F;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (b1 <= 0);
	FATAL_ERROR_IF (t1 <= 0);
	FATAL_ERROR_IF (b2 <= 0);
	FATAL_ERROR_IF (t2 <= 0);
	FATAL_ERROR_IF (b3 <= 0);
	FATAL_ERROR_IF (t3 <= 0);
	FATAL_ERROR_IF (h  <  0);
	FATAL_ERROR_IF (o  <  0);

	Cho = compute_Cho (h, o);

	mpz_init (PROD);
	mpz_init (F);

#define UPDATE(k) \
	do { \
		mpz_fac_ui (F, k); \
		mpz_mul (PROD, PROD, F); \
	} while (FALSE)

	mpz_set_ui (PROD, Cho);
	UPDATE (b1);
	UPDATE (t1);
	UPDATE (b2);
	UPDATE (t2);
	UPDATE (b3);
	UPDATE (t3);
	UPDATE (h);
	UPDATE (o);

#undef UPDATE

	mpz_set (res, PROD);

	mpz_clear (F);
	mpz_clear (PROD);
}

/*
 * Return the Extreme Point Ratio of a TSP(n) 3-toothed comb inequality.
 *
 * Note: the closed-form formula for EPR of 3-Toothed Combs is WRONG!
 * There are tours incident to the comb whose structure does not match
 * any of the "primtive comb" tours, so this formula undercounts the
 * incident tours.
 */

	void
tsp_3tc_epr (

mpq_ptr		res,	/* IN/OUT: Result */
int		b1,	/* IN: Parameter b1 */
int		t1,	/* IN: Parameter t1 */
int		b2,	/* IN: Parameter b2 */
int		t2,	/* IN: Parameter t2 */
int		b3,	/* IN: Parameter b3 */
int		t3,	/* IN: Parameter t3 */
int		h,	/* IN: Parameter h */
int		o	/* IN: Parameter o */
)
{
int		n;
mpz_t		TOTAL, COUNT;
mpq_t		RATIO;

	n = b1 + t1 + b2 + t2 + b3 + t3 + h + o;

	mpq_init (RATIO);
	mpz_init (TOTAL);
	mpz_init (COUNT);

	tsp_num_extreme_points (TOTAL, n);
	tsp_3tc_ep_count (COUNT, b1, t1, b2, t2, b3, t3, h, o);

	mpz_set (mpq_numref (RATIO), COUNT);
	mpz_set (mpq_denref (RATIO), TOTAL);
	mpq_canonicalize (RATIO);

	mpq_set (res, RATIO);

	mpz_clear (COUNT);
	mpz_clear (TOTAL);
	mpq_clear (RATIO);
}

/*
 * Compute the number of "primitive" 3-toothed combs (for the case
 * where b1=t1=b2=t2=b3=t3=1).
 */

	static
	int
compute_Cho (

int		h,		/* IN: Parameter h */
int		o		/* IN: Parameter o */
)
{
int		k;

	k = (h > 0) + (o > 0);
	switch (k) {
	case 0:		return (9);
	case 1:		return (18);
	case 2:		return (36);
	}

	FATAL_ERROR;
	return (0);
}

/*
 * The terms for sub-expression A of the CD squared for TSP(n) 3-toothed combs.
 */

static signed char	A_terms [] = {
	1, 0, 1, 1, 1, -1,
	2, 0, 1, 2, 1, -1,
	3, 0, 1, 3, 1, -1,
	2, 0, 1, 4, 1, -1,
	3, 0, 1, 5, 1, -1,
	1, 0, 1, 6, 1, -1,
	2, 0, 1, 7, 1, -1,
	-5, 0, 1, -1,
	3, 1, 1, 2, 1, -1,
	2, 1, 1, 3, 1, -1,
	3, 1, 1, 4, 1, -1,
	2, 1, 1, 5, 1, -1,
	2, 1, 1, 6, 1, -1,
	1, 1, 1, 7, 1, -1,
	-5, 1, 1, -1,
	1, 2, 1, 3, 1, -1,
	2, 2, 1, 4, 1, -1,
	3, 2, 1, 5, 1, -1,
	1, 2, 1, 6, 1, -1,
	2, 2, 1, 7, 1, -1,
	-5, 2, 1, -1,
	3, 3, 1, 4, 1, -1,
	2, 3, 1, 5, 1, -1,
	2, 3, 1, 6, 1, -1,
	1, 3, 1, 7, 1, -1,
	-5, 3, 1, -1,
	1, 4, 1, 5, 1, -1,
	1, 4, 1, 6, 1, -1,
	2, 4, 1, 7, 1, -1,
	-5, 4, 1, -1,
	2, 5, 1, 6, 1, -1,
	1, 5, 1, 7, 1, -1,
	-5, 5, 1, -1,
	1, 6, 1, 7, 1, -1,
	-5, 6, 1, -1,
	-5, 7, 1, -1,
	5, -1,
	0,
};

/*
 * The terms for sub-expression B of the CD squared for TSP(n) 3-toothed combs.
 */

static signed char	B_terms [] = {
	1, 0, 2, 1, 2, -1,
	2, 0, 2, 1, 1, 3, 1, -1,
	2, 0, 2, 1, 1, 5, 1, -1,
	2, 0, 2, 1, 1, 7, 1, -1,
	-1, 0, 2, 1, 1, -1,
	4, 0, 2, 2, 2, -1,
	8, 0, 2, 2, 1, 3, 1, -1,
	2, 0, 2, 2, 1, 4, 1, -1,
	2, 0, 2, 2, 1, 5, 1, -1,
	2, 0, 2, 2, 1, 6, 1, -1,
	2, 0, 2, 2, 1, 7, 1, -1,
	-4, 0, 2, 2, 1, -1,
	9, 0, 2, 3, 2, -1,
	2, 0, 2, 3, 1, 4, 1, -1,
	8, 0, 2, 3, 1, 5, 1, -1,
	2, 0, 2, 3, 1, 6, 1, -1,
	8, 0, 2, 3, 1, 7, 1, -1,
	-9, 0, 2, 3, 1, -1,
	4, 0, 2, 4, 2, -1,
	8, 0, 2, 4, 1, 5, 1, -1,
	2, 0, 2, 4, 1, 6, 1, -1,
	2, 0, 2, 4, 1, 7, 1, -1,
	-4, 0, 2, 4, 1, -1,
	9, 0, 2, 5, 2, -1,
	2, 0, 2, 5, 1, 6, 1, -1,
	8, 0, 2, 5, 1, 7, 1, -1,
	-9, 0, 2, 5, 1, -1,
	1, 0, 2, 6, 2, -1,
	2, 0, 2, 6, 1, 7, 1, -1,
	-1, 0, 2, 6, 1, -1,
	4, 0, 2, 7, 2, -1,
	-4, 0, 2, 7, 1, -1,
	2, 0, 1, 1, 2, 2, 1, -1,
	2, 0, 1, 1, 2, 4, 1, -1,
	2, 0, 1, 1, 2, 6, 1, -1,
	-1, 0, 1, 1, 2, -1,
	8, 0, 1, 1, 1, 2, 2, -1,
	12, 0, 1, 1, 1, 2, 1, 3, 1, -1,
	4, 0, 1, 1, 1, 2, 1, 4, 1, -1,
	4, 0, 1, 1, 1, 2, 1, 5, 1, -1,
	4, 0, 1, 1, 1, 2, 1, 6, 1, -1,
	4, 0, 1, 1, 1, 2, 1, 7, 1, -1,
	-10, 0, 1, 1, 1, 2, 1, -1,
	8, 0, 1, 1, 1, 3, 2, -1,
	4, 0, 1, 1, 1, 3, 1, 4, 1, -1,
	4, 0, 1, 1, 1, 3, 1, 5, 1, -1,
	4, 0, 1, 1, 1, 3, 1, 6, 1, -1,
	4, 0, 1, 1, 1, 3, 1, 7, 1, -1,
	-10, 0, 1, 1, 1, 3, 1, -1,
	8, 0, 1, 1, 1, 4, 2, -1,
	12, 0, 1, 1, 1, 4, 1, 5, 1, -1,
	4, 0, 1, 1, 1, 4, 1, 6, 1, -1,
	4, 0, 1, 1, 1, 4, 1, 7, 1, -1,
	-10, 0, 1, 1, 1, 4, 1, -1,
	8, 0, 1, 1, 1, 5, 2, -1,
	4, 0, 1, 1, 1, 5, 1, 6, 1, -1,
	4, 0, 1, 1, 1, 5, 1, 7, 1, -1,
	-10, 0, 1, 1, 1, 5, 1, -1,
	2, 0, 1, 1, 1, 6, 2, -1,
	4, 0, 1, 1, 1, 6, 1, 7, 1, -1,
	-4, 0, 1, 1, 1, 6, 1, -1,
	2, 0, 1, 1, 1, 7, 2, -1,
	-4, 0, 1, 1, 1, 7, 1, -1,
	1, 0, 1, 1, 1, -1,
	2, 0, 1, 2, 2, 4, 1, -1,
	2, 0, 1, 2, 2, 5, 1, -1,
	2, 0, 1, 2, 2, 6, 1, -1,
	2, 0, 1, 2, 2, 7, 1, -1,
	-4, 0, 1, 2, 2, -1,
	2, 0, 1, 2, 1, 3, 2, -1,
	4, 0, 1, 2, 1, 3, 1, 4, 1, -1,
	4, 0, 1, 2, 1, 3, 1, 5, 1, -1,
	4, 0, 1, 2, 1, 3, 1, 6, 1, -1,
	4, 0, 1, 2, 1, 3, 1, 7, 1, -1,
	-10, 0, 1, 2, 1, 3, 1, -1,
	2, 0, 1, 2, 1, 4, 2, -1,
	4, 0, 1, 2, 1, 4, 1, 5, 1, -1,
	-6, 0, 1, 2, 1, 4, 1, -1,
	8, 0, 1, 2, 1, 5, 2, -1,
	4, 0, 1, 2, 1, 5, 1, 7, 1, -1,
	-12, 0, 1, 2, 1, 5, 1, -1,
	-4, 0, 1, 2, 1, 6, 1, -1,
	2, 0, 1, 2, 1, 7, 2, -1,
	-6, 0, 1, 2, 1, 7, 1, -1,
	4, 0, 1, 2, 1, -1,
	8, 0, 1, 3, 2, 4, 1, -1,
	2, 0, 1, 3, 2, 5, 1, -1,
	8, 0, 1, 3, 2, 6, 1, -1,
	2, 0, 1, 3, 2, 7, 1, -1,
	-9, 0, 1, 3, 2, -1,
	2, 0, 1, 3, 1, 4, 2, -1,
	4, 0, 1, 3, 1, 4, 1, 5, 1, -1,
	4, 0, 1, 3, 1, 4, 1, 7, 1, -1,
	-12, 0, 1, 3, 1, 4, 1, -1,
	2, 0, 1, 3, 1, 5, 2, -1,
	4, 0, 1, 3, 1, 5, 1, 6, 1, -1,
	-12, 0, 1, 3, 1, 5, 1, -1,
	4, 0, 1, 3, 1, 6, 1, 7, 1, -1,
	-10, 0, 1, 3, 1, 6, 1, -1,
	-10, 0, 1, 3, 1, 7, 1, -1,
	9, 0, 1, 3, 1, -1,
	2, 0, 1, 4, 2, 6, 1, -1,
	2, 0, 1, 4, 2, 7, 1, -1,
	-4, 0, 1, 4, 2, -1,
	2, 0, 1, 4, 1, 5, 2, -1,
	4, 0, 1, 4, 1, 5, 1, 6, 1, -1,
	4, 0, 1, 4, 1, 5, 1, 7, 1, -1,
	-10, 0, 1, 4, 1, 5, 1, -1,
	-4, 0, 1, 4, 1, 6, 1, -1,
	2, 0, 1, 4, 1, 7, 2, -1,
	-6, 0, 1, 4, 1, 7, 1, -1,
	4, 0, 1, 4, 1, -1,
	8, 0, 1, 5, 2, 6, 1, -1,
	2, 0, 1, 5, 2, 7, 1, -1,
	-9, 0, 1, 5, 2, -1,
	4, 0, 1, 5, 1, 6, 1, 7, 1, -1,
	-10, 0, 1, 5, 1, 6, 1, -1,
	-10, 0, 1, 5, 1, 7, 1, -1,
	9, 0, 1, 5, 1, -1,
	-1, 0, 1, 6, 2, -1,
	2, 0, 1, 6, 1, 7, 2, -1,
	-4, 0, 1, 6, 1, 7, 1, -1,
	1, 0, 1, 6, 1, -1,
	-4, 0, 1, 7, 2, -1,
	4, 0, 1, 7, 1, -1,
	9, 1, 2, 2, 2, -1,
	8, 1, 2, 2, 1, 3, 1, -1,
	8, 1, 2, 2, 1, 4, 1, -1,
	2, 1, 2, 2, 1, 5, 1, -1,
	8, 1, 2, 2, 1, 6, 1, -1,
	2, 1, 2, 2, 1, 7, 1, -1,
	-9, 1, 2, 2, 1, -1,
	4, 1, 2, 3, 2, -1,
	2, 1, 2, 3, 1, 4, 1, -1,
	2, 1, 2, 3, 1, 5, 1, -1,
	2, 1, 2, 3, 1, 6, 1, -1,
	2, 1, 2, 3, 1, 7, 1, -1,
	-4, 1, 2, 3, 1, -1,
	9, 1, 2, 4, 2, -1,
	8, 1, 2, 4, 1, 5, 1, -1,
	8, 1, 2, 4, 1, 6, 1, -1,
	2, 1, 2, 4, 1, 7, 1, -1,
	-9, 1, 2, 4, 1, -1,
	4, 1, 2, 5, 2, -1,
	2, 1, 2, 5, 1, 6, 1, -1,
	2, 1, 2, 5, 1, 7, 1, -1,
	-4, 1, 2, 5, 1, -1,
	4, 1, 2, 6, 2, -1,
	2, 1, 2, 6, 1, 7, 1, -1,
	-4, 1, 2, 6, 1, -1,
	1, 1, 2, 7, 2, -1,
	-1, 1, 2, 7, 1, -1,
	2, 1, 1, 2, 2, 3, 1, -1,
	2, 1, 1, 2, 2, 4, 1, -1,
	8, 1, 1, 2, 2, 5, 1, -1,
	2, 1, 1, 2, 2, 6, 1, -1,
	8, 1, 1, 2, 2, 7, 1, -1,
	-9, 1, 1, 2, 2, -1,
	4, 1, 1, 2, 1, 3, 1, 4, 1, -1,
	4, 1, 1, 2, 1, 3, 1, 5, 1, -1,
	4, 1, 1, 2, 1, 3, 1, 6, 1, -1,
	4, 1, 1, 2, 1, 3, 1, 7, 1, -1,
	-10, 1, 1, 2, 1, 3, 1, -1,
	2, 1, 1, 2, 1, 4, 2, -1,
	4, 1, 1, 2, 1, 4, 1, 5, 1, -1,
	4, 1, 1, 2, 1, 4, 1, 7, 1, -1,
	-12, 1, 1, 2, 1, 4, 1, -1,
	2, 1, 1, 2, 1, 5, 2, -1,
	4, 1, 1, 2, 1, 5, 1, 6, 1, -1,
	-12, 1, 1, 2, 1, 5, 1, -1,
	4, 1, 1, 2, 1, 6, 1, 7, 1, -1,
	-10, 1, 1, 2, 1, 6, 1, -1,
	-10, 1, 1, 2, 1, 7, 1, -1,
	9, 1, 1, 2, 1, -1,
	2, 1, 1, 3, 2, 4, 1, -1,
	2, 1, 1, 3, 2, 5, 1, -1,
	2, 1, 1, 3, 2, 6, 1, -1,
	2, 1, 1, 3, 2, 7, 1, -1,
	-4, 1, 1, 3, 2, -1,
	8, 1, 1, 3, 1, 4, 2, -1,
	4, 1, 1, 3, 1, 4, 1, 5, 1, -1,
	4, 1, 1, 3, 1, 4, 1, 6, 1, -1,
	-12, 1, 1, 3, 1, 4, 1, -1,
	2, 1, 1, 3, 1, 5, 2, -1,
	-6, 1, 1, 3, 1, 5, 1, -1,
	2, 1, 1, 3, 1, 6, 2, -1,
	-6, 1, 1, 3, 1, 6, 1, -1,
	-4, 1, 1, 3, 1, 7, 1, -1,
	4, 1, 1, 3, 1, -1,
	2, 1, 1, 4, 2, 5, 1, -1,
	2, 1, 1, 4, 2, 6, 1, -1,
	8, 1, 1, 4, 2, 7, 1, -1,
	-9, 1, 1, 4, 2, -1,
	4, 1, 1, 4, 1, 5, 1, 6, 1, -1,
	4, 1, 1, 4, 1, 5, 1, 7, 1, -1,
	-10, 1, 1, 4, 1, 5, 1, -1,
	4, 1, 1, 4, 1, 6, 1, 7, 1, -1,
	-10, 1, 1, 4, 1, 6, 1, -1,
	-10, 1, 1, 4, 1, 7, 1, -1,
	9, 1, 1, 4, 1, -1,
	2, 1, 1, 5, 2, 6, 1, -1,
	2, 1, 1, 5, 2, 7, 1, -1,
	-4, 1, 1, 5, 2, -1,
	2, 1, 1, 5, 1, 6, 2, -1,
	-6, 1, 1, 5, 1, 6, 1, -1,
	-4, 1, 1, 5, 1, 7, 1, -1,
	4, 1, 1, 5, 1, -1,
	2, 1, 1, 6, 2, 7, 1, -1,
	-4, 1, 1, 6, 2, -1,
	-4, 1, 1, 6, 1, 7, 1, -1,
	4, 1, 1, 6, 1, -1,
	-1, 1, 1, 7, 2, -1,
	1, 1, 1, 7, 1, -1,
	1, 2, 2, 3, 2, -1,
	2, 2, 2, 3, 1, 5, 1, -1,
	2, 2, 2, 3, 1, 7, 1, -1,
	-1, 2, 2, 3, 1, -1,
	4, 2, 2, 4, 2, -1,
	8, 2, 2, 4, 1, 5, 1, -1,
	2, 2, 2, 4, 1, 6, 1, -1,
	2, 2, 2, 4, 1, 7, 1, -1,
	-4, 2, 2, 4, 1, -1,
	9, 2, 2, 5, 2, -1,
	2, 2, 2, 5, 1, 6, 1, -1,
	8, 2, 2, 5, 1, 7, 1, -1,
	-9, 2, 2, 5, 1, -1,
	1, 2, 2, 6, 2, -1,
	2, 2, 2, 6, 1, 7, 1, -1,
	-1, 2, 2, 6, 1, -1,
	4, 2, 2, 7, 2, -1,
	-4, 2, 2, 7, 1, -1,
	2, 2, 1, 3, 2, 4, 1, -1,
	2, 2, 1, 3, 2, 6, 1, -1,
	-1, 2, 1, 3, 2, -1,
	8, 2, 1, 3, 1, 4, 2, -1,
	12, 2, 1, 3, 1, 4, 1, 5, 1, -1,
	4, 2, 1, 3, 1, 4, 1, 6, 1, -1,
	4, 2, 1, 3, 1, 4, 1, 7, 1, -1,
	-10, 2, 1, 3, 1, 4, 1, -1,
	8, 2, 1, 3, 1, 5, 2, -1,
	4, 2, 1, 3, 1, 5, 1, 6, 1, -1,
	4, 2, 1, 3, 1, 5, 1, 7, 1, -1,
	-10, 2, 1, 3, 1, 5, 1, -1,
	2, 2, 1, 3, 1, 6, 2, -1,
	4, 2, 1, 3, 1, 6, 1, 7, 1, -1,
	-4, 2, 1, 3, 1, 6, 1, -1,
	2, 2, 1, 3, 1, 7, 2, -1,
	-4, 2, 1, 3, 1, 7, 1, -1,
	1, 2, 1, 3, 1, -1,
	2, 2, 1, 4, 2, 6, 1, -1,
	2, 2, 1, 4, 2, 7, 1, -1,
	-4, 2, 1, 4, 2, -1,
	2, 2, 1, 4, 1, 5, 2, -1,
	4, 2, 1, 4, 1, 5, 1, 6, 1, -1,
	4, 2, 1, 4, 1, 5, 1, 7, 1, -1,
	-10, 2, 1, 4, 1, 5, 1, -1,
	-4, 2, 1, 4, 1, 6, 1, -1,
	2, 2, 1, 4, 1, 7, 2, -1,
	-6, 2, 1, 4, 1, 7, 1, -1,
	4, 2, 1, 4, 1, -1,
	8, 2, 1, 5, 2, 6, 1, -1,
	2, 2, 1, 5, 2, 7, 1, -1,
	-9, 2, 1, 5, 2, -1,
	4, 2, 1, 5, 1, 6, 1, 7, 1, -1,
	-10, 2, 1, 5, 1, 6, 1, -1,
	-10, 2, 1, 5, 1, 7, 1, -1,
	9, 2, 1, 5, 1, -1,
	-1, 2, 1, 6, 2, -1,
	2, 2, 1, 6, 1, 7, 2, -1,
	-4, 2, 1, 6, 1, 7, 1, -1,
	1, 2, 1, 6, 1, -1,
	-4, 2, 1, 7, 2, -1,
	4, 2, 1, 7, 1, -1,
	9, 3, 2, 4, 2, -1,
	8, 3, 2, 4, 1, 5, 1, -1,
	8, 3, 2, 4, 1, 6, 1, -1,
	2, 3, 2, 4, 1, 7, 1, -1,
	-9, 3, 2, 4, 1, -1,
	4, 3, 2, 5, 2, -1,
	2, 3, 2, 5, 1, 6, 1, -1,
	2, 3, 2, 5, 1, 7, 1, -1,
	-4, 3, 2, 5, 1, -1,
	4, 3, 2, 6, 2, -1,
	2, 3, 2, 6, 1, 7, 1, -1,
	-4, 3, 2, 6, 1, -1,
	1, 3, 2, 7, 2, -1,
	-1, 3, 2, 7, 1, -1,
	2, 3, 1, 4, 2, 5, 1, -1,
	2, 3, 1, 4, 2, 6, 1, -1,
	8, 3, 1, 4, 2, 7, 1, -1,
	-9, 3, 1, 4, 2, -1,
	4, 3, 1, 4, 1, 5, 1, 6, 1, -1,
	4, 3, 1, 4, 1, 5, 1, 7, 1, -1,
	-10, 3, 1, 4, 1, 5, 1, -1,
	4, 3, 1, 4, 1, 6, 1, 7, 1, -1,
	-10, 3, 1, 4, 1, 6, 1, -1,
	-10, 3, 1, 4, 1, 7, 1, -1,
	9, 3, 1, 4, 1, -1,
	2, 3, 1, 5, 2, 6, 1, -1,
	2, 3, 1, 5, 2, 7, 1, -1,
	-4, 3, 1, 5, 2, -1,
	2, 3, 1, 5, 1, 6, 2, -1,
	-6, 3, 1, 5, 1, 6, 1, -1,
	-4, 3, 1, 5, 1, 7, 1, -1,
	4, 3, 1, 5, 1, -1,
	2, 3, 1, 6, 2, 7, 1, -1,
	-4, 3, 1, 6, 2, -1,
	-4, 3, 1, 6, 1, 7, 1, -1,
	4, 3, 1, 6, 1, -1,
	-1, 3, 1, 7, 2, -1,
	1, 3, 1, 7, 1, -1,
	1, 4, 2, 5, 2, -1,
	2, 4, 2, 5, 1, 7, 1, -1,
	-1, 4, 2, 5, 1, -1,
	1, 4, 2, 6, 2, -1,
	2, 4, 2, 6, 1, 7, 1, -1,
	-1, 4, 2, 6, 1, -1,
	4, 4, 2, 7, 2, -1,
	-4, 4, 2, 7, 1, -1,
	2, 4, 1, 5, 2, 6, 1, -1,
	-1, 4, 1, 5, 2, -1,
	2, 4, 1, 5, 1, 6, 2, -1,
	4, 4, 1, 5, 1, 6, 1, 7, 1, -1,
	-4, 4, 1, 5, 1, 6, 1, -1,
	2, 4, 1, 5, 1, 7, 2, -1,
	-4, 4, 1, 5, 1, 7, 1, -1,
	1, 4, 1, 5, 1, -1,
	-1, 4, 1, 6, 2, -1,
	2, 4, 1, 6, 1, 7, 2, -1,
	-4, 4, 1, 6, 1, 7, 1, -1,
	1, 4, 1, 6, 1, -1,
	-4, 4, 1, 7, 2, -1,
	4, 4, 1, 7, 1, -1,
	4, 5, 2, 6, 2, -1,
	2, 5, 2, 6, 1, 7, 1, -1,
	-4, 5, 2, 6, 1, -1,
	1, 5, 2, 7, 2, -1,
	-1, 5, 2, 7, 1, -1,
	2, 5, 1, 6, 2, 7, 1, -1,
	-4, 5, 1, 6, 2, -1,
	-4, 5, 1, 6, 1, 7, 1, -1,
	4, 5, 1, 6, 1, -1,
	-1, 5, 1, 7, 2, -1,
	1, 5, 1, 7, 1, -1,
	1, 6, 2, 7, 2, -1,
	-1, 6, 2, 7, 1, -1,
	-1, 6, 1, 7, 2, -1,
	1, 6, 1, 7, 1, -1,
	0,
};

/*
 * Return the Centroid Distance squared of a TSP(n) 3-toothed comb
 * inequality.
 */

	void
tsp_3tc_cd2 (

mpq_ptr		res,	/* IN/OUT: Result */
int		b1,	/* IN: Parameter b1 */
int		t1,	/* IN: Parameter t1 */
int		b2,	/* IN: Parameter b2 */
int		t2,	/* IN: Parameter t2 */
int		b3,	/* IN: Parameter b3 */
int		t3,	/* IN: Parameter t3 */
int		h,	/* IN: Parameter h */
int		o	/* IN: Parameter o */
)
{
int		n;
int		val [8];
mpz_ptr		N, D;
mpz_t		A, B;
mpq_t		DIST2;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (b1 <= 0);
	FATAL_ERROR_IF (t1 <= 0);
	FATAL_ERROR_IF (b2 <= 0);
	FATAL_ERROR_IF (t2 <= 0);
	FATAL_ERROR_IF (b3 <= 0);
	FATAL_ERROR_IF (t3 <= 0);
	FATAL_ERROR_IF (h  <  0);
	FATAL_ERROR_IF (o  <  0);

	n = b1 + t1 + b2 + t2 + b3 + t3 + h + o;

	val [0]	= b1;
	val [1] = t1;
	val [2] = b2;
	val [3] = t2;
	val [4] = b3;
	val [5] = t3;
	val [6] = h;
	val [7] = o;

	mpq_init (DIST2);
	mpz_init (A);
	mpz_init (B);

	N = mpq_numref (DIST2);
	D = mpq_denref (DIST2);

	eval_terms (A, A_terms, val);
	eval_terms (B, B_terms, val);

	mpz_mul (N, A, A);
	mpz_mul_ui (N, N, 2 * (n - 2));

	mpz_mul_ui (D, B, n - 1);

	mpq_canonicalize (DIST2);

	mpq_set (res, DIST2);

	mpz_clear (B);
	mpz_clear (A);
	mpq_clear (DIST2);
}

/*
 * Evaluate polynomial (given a description of its terms) at the given
 * variable values.
 */

	static
	void
eval_terms (

mpz_ptr			res,		/* IN/OUT: Result */
const signed char *	terms,		/* IN: Terms of polynomial */
const int *		val		/* IN: Array of variable values */
)
{
int		coeff, v, p;
mpz_t		SUM, TERM, TEMP;

	mpz_init (SUM);
	mpz_init (TERM);
	mpz_init (TEMP);

	for (;;) {
		coeff = *terms++;
		if (coeff EQ 0) break;

		mpz_set_si (TERM, coeff);
		for (;;) {
			v = *terms++;
			if (v < 0) break;
			p = *terms++;
			FATAL_ERROR_IF ((v >= 8) OR (p <= 0));
			mpz_ui_pow_ui (TEMP, val [v], p);
			mpz_mul (TERM, TERM, TEMP);
		}

		mpz_add (SUM, SUM, TERM);
	}

	mpz_set (res, SUM);

	mpz_clear (TEMP);
	mpz_clear (TERM);
	mpz_clear (SUM);
}

/*
 * Shut down the tsp_indicators module, freeing all memory it
 * statically holds.
 */

	void
tsp_indicators_shutdown (void)

{
	binomial_shutdown ();
}
