/***********************************************************************

	File:	zmemo.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Memoize function results of the form int* ==> mpz.

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "zmemo.h"

#include "fatal.h"
#include "gmpmisc.h"
#include "logic.h"
#include "new.h"
#include <stdlib.h>
#include <string.h>

/*
 * Local Types
 */

enum ZMType {
	MROW,		/* A row of ZMemo pointers */
	ZROW,		/* A row of mpz_ptrs */
};

struct ZMemo {
	int		size;	/* Number of elements in row */
	enum ZMType	type;	/* Type of row */
	union {
		struct ZMemo **		mrow;	/* Array of ZMemo pointers */
		mpz_ptr *		zrow;	/* Array of mpz_ptrs */
	} u;
};

/*
 * Local Functions
 */

/*
 * Create an empty ZMemo "mrow" object.
 */

	struct ZMemo *
zmemo_create_mrow (void)

{
struct ZMemo *		mp;

	mp = NEW (struct ZMemo);

	memset (mp, 0, sizeof (*mp));

	mp -> size	= 0;
	mp -> type	= MROW;
	mp -> u.mrow	= NULL;

	return (mp);
}

/*
 * Create an empty ZMemo "zrow" object.
 */

	struct ZMemo *
zmemo_create_zrow (void)

{
struct ZMemo *		mp;

	mp = NEW (struct ZMemo);

	memset (mp, 0, sizeof (*mp));

	mp -> size	= 0;
	mp -> type	= ZROW;
	mp -> u.zrow	= NULL;

	return (mp);
}

/*
 * Index the given mrow object, returning its n-th ZMemo object (which
 * can be NULL if it does not yet exist).  The given mrow is expanded
 * (if necessary) so that it has an element at index n.
 */

	struct ZMemo *
zmemo_get_mrow (

struct ZMemo *	mp,		/* IN: MROW type of memo to index */
int		n		/* IN: Element of row to get */
)
{
int		size, new_size;
struct ZMemo **	mrow;
struct ZMemo **	new_mrow;
struct ZMemo *	mp2;

	FATAL_ERROR_IF (mp EQ NULL);
	FATAL_ERROR_IF (mp -> type NE MROW);
	FATAL_ERROR_IF (n < 0);

	size	= mp -> size;
	mrow	= mp -> u.mrow;

	if (n >= size) {
		/* Expand this row */
		new_size = 2 * size;
		if (n >= new_size) {
			new_size = n + 1;
		}
		new_mrow = NEWA (new_size, struct ZMemo *);
		memcpy (new_mrow,
			mrow,
			size * sizeof (new_mrow [0]));
		memset (&new_mrow [size],
			0,
			(new_size - size) * sizeof (new_mrow [0]));
		if (mrow NE NULL) {
			free (mrow);
		}
		size	= new_size;
		mrow	= new_mrow;
		mp -> size	= size;
		mp -> u.mrow	= mrow;

		FATAL_ERROR_IF (n >= size);
	}

	mp2 = mrow [n];

	return (mp2);
}

/*
 * Set the n-th element of the given mrow to the given value (which must
 * not yet have a value).
 */

	void
zmemo_set_mrow (

struct ZMemo *	mp,		/* IN: ZROW type of memo to index */
int		n,		/* IN: Element of row to set */
struct ZMemo *	mp2		/* IN: Value to set */
)
{
int		size;
struct ZMemo **	mrow;

	FATAL_ERROR_IF (mp EQ NULL);
	FATAL_ERROR_IF (mp -> type NE MROW);
	FATAL_ERROR_IF (n < 0);

	size	= mp -> size;
	mrow	= mp -> u.mrow;

	/* The given element must already exist -- the caller	*/
	/* must have first called zmemo_get_mrow (mp, n) to see	*/
	/* if the value already exists or not.			*/
	FATAL_ERROR_IF (n >= size);

	FATAL_ERROR_IF (mrow [n] NE NULL);

	mrow [n] = mp2;
}

/*
 * Index the given zrow object, returning its n-th mpz_ptr (which
 * can be NULL if it does not yet exist).  The given zrow is expanded
 * (if necessary) so that it has an element at index n.
 */

	mpz_ptr
zmemo_get_zrow (

struct ZMemo *	mp,		/* IN: ZROW type of memo to index */
int		n		/* IN: Element of row to get */
)
{
int		size, new_size;
mpz_ptr *	zrow;
mpz_ptr *	new_zrow;
mpz_ptr		zp;

	FATAL_ERROR_IF (mp EQ NULL);
	FATAL_ERROR_IF (mp -> type NE ZROW);
	FATAL_ERROR_IF (n < 0);

	size	= mp -> size;
	zrow	= mp -> u.zrow;

	if (n >= size) {
		/* Expand this row */
		new_size = 2 * size;
		if (n >= new_size) {
			new_size = n + 1;
		}
		new_zrow = NEWA (new_size, mpz_ptr);
		memcpy (new_zrow,
			zrow,
			size * sizeof (new_zrow [0]));
		memset (&new_zrow [size],
			0,
			(new_size - size) * sizeof (new_zrow [0]));
		if (zrow NE NULL) {
			free (zrow);
		}
		size	= new_size;
		zrow	= new_zrow;
		mp -> size	= size;
		mp -> u.zrow	= zrow;

		FATAL_ERROR_IF (n >= size);
	}

	zp = zrow [n];

	return (zp);
}

/*
 * Set the n-th element of the given zrow to the given value (which must
 * not yet have a value).
 */

	void
zmemo_set_zrow (

struct ZMemo *	mp,		/* IN: ZROW type of memo to index */
int		n,		/* IN: Element of row to set */
mpz_srcptr	z		/* IN: Value to set */
)
{
int		size;
mpz_ptr *	zrow;
mpz_ptr		zp;

	FATAL_ERROR_IF (mp EQ NULL);
	FATAL_ERROR_IF (mp -> type NE ZROW);
	FATAL_ERROR_IF (n < 0);

	size	= mp -> size;
	zrow	= mp -> u.zrow;

	/* The given element must already exist -- the caller	*/
	/* must have first called zmemo_get_zrow (mp, n) to see	*/
	/* if the value already exists or not.			*/
	FATAL_ERROR_IF (n >= size);

	FATAL_ERROR_IF (zrow [n] NE NULL);

	zp = new_Z_vector (1);
	mpz_set (zp, z);
	zrow [n] = zp;
}

/*
 * Recursively free the given ZMemo object.
 */

	void
zmemo_free (

struct ZMemo *	mp		/* IN: ZMemo object to free */
)
{
int		i, size;
struct ZMemo **	mrow;
struct ZMemo *	mp2;
mpz_ptr *	zrow;
mpz_ptr		zp;

	if (mp EQ NULL) {
		return;
	}

	size	= mp -> size;

	switch (mp -> type) {
	case MROW:
		mrow	= mp -> u.mrow;
		for (i = 0; i < size; i++) {
			mp2 = mrow [i];
			if (mp2 NE NULL) {
				zmemo_free (mp2);
				mrow [i] = NULL;
			}
		}
		free (mrow);
		break;

	case ZROW:
		zrow	= mp -> u.zrow;
		for (i = 0; i < size; i++) {
			zp = zrow [i];
			if (zp NE NULL) {
				free_Z_vector (zp, 1);
				zrow [i] = NULL;
			}
		}
		free (zrow);
		break;

	default:
		FATAL_ERROR;
		break;
	}

	memset (mp, 0, sizeof (*mp));

	free (mp);
}
