/**********************************************************************

	$Id: plot_sthgp_sec_cd2_dx.c,v 1.1 2025/01/26 02:31:52 warme Exp $

	File:	plot_sthgp_sec_cd2_dx.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Plot DX_e values and sums of (DX_e)^2 values
	for the STHGP(n) subtour Centroid Distance.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

#include "fatal.h"
#include <gmp.h>
#include "logic.h"
#include <math.h>
#include "s3itypes.h"
#include <stdlib.h>
#include <stdio.h>
#include "sthgp_indicators.h"
#include "sthgp_sec_cd2_dx.h"
#include <string.h>


/*
 * Global Routines
 */

int		main (int, char **);

char *		me;

/*
 * Local Types
 */

enum Mode {
	MODE_PLOT_DX_0,
	MODE_PLOT_DX_1,
	MODE_PLOT_DX_S,
	MODE_PLOT_SUMS,
	MODE_PLOT_TAU,
};


/*
 * Local Routines
 */

static void		check_nverts (void);
static void		check_nverts_and_edge_size (void);
static void		decode_args (int argc, char ** argv);
static void		plot_dx_0 (void);
static void		plot_dx_1 (void);
static void		plot_dx_S (void);
static void		plot_sums (void);
static void		plot_tau (void);
static void		usage (void);

/*
 * Local Data
 */

static int		edge_size;
static enum Mode	mode;
static int		nverts;

/*
 * The main routine.
 *
 * Given the number of vertices N, it computes the centroid distance
 * for the K-vertex subtour for each 2 <= K <= N-1.
 */

	int
main (

int		argc,
char **		argv
)
{
	me = argv [0];

	decode_args (argc, argv);

	switch (mode) {
	case MODE_PLOT_DX_0:
		plot_dx_0 ();
		break;

	case MODE_PLOT_DX_1:
		plot_dx_1 ();
		break;

	case MODE_PLOT_DX_S:
		plot_dx_S ();
		break;

	case MODE_PLOT_SUMS:
		plot_sums ();
		break;

	case MODE_PLOT_TAU:
		plot_tau ();
		break;

	default:
		FATAL_ERROR;
	}

	sthgp_indicators_shutdown ();

	exit (0);
}

/*
 * Decode the given command line arguments.
 */

	static
	void
decode_args (

int		argc,
char **		argv
)
{
char *		ap;

	if (argc < 2) {
		usage ();
	}

	nverts		= -1;
	edge_size	= -1;

	ap = argv [1];

	if (strcmp (ap, "--dx0") EQ 0) {
		if (argc < 4) {
			usage ();
		}
		mode		= MODE_PLOT_DX_0;
		nverts		= atoi (argv [2]);
		edge_size	= atoi (argv [3]);
		check_nverts_and_edge_size ();
	}
	else if (strcmp (ap, "--dx1") EQ 0) {
		if (argc < 4) {
			usage ();
		}
		mode		= MODE_PLOT_DX_1;
		nverts		= atoi (argv [2]);
		edge_size	= atoi (argv [3]);
		check_nverts_and_edge_size ();
	}
	else if (strcmp (ap, "--dxS") EQ 0) {
		if (argc < 4) {
			usage ();
		}
		mode		= MODE_PLOT_DX_S;
		nverts		= atoi (argv [2]);
		edge_size	= atoi (argv [3]);
		check_nverts_and_edge_size ();
	}
	else if (strcmp (ap, "--sums") EQ 0) {
		if (argc < 3) {
			usage ();
		}
		mode		= MODE_PLOT_SUMS;
		nverts		= atoi (argv [2]);
		check_nverts ();
	}
	else if (strcmp (ap, "--tau") EQ 0) {
		if (argc < 3) {
			usage ();
		}
		mode		= MODE_PLOT_TAU;
		nverts		= atoi (argv [2]);
		check_nverts ();
	}
	else {
		usage ();
	}
}

/*
 * Display command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usages:\n"
		 "\t%s --dx0 NVerts EdgeSize\n"
		 "\t%s --dx1 NVerts EdgeSize\n"
		 "\t%s --dxS NVerts EdgeSize\n"
		 "\t%s --sums NVerts\n"
		 "\t%s --tau NVerts\n",
		 me, me, me, me, me);
	exit (1);
}

/*
 * Check the given number of vertices (which must be >= 3).
 */

	static
	void
check_nverts (void)

{
	if (nverts < 3) {
		fprintf (stderr, "%s: must have at least 3 vertices!\n", me);
		exit (1);
	}
}

/*
 * Check the given number of vertices and the given edge size.
 */

	static
	void
check_nverts_and_edge_size (void)

{
	check_nverts ();

	if ((edge_size < 2) OR (edge_size > nverts)) {
		fprintf (stderr,
			 "%s: Invalid edge size %d."
			 "  (Must be in [%d,%d].)\n",
			 me, edge_size, 2, nverts - 1);
		exit (1);
	}
}

/*
 * Plot DX(n,k,0,q) for given value of n.
 */

	static
	void
plot_dx_0 (void)

{
int		k, n, q;
mpq_t		VAL;
mpz_t		B, TMP1;
	
	n = nverts;

	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF ((edge_size < 2) OR (edge_size > n));

	mpq_init (VAL);

	mpz_init (B);
	mpz_init (TMP1);

	for (k = 2; k < n; k++) {

		q = edge_size;

		if (k + q > n) {
			/* Must stop here. */
			break;
		}

		sthgp_subtour_cd2_dx_TAU (VAL, n, k);
		sthgp_subtour_cd2_Beta (B, n, k);

		mpz_mul_ui (TMP1, B, q - 1);

		mpz_mul (mpq_numref (VAL), mpq_numref (VAL), TMP1);
		mpq_canonicalize (VAL);

		printf ("%d\t%.17g\n",
			k,
			mpq_get_d (VAL));
	}

	mpz_clear (TMP1);
	mpz_clear (B);

	mpq_clear (VAL);
}

/*
 * Plot DX(n,k,1,q) for given value of n.
 */

	static
	void
plot_dx_1 (void)

{
int		k, n, q;
mpq_t		VAL;
mpz_t		B, TMP1;
	
	n = nverts;

	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF ((edge_size < 2) OR (edge_size > n));

	mpq_init (VAL);

	mpz_init (B);
	mpz_init (TMP1);

	for (k = 2; k < n; k++) {

		q = edge_size;

		if (k + q > n + 1) {
			/* Must stop here. */
			break;
		}

		sthgp_subtour_cd2_dx_TAU (VAL, n, k);
		sthgp_subtour_cd2_Beta (B, n, k);

		mpz_mul_ui (TMP1, B, q);

		mpz_mul (mpq_numref (VAL), mpq_numref (VAL), TMP1);
		mpq_canonicalize (VAL);

		printf ("%d\t%.17g\n",
			k,
			mpq_get_d (VAL));
	}

	mpz_clear (TMP1);
	mpz_clear (B);

	mpq_clear (VAL);
}

/*
 * Plot DX(n,k,p,q) {with p >= 2) for given value of n.
 */

	static
	void
plot_dx_S (void)

{
int		k, n, p, q;
mpq_t		VAL;
mpz_t		A, B, TMP1, TMP2;
	
	n = nverts;

	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF ((edge_size < 2) OR (edge_size > n));

	mpq_init (VAL);

	mpz_init (A);
	mpz_init (B);
	mpz_init (TMP1);
	mpz_init (TMP2);

	sthgp_subtour_cd2_Alpha (A, n);

	for (k = 2; k < n; k++) {

		p = k;
		if (p > edge_size) {
			p = edge_size;
		}
		FATAL_ERROR_IF (p < 2);

		q = edge_size - p;

		sthgp_subtour_cd2_dx_TAU (VAL, n, k);
		sthgp_subtour_cd2_Beta (B, n, k);

		mpz_mul_ui (TMP1, A, p - 1);
		mpz_mul_ui (TMP2, B, (q > 0) ? (q - 1) : 0);

		mpz_sub (TMP1, TMP1, TMP2);
		mpz_abs (TMP1, TMP1);

		mpz_mul (mpq_numref (VAL), mpq_numref (VAL), TMP1);
		mpq_canonicalize (VAL);

		printf ("%d\t%.17g\n",
			k,
			mpq_get_d (VAL));
	}

	mpz_clear (TMP2);
	mpz_clear (TMP1);
	mpz_clear (B);
	mpz_clear (A);

	mpq_clear (VAL);
}

/*
 * Plot TAU(n,k) for given value of n.
 */

	static
	void
plot_tau (void)

{
int		k, n;
mpq_t		TMP1;
	
	n = nverts;

	FATAL_ERROR_IF (n < 3);

	mpq_init (TMP1);

	for (k = 2; k < n; k++) {

		sthgp_subtour_cd2_dx_TAU (TMP1, n, k);

		printf ("%d\t%.17g\n",
			k,
			mpq_get_d (TMP1));
	}

	mpq_clear (TMP1);
}

/*
 * Plot the sums of DX^2 values.
 */

	static
	void
plot_sums (void)

{
int		k, n;
mpq_t		TMP1, TMP2, TMP3;
	
	n = nverts;

	FATAL_ERROR_IF (n < 3);

	mpq_init (TMP1);
	mpq_init (TMP2);
	mpq_init (TMP3);

	for (k = 2; k < n; k++) {

		sthgp_subtour_cd2_dx_Delta_x_1 (TMP1, n, k);
		sthgp_subtour_cd2_dx_Delta_x_2 (TMP2, n, k);

		mpq_add (TMP3, TMP1, TMP2);

		printf ("%d\t%.17g\t%.17g\t%.17g\n",
			k,
			mpq_get_d (TMP1),
			mpq_get_d (TMP2),
			mpq_get_d (TMP3));
	}

	mpq_clear (TMP3);
	mpq_clear (TMP2);
	mpq_clear (TMP1);
}
