/**********************************************************************

	$Id: sthgp_sec_compare.c,v 1.1 2025/01/26 02:31:53 warme Exp $

	File:	sthgp_sec_compare.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Generate Postscript image indicating when the CD and EPR
	indicators disagree about the relative strength of STHGP(n)
	subtours k1 and k2.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

/*
 * Program to graphically compare how closely the centroid distance
 * and extreme point ratio indicators rank the relative strength of all
 * the subtour inequalities of STHGP(n).
 *
 * Let n >= 3, and let K = {k \in Z : 2 <= k < n}.
 * For each k \in K, let S \subset V such that |S| = k.
 * Let CDIST2[k] be the centroid distance squared for subtour S,
 * and let NXP[k] be the number of extreme points incident to subtour S.
 *
 * For (i,j) \in K^2, define:
 *	d2(i,j)		= '<<' if CDIST2[i] > CDIST2[j],
 *			  '==' if CDIST2[i] = CDIST2[j],
 *			  '>>' if CDIST2[i] < CDIST2[j], and
 *
 *	epr(i,j)	= '<<' if NXP[i] < NXP[j],
 *			  '==' if NXP[i] = NXP[j],
 *			  '>>' if NXP[i] < NXP[j].
 *
 * Thus d2(k1,k2) indicates whether subtour k1 is weaker, equal or
 * stronger than subtour k2 according to the CD indicator, and
 * epr(k1,k2) indicates whether subtour k1 is weaker, equal or stronger
 * than subtour k2 according to the extreme point ratio indicator.
 *
 * We generate a postscript plot giving a black box at coordinates
 * (k1,k2) for all (k1,k2) \in K^2 such that CDIST2(k1,k2) NE XP(k1,k2),
 * i.e., pairs of subtours that the two indicators rank differently.
 *
 * We demonstrate that the discrepancies are relatively sparse and
 * concern the weakest subtours.
 */

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include <math.h>
#include "new.h"
#include "s3itypes.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include "sthgp_indicators.h"
#include <string.h>

int		main (int argc, char ** argv);

char *		me;

/*
 * Local Functions
 */

static int		decode_args (int argc, char ** argv);
static void		usage (void);

/*
 * Local Data
 */

static char	epsheader [];

/*
 * The main program.
 */

	int
main (

int		argc,
char **		argv
)
{
int		i, j, k, n, cmp;
mpz_ptr		NXP;
mpq_ptr		CDIST2;
char *		epr;
char *		d2;

	me = argv [0];

	n = decode_args (argc, argv);
	
	/* Emit Encapsulated Postscript Plot. */
	printf ("%s", epsheader);

	/* Compute the number of extreme points incident to each subtour. */
	NXP = new_Z_vector (n + 1);
	sthgp_all_subtour_ep_count (NXP, n);

#if 0
	printf ("\n%% Incident Extreme Points:\n");
	for (k = 2; k < n; k++) {
		gmp_printf ("%% %d:\t%Zd\n", k, &NXP [k]);
	}
#endif

	/* Compute the centroid distances squared. */
	CDIST2 = new_Q_vector (n + 1);
	sthgp_all_subtour_cd2 (CDIST2, n);

#if 0
	printf ("\n%% Centroid Distances Squared / Distances:\n");
	for (k = 2; k < n; k++) {
		double x = mpq_get_d (&CDIST2 [k]);
		printf ("%% %d:\t%.17g\t%.17g\n", k, x, sqrt (x));
	}
#endif

	/* Find weakest extreme point ratio. */
	k = 2;
	for (i = 3; i < n; i++) {
		if (mpz_cmp (&NXP [i], &NXP [k]) < 0) {
			k = i;
		}
	}
	printf ("%% N = %d\n", n);
	printf ("%% Weakest extreme point ratio at K = %d\n", k);

	/* Find weakest centroid distance. */
	k = 2;
	for (i = 3; i < n; i++) {
		if (mpq_cmp (&CDIST2 [i], &CDIST2 [k]) > 0) {
			k = i;
		}
	}
	printf ("%% Weakest centroid distance at K = %d\n", k);

	/* Compute epr[i,j] and d2[i,j]. */
	epr = NEWA (n * n, char);
	d2  = NEWA (n * n, char);
	memset (epr, 0, n * n);
	memset (d2,  0, n * n);

	for (i = 2; i < n; i++) {
		for (j = 2; j < n; j++) {
			cmp = mpz_cmp (&NXP [i], &NXP [j]);
			if (cmp < 0) cmp = -1;
			else if (cmp > 0) cmp = 1;
			epr [i * n + j] = "<=>" [1 + cmp];

			cmp = mpq_cmp (&CDIST2 [i], &CDIST2 [j]);
			if (cmp < 0) cmp = -1;
			else if (cmp > 0) cmp = 1;
			d2 [i * n + j] = ">=<" [1 + cmp];
		}
	}

	printf ("%%%%BeginSetup\n");
	printf ("\t%d SetAxes\n", n);
	printf ("%%%%EndSetup\n"
		"\n"
		"%%%%Page: 1 1\n"
		"BeginPlot\n");
	for (i = 2; i < n; i++) {
		for (j = 2; j < n; j++) {
			if (epr [i * n + j] NE d2 [i * n + j]) {
				printf ("\t%d %d Box\n", i, j);
			}
		}
	}
	printf ("EndPlot\n"
		"%%%%EOF\n");

	free (d2);
	free (epr);

	free_Q_vector (CDIST2, n + 1);
	free_Z_vector (NXP, n + 1);

	sthgp_indicators_shutdown ();

	exit (0);
}

/*
 * Decode command line arguments.
 */

	static
	int
decode_args (

int		argc,
char **		argv
)
{
int		n;

	n	= 10;
	if (argc >= 2) {
		n = atoi (argv [1]);
	}
	if (n < 3) {
		fprintf (stdout, "%s: Must have N >= 3.\n", me);
		usage ();
	}

	return (n);
}

/*
 * Print command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage: %s [ num_verts ]\n",
		 me);

	exit (1);
}

/*
 * Define the Encapsulated Postscript header stuff.
 */

static char epsheader [] =
	"%!PS-Adobe-3.0 EPSF-3.0\n"
	"%%Creator: sthgp_compare Version 1.0\n"
	"%%Copyright: Copyright (c) 1998, 2021 by David M. Warme."
	" This work is licensed under a Creative Commons Attribution 4.0"
	" International License.\n"
	"%%LanguageLevel: 1\n"
	"%%BoundingBox: 53 143 559 649\n"
	"%%HiResBoundingBox: 53.500006 143.500003 558.499975 648.499972\n"
	"%%DocumentNeededResources: font Times-Roman\n"
	"%%DocumentSuppliedResources: procset STHGPCompareProcs 1.0 0\n"
	"%%EndComments\n"
	"%%BeginProlog\n"
	"%%BeginProcSet: STHGPCompareProcs 1.0 0\n"
	"/inch { 72 mul } def\n"
	"/PaperWidth	8.5 inch	def\n"
	"/PaperHeight	11 inch		def\n"
	"/BigBoxSize	7 inch		def\n"
	"/BoxHeight	BigBoxSize	def\n"
	"/BoxWidth	BigBoxSize	def\n"
	"/MinX		0		def\n"
	"/MaxX		1		def\n"
	"/MinY		0		def\n"
	"/MaxY		1		def\n"
	"\n"
	"% Define the clipping region to be the inside edge of the box.\n"
	"\n"
	"/ClipHeight	BigBoxSize currentlinewidth sub def\n"
	"/ClipWidth	ClipHeight	def\n"
	"/PlotHeight	ClipHeight	def\n"
	"/PlotWidth	ClipWidth	def\n"
	"/XlateX { MinX sub MaxX MinX sub div PlotWidth mul } def\n"
	"/XlateY { MinY sub MaxY MinY sub div PlotHeight mul } def\n"
	"/XlateXY { exch XlateX exch XlateY } def\n"
	"/Dup2 { dup 3 -1 roll dup 4 1 roll exch } def\n"
	"\n"
	"% Define the plotting region for given n.\n"
	"% We want coodinates 0 ... n+2 so that 2 ... n are located"
	" symmetrically\n"
	"% inside the box.\n"
	"/SetAxes {	% n\n"
	"	dup 2 add /MaxX exch def\n"
	"	dup 2 add /MaxY exch def\n"
	"	/MinX 0 def\n"
	"	/MinY 0 def\n"
	"} def\n"
	"/BeginPlot {\n"
	"	gsave\n"
	"	PaperWidth BoxWidth sub 2 div\n"
	"	 PaperHeight BoxHeight sub 2 div\n"
	"	  translate\n"
	"	newpath\n"
	"	0 0 moveto\n"
	"	BoxWidth 0 lineto\n"
	"	BoxWidth BoxHeight lineto\n"
	"	0 BoxHeight lineto\n"
	"	closepath\n"
	"	stroke\n"
	"	gsave\n"
	"	% establish a clipping region at the inside edge of the box\n"
	"	currentlinewidth 2 div dup moveto\n"
	"	ClipWidth 0 rlineto\n"
	"	0 ClipHeight rlineto\n"
	"	ClipWidth neg 0 rlineto\n"
	"	closepath\n"
	"	clip\n"
	"	% establish the origin for the plotting region\n"
	"	currentlinewidth 2 div\n"
	"	 dup\n"
	"	  translate\n"
	"} def\n"
	"/Box {	% k1 k2\n"
	"	10 dict begin\n"
	"	/k2 exch def\n"
	"	/k1 exch def\n"
	"	k1 k2 XlateXY\n"
	"	/y0 exch def\n"
	"	/x0 exch def\n"
	"	k1 1 add k2 1 add XlateXY\n"
	"	/y1 exch def\n"
	"	/x1 exch def\n"
	"	newpath\n"
	"	x0 y0 moveto\n"
	"	x1 y0 lineto\n"
	"	x1 y1 lineto\n"
	"	x0 y1 lineto\n"
	"	closepath\n"
	"	fill\n"
	"	end\n"
	"} def\n"
	"/EndPlot {\n"
	"	showpage\n"
	"} def\n"
	"%%EndProcSet\n"
	"%%EndProlog\n"
	"\n";
