/***********************************************************************

	$Id: qpdist.c,v 1.1 2025/01/26 02:32:02 warme Exp $

	File:	qpdist.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Use quadratic programming to measure centroid distances
	from first principles.

************************************************************************

	Modification Log:

	a-1:	08/18/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "bitmaskmacros.h"
#include "fatal.h"
#include <limits.h>
#include "logic.h"
#include <math.h>
#include "memory.h"
#include "polytope.h"
#include "qpdist2.h"
#include "s3itypes.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cplex.h"


/*
 * Global Routines
 */

int		main (int, char **);

char *		me;

/*
 * Local Routines
 */

static void		decode_params (int, char **);
static void		usage (void);


/*
 * Local Variables
 */

enum CD_Type		cd_type;
static char *		facet_type;
static char *		polytope_type;
static int		verbose;

/*
 * The main routine for the hypertree generator.
 */

	int
main (

int		argc,
char **		argv
)

{
struct Polytope *	poly;
struct Constraint *	cp;
double			dist2;

	setbuf (stdout, NULL);

	decode_params (argc, argv);

	if (verbose) {
		printf ("    === %s centroid distance ===\n",
			(cd_type EQ QPDIST2_NORMAL_CD2) ? "Strong" : "Weak");
	}

	poly = make_polytope_by_name (polytope_type);
	if (poly EQ NULL) {
		/* Error message already printed. */
		exit (1);
	}

	/* Get the constraint corresponding to the specified facet. */
	cp = poly -> get_constraint (poly, facet_type);
	if (cp EQ NULL) {
		fprintf (stderr, "%s: Constraint `%s' unknown.\n",
			 me, facet_type);
		exit (1);
	}

	dist2 = cd2_via_qp (cp, poly, cd_type);

	printf ("Minimum distance [%s,%s,%s] = %.17g\n",
		polytope_type,
		facet_type,
		((cd_type EQ QPDIST2_NORMAL_CD2) ? "chull" : "ahull"),
		sqrt (dist2));

	free_constraint_list (cp);

	poly -> free (poly);
	poly = NULL;

	exit (0);
}

/*
 * Decode the command line arguments.
 */

	static
	void
decode_params (

int		argc,
char **		argv
)
{
char *		ap;
char		c;

	--argc;
	me = *argv++;

	cd_type		= QPDIST2_WEAK_CD2;
	polytope_type	= NULL;
	facet_type	= NULL;
	verbose		= 0;

	while (argc > 0) {
		--argc;
		ap = *argv++;
		if (*ap NE '-') {
			usage ();
		}
		else {
			++ap;
			while ((c = *ap++) NE '\0') {
				switch (c) {

				case 'c':
					/* Use convex hull. */
					cd_type = QPDIST2_NORMAL_CD2;
					break;

				case 'f':
					if (*ap EQ '\0') {
						if (argc <= 0) usage ();
						--argc;
						ap = *argv++;
					}
					facet_type = ap;
					ap = "";
					break;

				case 'p':
					if (*ap EQ '\0') {
						if (argc <= 0) usage ();
						--argc;
						ap = *argv++;
					}
					polytope_type = ap;
					ap = "";
					break;

				case 'v':
					++verbose;
					break;

				default:
					usage ();
					break;
				}
			}
		}
	}

	if ((polytope_type EQ NULL) OR (facet_type EQ NULL)) {
		usage ();
	}
}

/*
 * This routine prints out the proper usage and exits.
 */

static char *	arg_doc [] = {
	"",
	"\t-p P\tSpecifies the polytope P.  Examples include:",
	"\t\t-p 'TSP(8)' and -p 'STHGP(7)'.",
	"\t-f F\tSpecifies facet type F.  Examples for TSP(n) include:",
	"\t\t-f 'sec-4', -f 'nonneg' and -f '3comb-A-B-C-D-E-F-G'.",
	"\t\tExamples for STHGP(n) include:",
	"\t\t-f 'sec-3' and -f 'nonneg-4'.",
	"\t-c\tMeasure distance to convex hull of facet.  (Default",
	"\t\tis to measure distance to affine hull of facet.)",
	"\t-v\tVerbose.  Print more info.  Can be used more than once",
	"\t\tto increase amount printed.\n",
	"",
	NULL
};

	static
	void
usage (void)

{
char **		pp;
char *		p;

	(void) fprintf (stderr,
			"\nUsage: %s [-cv] -p polytope -f facet\n",
			me);
	pp = &arg_doc [0];
	while ((p = *pp++) NE NULL) {
		(void) fprintf (stderr, "%s\n", p);
	}
	exit (1);
}
