/***********************************************************************

	$Id: validate_stgp.c,v 1.1 2025/01/26 02:32:02 warme Exp $

	File:	validate_stgp.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Validate all indicators for STGP(n).

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "config.h"

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "math.h"
#include "new.h"
#include "polytope.h"
#include "qpdist2.h"
#include <stdio.h>
#include <stdlib.h>
#include "stgp_indicators.h"
#include "stgp_polytope.h"
#include <string.h>

/*
 * Local Functions
 */

static void	validate_stgp_indicators (int n);

char *		me;

/*
 * Main program to validate all indicators for STGP(n).
 */

	int
main (

int		argc,
char **		argv
)
{
int		n;

	me = argv [0];

	printf ("Validating strength indicators for STGP(n).\n");

#ifndef CPLEX
	printf ("\n"
		"Warning: No CPLEX"
		" -- skipping validation of Centroid Distances.\n");
#endif

	for (n = 3; n <= 9; n++) {
		validate_stgp_indicators (n);
	}

	exit (0);
}

/*
 * Validate all indicators for STGP(n).
 */

	static
	void
validate_stgp_indicators (

int		n
)
{
int			j, k, e, dim, nmasks, nf;
const bitmap_t *	mask;
struct Polytope *	poly;
struct IVecSet *	xpts;
struct Constraint *	cp;
int **			edges;
int *			list;
int *			xpsum;
mpz_t			COUNT;
mpq_t			DIST2;
char			cname [32];

	printf ("\n"
		"n = %d:\n", n);

	mpz_init (COUNT);
	mpq_init (DIST2);

	/* Make the polytope STGP(n). */
	poly	= make_STGP_polytope (n);

	dim	= poly -> dim;

	/* Get list of all edges. */
	edges	= get_STGP_edge_list (poly);

	/* Get list of all extreme points. */
	xpts	= poly -> get_extreme_points (poly);

	FATAL_ERROR_IF (xpts -> dim NE dim);

	/* Get number of extreme points via closed-form formula. */
	stgp_num_extreme_points (COUNT, n);

	gmp_printf ("  Extreme Points:\t%Zd", COUNT);
	if (mpz_cmp_ui (COUNT, xpts -> nvecs) NE 0) {
		printf ("\t*** Enumeration gives %d ***", xpts -> nvecs);
	}
	printf ("\n");

	printf ("  Centroid Edge Counts:\n");
	/* Let e \in E.  Get the number of spanning trees containint	*/
	/* edge e using closed-form formula.				*/

	stgp_num_edge_extreme_points (COUNT, n);
	gmp_printf ("\t%Zd\n", COUNT);

	/* Add up all extreme points. */
	xpsum = NEWA (dim, int);
	memset (xpsum, 0, dim * sizeof (xpsum [0]));
	nmasks	= xpts -> nmasks;
	mask	= &(xpts -> array [0]);
	for (j = 0; j < xpts -> nvecs; j++, mask += nmasks) {
		for (e = 0; e < dim; e++) {
			if (BITON (mask, e)) {
				++(xpsum [e]);
			}
		}
	}

	/* Verify that each edge is contained in the correct number of	*/
	/* spanning trees.						*/
	for (e = 0; e < dim; e++) {
		k = edges [e + 1] - edges [e];
		if (mpz_cmp_ui (COUNT, xpsum [e]) NE 0) {
			printf ("\t*** Spanning tree count for"
				" edge %d = %d ***\n",
				e, xpsum [e]);
		}
	}
	free (xpsum);

	printf ("  Non-Negativity Extreme Point Counts:\n");
	stgp_nonneg_ep_count (COUNT, n);

	cp = poly -> get_constraint (poly, "nonneg");

	nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);
	free (list);
	free_constraint_list (cp);

	gmp_printf ("\t%Zd", COUNT);
	if (mpz_cmp_ui (COUNT, nf) NE 0) {
		printf ("\t*** Enumeration gives %d ***", nf);
	}
	printf ("\n");

	printf ("  Subtour Extreme Point Counts:\n");
	for (k = 2; k < n; k++) {
		stgp_subtour_ep_count (COUNT, n, k);

		sprintf (cname, "subtour-%d", k);
		cp = poly -> get_constraint (poly, cname);

		nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);
		free (list);
		free_constraint_list (cp);

		gmp_printf ("    k = %d:\t%Zd", k, COUNT);
		if (mpz_cmp_ui (COUNT, nf) NE 0) {
			printf ("\t*** Enumeration gives %d ***", nf);
		}
		printf ("\n");
	}

	/* We need CPLEX in order to compute Centroid Distances	*/
	/* (squared) via quadratic programming.			*/

#if CPLEX
	{ double	relerr, max_relerr, dist2, xdist2;
	  printf ("  Non-Negativity Centroid Distance Squared:\n");
	  max_relerr = 0.0;
	  stgp_nonneg_cd2 (DIST2, n);

	  cp = poly -> get_constraint (poly, "nonneg");

	  dist2 = cd2_via_qp (cp, poly, QPDIST2_WEAK_CD2);

	  xdist2 = mpq_get_d (DIST2);

	  relerr = fabs (dist2 - xdist2) / fabs (xdist2);

	  printf ("\t%.17g\t%.17g\t%.17g\n",
		  xdist2, dist2, relerr);
	  if (relerr > max_relerr) {
		max_relerr = relerr;
	  }
	  printf ("\tMax Relative Error:\t%.17g\n", max_relerr);

	  printf ("  Subtour Centroid Distance Squared:\n");
	  max_relerr = 0.0;
	  for (k = 2; k < n; k++) {
		stgp_subtour_cd2 (DIST2, n, k);

		sprintf (cname, "subtour-%d", k);
		cp = poly -> get_constraint (poly, cname);

		dist2 = cd2_via_qp (cp, poly, QPDIST2_WEAK_CD2);

		xdist2 = mpq_get_d (DIST2);

		relerr = fabs (dist2 - xdist2) / fabs (xdist2);

		printf ("    k = %d:\t%.17g\t%.17g\t%.17g\n",
			k, xdist2, dist2, relerr);
		if (relerr > max_relerr) {
			max_relerr = relerr;
		}
	  }
	  printf ("\tMax Relative Error:\t%.17g\n", max_relerr);
	}
#endif

	/* Free up the polytope. */
	poly -> free (poly);

	mpq_clear (DIST2);
	mpz_clear (COUNT);
}
