/***********************************************************************

	File:	validate_sthgp_angles.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Code to test / verify the math for computing the angle
	between two STHGP subtour inequalities.

************************************************************************

	Modification Log:

	a-1:	09/10/2023	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "sthgp_polytope.h"

#include "bitmaskmacros.h"
#include "fatal.h"
#include <limits.h>
#include "logic.h"
#include <math.h>
#include "new.h"
#include "polytope.h"
#include "s3itypes.h"
#include "scanutils.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include "sthgp_polytope.h"
#include <string.h>

/*
 * Global Data
 */

char *		me;

/*
 * An object to hold a complete hypergraph H=(V,E) having given number
 * of vertices.  This application is only interested in V and E --
 * there is no need to tie into a general polytope framework or
 * recursively enumerate spanning tree incidence vectors.
 */

struct Hypergraph {
	int		nverts;		/* Number of vertices */
	int		nedges;		/* Number of hyperedges */
	int **		edge_list;	/* To access all vertices of edge e: */
					/* ep1 = edge_list [e];		*/
					/* ep2 = edge_list [e+1];	*/
					/* while (ep1 < ep2) {		*/
					/*	v = *ep1++;		*/
					/* }				*/
	const char **	edge_names;	/* Printable name for each edge */
	int		maxlen;		/* Length of longest edge_names[e] */
	int *		ahull_vec;	/* Affine hull vector */
	struct Polytope *
			poly;		/* STHGP polytope owning edge_list */
};

/*
 * Local Routines
 */

static int64s		Alpha (int n);
static int64s		B (int n);
static int64s		Beta (int n, int k);
static int64s		C (int n);
static int *		classify_edges (struct Hypergraph *	hgp,
					const int *		vclass);
static int *		classify_vertices (struct Hypergraph *	hgp,
					   bitmap_t		a1verts,
					   bitmap_t		a2verts);
static const char **	create_edge_names (int		nverts,
					   int		nedges,
					   int **	edge_list);
static int64s		D (int n);
static void		f (int n, int p, int q, int r);
static void		free_hypergraph (struct Hypergraph * hgp);
static int		int_dot_prod (int		n,
				      const int *	v1,
				      const int *	v2);
static int64s		Gamma (int n, int k);
static long long	long_dot_prod (int		n,
				       const int *	v1,
				       const int *	v2);
static struct Hypergraph *
			make_complete_hypergraph (int nverts);
static int *		make_subtour (struct Hypergraph *	hgp,
				      bitmap_t			verts);
static int64s		Mu (int n, int k);
static int64s		mul_2exp (int64s n, int exp);
static int *		project_subtour_onto_ahull (
					struct Hypergraph *	hgp,
					const int *		avec);
static bool		test_f (struct Hypergraph *	hgp,
				int			n,
				int			p,
				int			q,
				int			r);
static void		test_f_driver (int max_n);
static void		test_alpha_beta (int max_n);
static int64s		W (int n, int p, int q, int r);

/*
 * Main routine.
 */

	int
main (

int		argc,
char **		argv
)
{
int		n, p, q, r;

	me = argv [0];
	if (argc >= 5) {
		n = atoi (argv [1]);
		p = atoi (argv [2]);
		q = atoi (argv [3]);
		r = atoi (argv [4]);
		f (n, p, q, r);
	}
	else {
		int max_n = 15;

		test_f_driver (max_n);

#if 0
		test_alpha_beta (max_n);
#endif
	}

	/* Eliminate compiler warnings. */
	(void) test_alpha_beta;
	(void) classify_vertices;
	(void) classify_edges;

	exit (0);
}

/*
 * The function f(n,p,q,r).
 * Computes and displays the cosine of the angle from first principles.
 */

	static
	void
f (

int		n,
int		p,
int		q,
int		r
)
{
int			nedges;
long long		numer, mag1, mag2;
bitmap_t		a1verts, a2verts;
struct Hypergraph *	hgp;
int *			a1vec;
int *			a2vec;
int *			a1hat;
int *			a2hat;
double			cos_phi, phi, theta;

	if (n < 3) {
		fprintf (stderr, "Must have n >= 3.\n");
		exit (1);
	}
	if (p < 0) {
		fprintf (stderr, "Must have p >= 0.\n");
		exit (1);
	}
	if (q < 0) {
		fprintf (stderr, "Must have q >= 0.\n");
		exit (1);
	}
	if (r < 0) {
		fprintf (stderr, "Must have r >= 0.\n");
		exit (1);
	}
	if (p + r < 2) {
		fprintf (stderr, "Must have p + r >= 2.\n");
		exit (1);
	}
	if (q + r < 2) {
		fprintf (stderr, "Must have q + r >= 2.\n");
		exit (1);
	}
	if (p + q + r > n) {
		fprintf (stderr, "Must have p + q + r <= n.\n");
		exit (1);
	}

	hgp = make_complete_hypergraph (n);

	nedges	= hgp -> nedges;

	a1verts = ((1 << (p + r)) - 1);
	a2verts	= ((1 << (q + r)) - 1);

	a2verts <<= p;

	a1vec	= make_subtour (hgp, a1verts);
	a2vec	= make_subtour (hgp, a2verts);
	
	a1hat	= project_subtour_onto_ahull (hgp, a1vec);
	a2hat	= project_subtour_onto_ahull (hgp, a2vec);

	numer	= long_dot_prod (nedges, a1hat, a2hat);
	mag1	= long_dot_prod (nedges, a1hat, a1hat);
	mag2	= long_dot_prod (nedges, a2hat, a2hat);

	cos_phi	= ((double) numer) /
		   (sqrt ((double) mag1) * sqrt ((double) mag2));

	printf ("cos(phi) = %Ld / sqrt (%Ld * %Ld) = %.17g\n",
		numer, mag1, mag2, cos_phi);

	phi	= acos (cos_phi);

	theta	= 180.0 - 180.0 * phi / 3.14159265358979323;

	printf ("theta = %.17g\n", theta);

	free (a2hat);
	free (a1hat);
	free (a2vec);
	free (a1vec);
	free_hypergraph (hgp);
}

/*
 * Test the computation of f(n,p,q,r) for all valid combos.
 */

	static
	void
test_f_driver (

int			max_n		//- IN: Maximum value of n
)
{
int			n, k, p, q, r, lim;
bool			problem_detected;
struct Hypergraph *	hgp;

	problem_detected = FALSE;

	for (n = 3; n <= max_n; n++) {
		hgp = make_complete_hypergraph (n);

		for (k = 2; k <= n; k++) {
			for (r = 0; r <= k; r++) {
				lim = n - k - r;
				for (p = 0; p <= lim; p++) {
					q = lim - p;

					if (p + r < 2) continue;
					if (q + r < 2) continue;

					if (test_f (hgp, n, p, q, r)) {
						problem_detected = TRUE;
					}
				}
			}
		}

		free_hypergraph (hgp);
	}
	if (NOT problem_detected) {
		printf ("All f(n,p,q,r) are correct!\n");
	}
}

/*
 * Test computation of f(n,p,q,r).
 * Returns TRUE if any problems were detected.
 */

	static
	bool
test_f (

struct Hypergraph *	hgp,		/* IN: The hypergraph for given n */
int			n,		/* IN: The value n */
int			p,		/* IN: The value p */
int			q,		/* IN: The value q */
int			r		/* IN: The value r */
)
{
int		nverts, nedges;
int64s		alpha_n;
int64s		numer, mag1, mag2, numer_2, mag1_2, mag2_2;
bitmap_t	pmask, qmask, rmask, a1mask, a2mask;
bool		have_problem;
bool		header_printed;
int *		a1vec;
int *		a2vec;
int *		a1hat;
int *		a2hat;

	FATAL_ERROR_IF (hgp EQ NULL);
	FATAL_ERROR_IF (n < 3);
	FATAL_ERROR_IF ((p < 0) OR (q < 0) OR (r < 0));
	FATAL_ERROR_IF (p + q + r > n);

	nverts	= hgp -> nverts;
	nedges	= hgp -> nedges;

	FATAL_ERROR_IF (n NE nverts);

	/* Compute pmask, qmask, rmask. */
	pmask = (1 << p) - 1;
	qmask = (1 << q) - 1;
	rmask = (1 << r) - 1;

	/* (Shared vertices (of rmask) are in the middle.) */
	rmask <<= p;
	qmask <<= (p + r);

	a1mask	= pmask | rmask;
	a2mask	= qmask | rmask;

	a1vec	= make_subtour (hgp, a1mask);
	a2vec	= make_subtour (hgp, a2mask);

	a1hat	= project_subtour_onto_ahull (hgp, a1vec);
	a2hat	= project_subtour_onto_ahull (hgp, a2vec);

	numer	= long_dot_prod (nedges, a1hat, a2hat);
	mag1	= long_dot_prod (nedges, a1hat, a1hat);
	mag2	= long_dot_prod (nedges, a2hat, a2hat);

	alpha_n	= Alpha(n);

	numer_2	= Alpha(n) * W(n,p,q,r) - Beta(n,p+r)*Beta(n,q+r);
	mag1_2	= Mu (n,p+r);
	mag2_2	= Mu (n,q+r);

	if (alpha_n > 0) {
		FATAL_ERROR_IF ((numer % alpha_n) NE 0);
		FATAL_ERROR_IF ((mag1  % alpha_n) NE 0);
		FATAL_ERROR_IF ((mag2  % alpha_n) NE 0);
		numer /= alpha_n;
		mag1  /= alpha_n;
		mag2  /= alpha_n;
	}

	have_problem	= FALSE;
	header_printed	= FALSE;
	if ((numer NE numer_2) OR
	    (mag1  NE mag1_2 ) OR
	    (mag1  NE mag1_2 )) {
		/* Houston... we have a problem. */
		have_problem = TRUE;

		if (NOT header_printed) {
			printf ("a1.a2 (n=%d, p=%d, q=%d, r=%d):", n, p, q, r);
			header_printed = TRUE;
		}
		printf (" %lld/sqrt (%lld %lld) : %lld/sqrt(%lld %lld)\n",
			numer, mag1, mag2, numer_2, mag1_2, mag2_2);
	}

	free (a2hat);
	free (a1hat);
	free (a2vec);
	free (a1vec);

	return (have_problem);
}

/*
 * Return Alpha(n).
 */

	static
	int64s
Alpha (

int		n
)
{
int64s		res;

	res = (n * n - 3*n + 4);
	res = mul_2exp (res, n - 2);
	--res;
	return (res);
}

/*
 * Return Beta(n,k).
 */

	static
	int64s
Beta (

int		n,
int		k
)
{
int64s		res, t1, t2, t3;

	t1 = Gamma (n, k);
	t2 = mul_2exp (n - k, n - k - 1);
	t3 = D(k);

	res = t1 + t2 * t3;

	return (res);
}

/*
 * Return Gamma(n,k).
 */

	static
	int64s
Gamma (

int		n,
int		k
)
{
int64s		t1, res;

	t1	= Alpha (k);
	res	= mul_2exp (t1, n - k);

	return (res);
}

/*
 * Return W(n,p,q,r).
 */

	static
	int64s
W (

int		n,
int		p,
int		q,
int		r
)
{
int64s		t1, t2, t3, t4, res;

	t1	= Alpha (r);
	t1	= mul_2exp (t1, p + q);
	t2	= D (p) * D (q);
	t3	= B (p) * B (q) * C (r);
	t4	= B (p + q) * D (r);

	res	= t1 + t2 + t3 + t4;
	res	= mul_2exp (res, n - p - q - r);

	return (res);
}

/*
 * Return Mu(n,k).
 */

	static
	int64s
Mu (

int		n,
int		k
)
{
int64s		t1, t2, t3, res;

	t1	= Alpha (n);
	t2	= Gamma (n, k);
	t3	= Beta (n, k);

	res	= t1 * t2 - t3 * t3;

	return (res);
}

/*
 * Return B(n).
 */

	static
	int64s
B (

int		n
)
{
int64s		res;

	res = mul_2exp (n, n - 1);

	return (res);
}

/*
 * Return C(n).
 */

	static
	int64s
C (

int		n
)
{
int64s		res;

	res = 1;
	res <<= n;
	--res;

	return (res);
}

/*
 * Return D(n).
 */

	static
	int64s
D (

int		n
)
{
int64s		res;

	res = B(n) - C(n);

	return (res);
}

/*
 * Return n*2^exp.
 */

	static
	int64s
mul_2exp (

int64s		n,
int		exp
)
{
	if (exp > 0) {
		n <<= exp;
	}
	else if (exp < 0) {
		exp = - exp;
		while (exp > 0) {
			FATAL_ERROR_IF ((n & 0x01) NE 0);
			n >>= 1;
			--exp;
		}
	}
	return (n);
}

/*
 * Test the alpha(n) and beta(n,k) functions by computing the corresponding
 * dot-products c.c and a.c.
 */

	static
	void
test_alpha_beta (

int			max_n		//- IN: Maximum value of n
)
{
int			i, k, n, nedges, c_dot_c, a_dot_a, a_dot_c;
bitmap_t		vert_mask;
bool			problem_detected;
struct Hypergraph *	hgp;
int *			avec;
int *			cvec;
int *			ahat;

	printf ("\n"
		"|V|\t|E|\tc.c\n"
		"------------------------\n");
	for (n = 3; n <= max_n; n++) {
		hgp = make_complete_hypergraph (n);

		nedges	= hgp -> nedges;
		cvec	= hgp -> ahull_vec;

		/* Compute affine hull vector mag-squared. */
		c_dot_c = int_dot_prod (nedges, cvec, cvec);

		printf ("%d\t%d\t%d\n", hgp -> nverts, hgp -> nedges, c_dot_c);

		free_hypergraph (hgp);
	}

	printf ("\n"
		"n\tk\ta.a\ta.c\n"
		"------------------------\n");
	for (n = 3; n <= max_n; n++) {
		hgp = make_complete_hypergraph (n);

		nedges	= hgp -> nedges;
		cvec	= hgp -> ahull_vec;

		for (k = 2; k < n; k++) {
			/* Make vertex mask containing first k vertices. */
			vert_mask = ((1 << k) - 1);

			/* Make k-vertex subtour. */
			avec = make_subtour (hgp, vert_mask);

			a_dot_a = int_dot_prod (nedges, avec, avec);
			a_dot_c = int_dot_prod (nedges, avec, cvec);

			printf ("%d\t%d\t%d\t%d\n",
				n, k, a_dot_a, a_dot_c);

			free (avec);
		}
		printf ("\n");

		free_hypergraph (hgp);
	}

	problem_detected = FALSE;

	for (n = 3; n <= max_n; n++) {
		hgp = make_complete_hypergraph (n);

		nedges	= hgp -> nedges;
		cvec	= hgp -> ahull_vec;

		for (k = 2; k < n; k++) {
			/* Make vertex mask containing first k vertices. */
			vert_mask = ((1 << k) - 1);

			/* Make k-vertex subtour. */
			avec = make_subtour (hgp, vert_mask);

			ahat = project_subtour_onto_ahull (hgp, avec);

			i = int_dot_prod (nedges, ahat, cvec);
			if (i NE 0) {
				printf ("Warning: n=%d, k=%d: ahat.c NE 0.\n",
					n, k);
				problem_detected = TRUE;
			}

			free (ahat);
			free (avec);
		}

		free_hypergraph (hgp);
	}
	if (NOT problem_detected) {
		printf ("All projected subtours are normal to affine hull!\n");
	}
}

/*
 * Return an object representing the complete hypergraph H=(V,E)
 * such that |V| = nverts.
 */

	static
	struct Hypergraph *
make_complete_hypergraph (

int		nverts		/* IN: number of vertices in complete graph */
)
{
int			j, k;
int			nedges, maxlen;
struct Hypergraph *	hgp;
struct Polytope *	poly;
int **			edge_list;
const char **		edge_names;
int *			ahull_vec;
int *			ip1;
int *			ip2;

	if (nverts < 2) {
		fprintf (stderr, "STHGP(n) requires n >= 2.\n");
		return (NULL);
	}

	hgp = NEW (struct Hypergraph);
	memset (hgp, 0, sizeof (*hgp));

	poly = make_STHGP_polytope (nverts);

	nedges	= (1 << nverts) - nverts - 1;

	FATAL_ERROR_IF (nedges NE poly -> dim);

	edge_list = get_STHGP_edge_list (poly);

	edge_names = create_edge_names (nverts, nedges, edge_list);

	maxlen = 0;
	for (j = 0; j < nedges; j++) {
		k = strlen (edge_names [j]);
		if (k > maxlen) {
			maxlen = k;
		}
	}

	/* Create the affine hull vector. */
	ahull_vec	= NEWA (nedges, int);
	for (j = 0; j < nedges; j++) {
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		k = ip2 - ip1;
		ahull_vec [j] = k - 1;
	}

	hgp -> nverts		= nverts;
	hgp -> nedges		= nedges;
	hgp -> edge_list	= edge_list;
	hgp -> edge_names	= edge_names;
	hgp -> maxlen		= maxlen;
	hgp -> ahull_vec	= ahull_vec;
	hgp -> poly		= poly;

	return (hgp);
}

/*
 * Create names for each of the edges.
 */

	static
	const char **
create_edge_names (

int		nverts,		/* IN: number of vertices */
int		nedges,		/* IN: number of edges */
int **		edge_list	/* IN: vertices contained in each edge */
)
{
int		j, v;
const char **	edge_names;
char *		s;
int *		ip1;
int *		ip2;
char		buffer [128];

	edge_names = NEWA (nedges, const char *);

	if (nverts <= 10) {
		/* We can use single digits for each vertex. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				*s++ = '0' + v;
			}
			*s = '\0';
			edge_names [j] = strdup (buffer);
		}
	}
	else {
		/* At least one vertex requires more than 1 digit. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				sprintf (s, "_%d", v);
				s = strchr (s, '\0');
			}
			edge_names [j] = strdup (buffer);
		}
	}
	return (edge_names);
}

/*
 * Return the left-hand-side vector for the subtour corresponding to
 * the given set of vertices.
 */

	static
	int *
make_subtour (

struct Hypergraph *	hgp,	/* IN: Hypergraph to make subtour for */
bitmap_t		verts	/* IN: Set of vertices in subtour */
)
{
int		j, k, v, nverts, nedges;
int **		edge_list;
int *		sec;
int *		ip1;
int *		ip2;

	FATAL_ERROR_IF (hgp EQ NULL);

	nverts		= hgp -> nverts;
	nedges		= hgp -> nedges;
	edge_list	= hgp -> edge_list;

	FATAL_ERROR_IF (verts >= (1 << nverts));

	sec = NEWA (nedges, int);
	for (j = 0; j < nedges; j++) {
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		k = 0;
		while (ip1 < ip2) {
			v = *ip1++;
			FATAL_ERROR_IF ((v < 0) OR (v >= nverts));
			if ((verts & (1 << v)) NE 0) {
				/* Vertex v is in given subtour S. */
				++k;
			}
		}
		if (k > 0) {
			--k;
		}
		sec [j] = k;
	}

	return (sec);
}

/*
 * Project the given subtour (avec) onto the affine hull:
 *
 *	ahat = (cvec.cvec) * avec - (avec.cvec) * cvec,
 *
 * where cvec is the affine hull.
 */

	static
	int *
project_subtour_onto_ahull (

struct Hypergraph *	hgp,		/* IN: Hypergraph */
const int *		avec		/* IN: Subtour to project */
)
{
int		j, nedges, a_dot_c, c_dot_c;
int *		ahat;
int *		cvec;

	nedges	= hgp -> nedges;
	cvec	= hgp -> ahull_vec;

	a_dot_c	= int_dot_prod (nedges, avec, cvec);
	c_dot_c	= int_dot_prod (nedges, cvec, cvec);

	ahat	= NEWA (nedges, int);
	for (j = 0; j < nedges; j++) {
		ahat [j] = c_dot_c * avec [j] - a_dot_c * cvec [j];
	}

	return (ahat);
}

/*
 * Return a vector that properly classifies each vertex with respect to the
 * given pair of subtours S1 and S2 (as specified by a1verts and a2verts).
 *
 *	0	==> (v \notin S1) AND (v \notin S2)
 *	1	==> (v \in    S1) AND (v \notin S2)
 *	2	==> (v \notin S1) AND (v \in    S2)
 *	3	==> (v \in    S1) AND (v \in    S2)
 */

	static
	int *
classify_vertices (

struct Hypergraph *	hgp,		/* IN: Hypergraph */
bitmap_t		a1verts,	/* IN: Vertices in S1 */
bitmap_t		a2verts		/* IN: Vertices in S2 */
)
{
int		i, nverts, bits;
int *		vclass;

	nverts	= hgp -> nverts;
	vclass	= NEWA (nverts, int);

	for (i = 0; i < nverts; i++) {
		bits = 0;
		if ((a1verts & (1 << i)) NE 0) {
			bits |= 0x01;
		}
		if ((a2verts & (1 << i)) NE 0) {
			bits |= 0x02;
		}
		vclass [i] = bits;
	}

	return (vclass);
}

/*
 * Return a vector that properly classifies each edge.
 */

	static
	int *
classify_edges (

struct Hypergraph *	hgp,		/* IN: Hypergraph */
const int *		vclass		/* IN: Class of each vertex */
)
{
int		j, k, v, nverts, nedges, bits;
int **		edge_list;
int *		eclass;
int *		ip1;
int *		ip2;

	nverts		= hgp -> nverts;
	nedges		= hgp -> nedges;
	edge_list	= hgp -> edge_list;

	eclass	= NEWA (nedges, int);

	for (j = 0; j < nedges; j++) {
		bits = 0;
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		while (ip1 < ip2) {
			v = *ip1++;
			FATAL_ERROR_IF ((v < 0) OR (v >= nverts));

			k = vclass [v];
			bits |= (1 << k);
		}
		eclass [j] = bits;
	}

	return (eclass);
}

/*
 * Free up the given Hypergraph object.
 */

	static
	void
free_hypergraph (

struct Hypergraph *	hgp	/* IN: Hypergraph to free */
)
{
int			i, nedges;
const char **		edge_names;
const char *		name;
struct Polytope *	poly;

	if (hgp NE NULL) {
		edge_names = hgp -> edge_names;
		if (edge_names NE NULL) {
			nedges	= hgp -> nedges;
			for (i = 0; i < nedges; i++) {
				name = edge_names [i];
				if (name NE NULL) {
					free ((void *) name);
					edge_names [i] = NULL;
				}
			}
			free (edge_names);
		}
		poly = hgp -> poly;
		if (poly NE NULL) {
			/* Freeing the Polytope also frees edge_list. */
			poly -> free (poly);
		}

		memset (hgp, 0, sizeof (*hgp));

		free (hgp);
	}
}

/*
 * Return the dot-product of two "int" vectors.
 */

	static
	int
int_dot_prod (

int		n,		/* IN: Length of vectors */
const int *	v1,		/* IN: First vector */
const int *	v2		/* IN: Second vector */
)
{
int		i, dprod;

	dprod = 0;
	for (i = 0; i < n; i++) {
		dprod += (v1 [i] * v2 [i]);
	}
	return (dprod);
}

/*
 * Return the dot-product of two "int" vectors using "long long" arithmetic.
 */

	static
	long long
long_dot_prod (

int		n,		/* IN: Length of vectors */
const int *	v1,		/* IN: First vector */
const int *	v2		/* IN: Second vector */
)
{
int		i;
long long	dprod, x1, x2;

	dprod = 0;
	for (i = 0; i < n; i++) {
		x1	= v1 [i];
		x2	= v2 [i];
		dprod += (x1 * x2);
	}
	return (dprod);
}
