/**********************************************************************

	$Id: validate_sthgp_sec_cd2_dx.c,v 1.1 2025/01/26 02:32:02 warme Exp $

	File:	validate_sthgp_sec_cd2_dx.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Plot or test DX_e values and sums of (DX_e)^2 values
	for the STHGP(n) subtour Centroid Distance.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

#include "fatal.h"
#include <gmp.h>
#include "logic.h"
#include <math.h>
#include "s3itypes.h"
#include <stdlib.h>
#include <stdio.h>
#include "sthgp_indicators.h"
#include "sthgp_sec_cd2_dx.h"
#include <string.h>


/*
 * Global Routines
 */

int		main (int, char **);

char *		me;


/*
 * Local Routines
 */

static void		test_sums (int n);
static void		usage (void);

/*
 * The main routine.
 *
 * Given the number of vertices N, it computes the centroid distance
 * for the K-vertex subtour for each 2 <= K <= N-1.
 */

	int
main (

int		argc,
char **		argv
)
{
int		n;

	me = argv [0];

	if (argc NE 2) {
		usage ();
	}

	n = atoi (argv [1]);
	if (n < 3) {
		fprintf (stderr, "%s: must have at least 3 vertices.\n", me);
		exit (1);
	}

	test_sums (n);

	sthgp_indicators_shutdown ();

	exit (0);
}

/*
 * Display command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr, "Usage: %s NVerts\n", me);
	exit (1);
}

/*
 * Test the correctness of the formulae for sums of DX^2 values.
 */

	static
	void
test_sums (

int		n
)
{
int		k;
bool		any_mismatches;
mpq_t		DIST2, TMP1, TMP2, TMP3;

	FATAL_ERROR_IF (n < 3);

	mpq_init (DIST2);
	mpq_init (TMP1);
	mpq_init (TMP2);
	mpq_init (TMP3);

	any_mismatches = FALSE;

	for (k = 2; k < n; k++) {
		sthgp_subtour_cd2 (DIST2, n, k);
#if 0
		printf ("K = %d\tdist2 = ", k);
		mpz_out_str (stdout, 10, mpq_numref (DIST2));
		printf (" / ");
		mpz_out_str (stdout, 10, mpq_denref (DIST2));
		{ double d2;
		  d2 = mpq_get_d (DIST2);
		  printf (" = %.17g\n", sqrt (d2));
		}
#endif

		sthgp_subtour_cd2_dx_Delta_x_1 (TMP1, n, k);
		sthgp_subtour_cd2_dx_Delta_x_2 (TMP2, n, k);

		mpq_add (TMP3, TMP1, TMP2);

		if (mpq_cmp (TMP3, DIST2) NE 0) {
			printf ("(%d,%d) MISMATCH!\n", n, k);
			printf ("\t%.17g\t%.17g\n",
				mpq_get_d (DIST2),
				mpq_get_d (TMP3));
			any_mismatches = TRUE;
		}
	}
	if (NOT any_mismatches) {
		printf ("All results match!\n");
	}

	mpq_clear (TMP3);
	mpq_clear (TMP2);
	mpq_clear (TMP1);
	mpq_clear (DIST2);
}
