/***********************************************************************

	$Id: validate_tsp.c,v 1.1 2025/01/26 02:32:03 warme Exp $

	File:	validate_tsp.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Validate all indicators for TSP(n).

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "config.h"

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "math.h"
#include "new.h"
#include "polytope.h"
#include "qpdist2.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tsp_indicators.h"
#include "tsp_polytope.h"

/*
 * Local Types
 */

struct Comb3T {			/* Parameters for a 3-toothed comb */
	char		b1;
	char		t1;
	char		b2;
	char		t2;
	char		b3;
	char		t3;
	char		h;
	char		o;
	struct Comb3T *	next;
};

/*
 * Local Functions
 */

static struct Comb3T *
		enumerate_3_toothed_combs (int n);
static void	free_3_toothed_combs (struct Comb3T * combs);
static struct Comb3T **
		recurse_3tc (int		depth,
			     int		max_depth,
			     int		nleft,
			     char *		stack,
			     struct Comb3T **	hookp);
static void	validate_tsp_indicators (int n);

char *		me;

/*
 * Main program to validate all indicators for TSP(n).
 */

	int
main (

int		argc,
char **		argv
)
{
int		n;

	me = argv [0];

	printf ("Validating strength indicators for TSP(n).\n");

#ifndef CPLEX
	printf ("\n"
		"Warning: No CPLEX"
		" -- skipping validation of Centroid Distances.\n");
#endif

	for (n = 3; n <= 12; n++) {
		validate_tsp_indicators (n);
	}

	exit (0);
}

/*
 * Validate all indicators for TSP(n).
 */

	static
	void
validate_tsp_indicators (

int		n	/* IN: Number of vertices */
)
{
int			j, k, e, dim, nmasks, nf;
const bitmap_t *	mask;
struct Polytope *	poly;
struct IVecSet *	xpts;
struct Constraint *	cp;
int *			list;
int *			xpsum;
struct Comb3T *		combs;
struct Comb3T *		combp;
mpz_t			COUNT;
mpq_t			DIST2;
char			cname [256];

	printf ("\n"
		"n = %d:\n", n);

	mpz_init (COUNT);
	mpq_init (DIST2);

	/* Make the polytope TSP(n). */
	poly	= make_TSP_polytope (n);

	dim	= poly -> dim;

	/* Get list of all extreme points. */
	xpts	= poly -> get_extreme_points (poly);

	FATAL_ERROR_IF (xpts -> dim NE dim);

	/* Get number of extreme points via closed-form formula. */
	tsp_num_extreme_points (COUNT, n);

	gmp_printf ("  Extreme Points:\t%Zd", COUNT);
	if (mpz_cmp_ui (COUNT, xpts -> nvecs) NE 0) {
		printf ("\t*** Enumeration gives %d ***", xpts -> nvecs);
	}
	printf ("\n");

	/* Let e \in E such the |e| = k.  Get the number of tours	*/
	/* containing edge e using closed-form formula.			*/
	tsp_num_edge_extreme_points (COUNT, n);
	gmp_printf ("  Centroid Edge Counts: %Zd\n", COUNT);

	/* Add up all extreme points. */
	xpsum = NEWA (dim, int);
	memset (xpsum, 0, dim * sizeof (xpsum [0]));
	nmasks	= xpts -> nmasks;
	mask	= &(xpts -> array [0]);
	for (j = 0; j < xpts -> nvecs; j++, mask += nmasks) {
		for (e = 0; e < dim; e++) {
			if (BITON (mask, e)) {
				++(xpsum [e]);
			}
		}
	}
	/* Verify that each edge is contained in the correct number of	*/
	/* tours.							*/
	for (e = 0; e < dim; e++) {
		if (mpz_cmp_ui (COUNT, xpsum [e]) NE 0) {
			printf ("\t*** Count mismatch: edge %d = %d\n",
				e, xpsum [e]);
		}
	}
	free (xpsum);

	printf ("  Non-Negativity Extreme Point Counts:\n");
	tsp_nonneg_ep_count (COUNT, n);

	cp = poly -> get_constraint (poly, "nonneg");

	nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);
	free (list);
	free_constraint_list (cp);

	gmp_printf ("\t%Zd", COUNT);
	if (mpz_cmp_ui (COUNT, nf) NE 0) {
		printf ("\t*** Enumeration gives %d ***", nf);
	}
	printf ("\n");

	printf ("  Subtour Extreme Point Counts:\n");
	for (k = 2; k <= n - 2; k++) {
		tsp_subtour_ep_count (COUNT, n, k);

		sprintf (cname, "subtour-%d", k);
		cp = poly -> get_constraint (poly, cname);

		nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);
		free (list);
		free_constraint_list (cp);

		gmp_printf ("    k = %d:\t%Zd", k, COUNT);
		if (mpz_cmp_ui (COUNT, nf) NE 0) {
			printf ("\t*** Enumeration gives %d ***", nf);
		}
		printf ("\n");
	}

	combs = enumerate_3_toothed_combs (n);

#if 0
	/* Note: the closed-form formula for EPR of 3-Toothed Combs	*/
	/* is WRONG!  There are tours incident to the comb whose	*/
	/* structure does not match any of the "primtive comb" tours,	*/
	/* so this formula undercounts the incident tours.		*/
	/* Enabling this code verifies these facts.			*/
	if (combs NE NULL) {
		printf ("  3-Toothed Comb Extreme Point Counts:\n");
		for (combp = combs; combp NE NULL; combp = combp -> next) {
			/* Count incident extreme points using	*/
			/* closed-form formula.			*/
			tsp_3tc_ep_count (COUNT,
					  combp -> b1,
					  combp -> t1,
					  combp -> b2,
					  combp -> t2,
					  combp -> b3,
					  combp -> t3,
					  combp -> h,
					  combp -> o);

			/* Get constraint for this 3-toothed comb. */
			sprintf (cname,
				 "3comb-%d-%d-%d-%d-%d-%d-%d",
				 combp -> b1,
				 combp -> t1,
				 combp -> b2,
				 combp -> t2,
				 combp -> b3,
				 combp -> t3,
				 combp -> h);
			cp = poly -> get_constraint (poly, cname);

			nf = get_extreme_points_incident_to_constraint (&list, cp, xpts);
			free (list);
			free_constraint_list (cp);

			gmp_printf ("    %d %d %d %d %d %d %d %d:\t%Zd",
				    combp -> b1,
				    combp -> t1,
				    combp -> b2,
				    combp -> t2,
				    combp -> b3,
				    combp -> t3,
				    combp -> h,
				    combp -> o,
				    COUNT);
			if (mpz_cmp_ui (COUNT, nf) NE 0) {
				printf ("\t*** Enumeration gives %d ***", nf);
			}
			printf ("\n");
		}
	}
#endif

	/* We need CPLEX in order to compute Centroid Distances	*/
	/* (squared) via quadratic programming.			*/

#if CPLEX
	if (n >= 4) {
		{ double	relerr, max_relerr, dist2, xdist2;
		  printf ("  Non-Negativity Centroid Distance Squared:\n");
		  max_relerr = 0.0;

		  tsp_nonneg_cd2 (DIST2, n);

		  cp = poly -> get_constraint (poly, "nonneg");

		  dist2 = cd2_via_qp (cp, poly, QPDIST2_NORMAL_CD2);

		  xdist2 = mpq_get_d (DIST2);

		  relerr = fabs (dist2 - xdist2) / fabs (xdist2);

		  printf ("\t%.17g\t%.17g\t%.17g\n",
			  xdist2, dist2, relerr);
		  if (relerr > max_relerr) {
			max_relerr = relerr;
		  }
		  printf ("\tMax Relative Error:\t%.17g\n", max_relerr);

		  printf ("  Subtour Centroid Distance Squared:\n");
		  max_relerr = 0.0;
		  for (k = 2; k <= n - 2; k++) {
			tsp_subtour_cd2 (DIST2, n, k);

			sprintf (cname, "subtour-%d", k);
			cp = poly -> get_constraint (poly, cname);

			dist2 = cd2_via_qp (cp, poly, QPDIST2_NORMAL_CD2);

			xdist2 = mpq_get_d (DIST2);

			relerr = fabs (dist2 - xdist2) / fabs (xdist2);

			printf ("    k = %d:\t%.17g\t%.17g\t%.17g\n",
				k, xdist2, dist2, relerr);
			if (relerr > max_relerr) {
				max_relerr = relerr;
			}
		  }
		  printf ("\tMax Relative Error:\t%.17g\n", max_relerr);
		}
	}

	if (combs NE NULL) {
		{ double	relerr, max_relerr, dist2, xdist2;
		  printf ("  3-Toothed Comb Centroid Distance Squared:\n");
		  max_relerr = 0.0;
		  for (combp = combs; combp NE NULL; combp = combp -> next) {
			/* Compute CD squared using closed-form formula. */
			tsp_3tc_cd2 (DIST2,
				     combp -> b1,
				     combp -> t1,
				     combp -> b2,
				     combp -> t2,
				     combp -> b3,
				     combp -> t3,
				     combp -> h,
				     combp -> o);

			/* Get constraint for this 3-toothed comb. */
			sprintf (cname,
				 "3comb-%d-%d-%d-%d-%d-%d-%d",
				 combp -> b1,
				 combp -> t1,
				 combp -> b2,
				 combp -> t2,
				 combp -> b3,
				 combp -> t3,
				 combp -> h);
			cp = poly -> get_constraint (poly, cname);

			dist2 = cd2_via_qp (cp, poly, QPDIST2_WEAK_CD2);

			xdist2 = mpq_get_d (DIST2);

			relerr = fabs (dist2 - xdist2) / fabs (xdist2);

			printf ("    %d %d %d %d %d %d %d %d:\t%.17g\t%.17g\t%.17g\n",
				combp -> b1,
				combp -> t1,
				combp -> b2,
				combp -> t2,
				combp -> b3,
				combp -> t3,
				combp -> h,
				combp -> o,
				xdist2, dist2, relerr);
			if (relerr > max_relerr) {
				max_relerr = relerr;
			}
		  }
		  printf ("\tMax Relative Error:\t%.17g\n", max_relerr);
		}
	}
#endif

	free_3_toothed_combs (combs);

	/* Free up the polytope. */
	poly -> free (poly);

	mpq_clear (DIST2);
	mpz_clear (COUNT);
}

/*
 * Return a list of all valid 3-toothed comb parameters for given n.
 */

	static
	struct Comb3T *
enumerate_3_toothed_combs (

int		n	/* IN: Number of vertices */
)
{
struct Comb3T *		combs;
struct Comb3T **	hookp;
char			stack [8];

	if (n < 6) {
		/* No 3-toothed combs unless n >= 6. */
		return (NULL);
	}

	combs	= NULL;
	hookp	= &combs;

	/* Note: we implicitly allocate one initial vertex each	*/
	/* to b1, t1, b2, t2, b3 and t3, partitioning the n-6	*/
	/* remaining vertices among all 8 subsets.		*/

	hookp = recurse_3tc (0, 7, n - 6, stack, hookp);

	return (combs);
}

/*
 * Recursive portion of 3-toothed comb enumeration.
 */

	static
	struct Comb3T **
recurse_3tc (

int			depth,		/* IN: Current recursion depth */
int			max_depth,	/* IN: Max recursion depth */
int			nleft,		/* IN: Num vertices left */
char *			stack,		/* IN/OUT: Stack of subset counts */
struct Comb3T **	hookp		/* IN: Hook ptr for new combs */
)
{
int			i;
struct Comb3T *		comb;

	if (depth >= max_depth) {
		stack [depth] = nleft;

		/* We now have a complete solution. */
		comb = NEW (struct Comb3T);
		comb -> b1	= 1 + stack [0];
		comb -> t1	= 1 + stack [1];
		comb -> b2	= 1 + stack [2];
		comb -> t2	= 1 + stack [3];
		comb -> b3	= 1 + stack [4];
		comb -> t3	= 1 + stack [5];
		comb -> h	= 0 + stack [6];
		comb -> o	= 0 + stack [7];
		comb -> next	= NULL;
		*hookp = comb;
		hookp = &(comb -> next);
		return (hookp);
	}

	for (i = 0; i <= nleft; i++) {
		stack [depth] = i;

		hookp = recurse_3tc (depth + 1,
				     max_depth,
				     nleft - i,
				     stack,
				     hookp);
	}

	return (hookp);
}

/*
 * Free up the given list of 3-toothed comb parameters.
 */

	static
	void
free_3_toothed_combs (

struct Comb3T *		combs	/* IN: List of 3-toothed combs to free */
)
{
struct Comb3T *		tmp;

	while (combs NE NULL) {
		tmp = combs -> next;
		memset (combs, 0, sizeof (*combs));
		free (combs);
		combs = tmp;
	}
}
